//-----------------------------------------------------------------------------
// Title         : HT Channel Interface
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////
// File          : csco_ht_channel_interface.e
// Author        : Brian Van Essen  <vanessen@cisco.com>
// Created       : 25.01.2001
// Last modified : 30.04.2001
////////////////////////////////////////////////////////////////////////////
// Description :
// The Channel is the logical interface to the HT link.  It handles all HT traffic at the packet level
// an manages the inbound and outbound packet ordering.  It takes the HT packets and helps to
// assemble/disassemble the packet for interaction with the link interface.  The channel
// allows the device to present packets to be transmitted and with handle the scheduling.
// Additionally the channel will present received packets to the device.
////////////////////////////////////////////////////////////////////////////////
// Modification history :
// 25.01.2001 : created
////////////////////////////////////////////////////////////////////////////////

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****

<'
struct rw_sequence_item_t {
  command: csco_ht_packet_encoding;
    keep soft command == Nop;
  !addr: uint(bits:35);
  !accessing_next_cache_line: bool;
};

struct last_command_with_data_t {
  command: csco_ht_packet_encoding;
  data_length: uint;
  data_time: uint;
};

////////////////////////////////////////////////////////////////////////////////
// The Channel is the logical interface to the HT link.  It handles all HT traffic at the packet level
// an manages the inbound and outbound packet ordering.  It takes the HT packets and helps to
// assemble/disassemble the packet for interaction with the link interface.  The channel
// allows the device to present packets to be transmitted and with handle the scheduling.
// Additionally the channel will present received packets to the device.
////////////////////////////////////////////////////////////////////////////////
unit csco_ht_channel_interface {
  parent_device : csco_ht_device;
  keep parent_device == get_enclosing_unit(csco_ht_device);
  event csco_ht_queues_serviced;
  event csco_ht_rx_packet_retired;
  ////////////////////////////////////////////////////////////////////////////////
  // A number used to identifu which interface this is on given device.
  channel_identifier : uint;
  // A reference to the CSR and general control block.
  control_block : csco_ht_device_control_block;
  keep control_block == parent_device.control_block;
  ////////////////////////////////////////////////////////////////////////////////
  // General Link Information
  ////////////////////////////////////////////////////////////////////////////////
  // Instantiate the link associated with this device.
  csco_ht_link : csco_ht_link_interface is instance;
  keep csco_ht_link.link_identifier == channel_identifier;
  ////////////////////////////////////////////////////////////////////////////////
  // Link Initialization Information
  ////////////////////////////////////////////////////////////////////////////////
  // Number of cycles to wait after RESET is deasserted before initiating the SYNC SEQUENCE.
  initiate_sync_sequence_delay : uint;
  keep soft initiate_sync_sequence_delay == 0;
  // This is a flag to indicate that the receiver has detected the assertion of the targets ctl signal, hence the sync sequence can begin.
  initiate_sync_sequence : bool;
  keep initiate_sync_sequence == FALSE;
  // Number of cycles for the sync sequence, minimun of CSCO_HT_MIN_SYNC_SEQUENCE_CYCLES
  sync_sequence_length : uint;
  keep sync_sequence_length >= CSCO_HT_MIN_SYNC_SEQUENCE_CYCLES;
  keep soft sync_sequence_length == CSCO_HT_MIN_SYNC_SEQUENCE_CYCLES;
  ////////////////////////////////////////////////////////////////////////////////
  // Link Reset Information
  ////////////////////////////////////////////////////////////////////////////////
  // The is a counter to track how many sync packets have been TX by this channel
  !num_sync_packets_tx : uint;
  ////////////////////////////////////////////////////////////////////////////////
  // Link TX control Information
  ////////////////////////////////////////////////////////////////////////////////
  // Monitor the level of the tx_scheduling_queue. If the watermark level is exceeded print a message.
  tx_scheduling_queue_watermark : uint;
  keep soft tx_scheduling_queue_watermark == 10;

  // These are counters that are used to track how many buffers are available on the other side of this link.
  // They must be at least 4 bits wide and are not allowed to overflow.
  available_target_Posted_queue_buffers : uint;
  keep available_target_Posted_queue_buffers == 0;
  available_target_NonPosted_queue_buffers : uint;
  keep available_target_NonPosted_queue_buffers == 0;
  available_target_Response_queue_buffers : uint;
  keep available_target_Response_queue_buffers == 0;

  available_target_Posted_Data_queue_buffers : uint;
  keep available_target_Posted_Data_queue_buffers == 0;
  available_target_NonPosted_Data_queue_buffers : uint;
  keep available_target_NonPosted_Data_queue_buffers == 0;
  available_target_Response_Data_queue_buffers : uint;
  keep available_target_Response_Data_queue_buffers == 0;

  !available_target_Posted_queue_buffers_bottom: bool;
  !available_target_NonPosted_queue_buffers_bottom: bool;
  !available_target_Response_queue_buffers_bottom: bool;
  !available_target_Posted_Data_queue_buffers_bottom: bool;
  !available_target_NonPosted_Data_queue_buffers_bottom: bool;
  !available_target_Response_Data_queue_buffers_bottom: bool;

  ////////////////////////////////////////////////////////////////////////////////
  // Link TX scheduling and Knob Information
  ////////////////////////////////////////////////////////////////////////////////
  // Set up weights that will influence the priority and selection frequency of packets to be transmitted in the tx_scheduling_queue
  // VC queue weights must be positive;
  Posted_VC_queue_weight : uint;
  keep Posted_VC_queue_weight > 0;
  keep soft Posted_VC_queue_weight == 1;
  NonPosted_VC_queue_weight : uint;
  keep NonPosted_VC_queue_weight > 0;
  keep soft NonPosted_VC_queue_weight == 1;
  Response_VC_queue_weight : uint;
  keep Response_VC_queue_weight > 0;
  keep soft Response_VC_queue_weight == 1;
  // A weight to represent the priority of packets earlier in the tx_scheduling_queue to be selected for transmission.
  scheduling_timing_weight : uint;
  keep scheduling_timing_weight >= 0;
  keep soft scheduling_timing_weight == 1;
  // A weight to represent the priority of outstanding data packets to be selected for transmission.
  data_scheduling_weight : uint;
  keep data_scheduling_weight >= 0;
  keep soft data_scheduling_weight == 1;

  // These are a set of knobs that allow a user to determing the exact ordering requirements of the HT IO traffic.
  // These are generated from the ordering section 7.1 from the 1.0 HT spec.
  Posted_Request_with_PassPW_pass_Posted_Request : bool;
  keep soft Posted_Request_with_PassPW_pass_Posted_Request == TRUE;

  NonPosted_Request_with_PassPW_pass_Posted_Request : bool;
  keep soft NonPosted_Request_with_PassPW_pass_Posted_Request == TRUE;
  NonPosted_Request_pass_NonPosted_Request : bool;
  keep soft NonPosted_Request_pass_NonPosted_Request == TRUE;
  NonPosted_Request_pass_Response : bool;
  keep soft NonPosted_Request_pass_Response == TRUE;

  Response_with_PassPW_pass_Posted_Request : bool;
  keep soft Response_with_PassPW_pass_Posted_Request == TRUE;
  Response_pass_Response : bool;
  keep soft Response_pass_Response == TRUE;

  // If this flag is set, the transmitter will release rx buffers in such a ways as to allow the other side of the link
  // to starve.  Once starvation has occured, the transmitter will wait a certain number of cycles before releasing buffers.
  create_target_starvation : bool;
  keep soft create_target_starvation == FALSE;
  starvation_buffer_release_percentage : uint;
  keep starvation_buffer_release_percentage in range [0..100];
  keep soft starvation_buffer_release_percentage == 10;
  // How long to wait for buffer releases once starvation has occurred
// BVE Fix This
  starvation_cycle_count : uint;
  keep starvation_cycle_count == 0;
  // Nop's will be sent as a percentage of the BW.
  nop_bandwidth_percent : uint;
  keep nop_bandwidth_percent in range [1..99];  // Don't allow it to be 0 or 100 percent.
  keep soft nop_bandwidth_percent == 10; // By default set Nop's to consume 10% of the bandwidth.

  // If this flag is set every nop packet will release the maximum available number of buffers.  Otherwise a random number
  // will be released with each packet.
  nop_maximum_buffer_release : bool;
  keep soft nop_maximum_buffer_release == TRUE;

  // This field is the percent chance that a packet that is being transmitted will be interrupted.  If it is, a new packet will
  // be transmitted for this 4 byte window.
  interrupt_tx_packet_percentage : uint;
  keep interrupt_tx_packet_percentage in range [0..100];
  keep soft interrupt_tx_packet_percentage == 25; // By default set the change interrupt to 25%.
  ////////////////////////////////////////////////////////////////////////////////
  // coverage events
  ////////////////////////////////////////////////////////////////////////////////
  event intrp_tx_data_pkt;
  event intrp_tx_ctl_pkt;
  // coverage fields
  ////////////////////////////////////////////////////////////////////////////////
  tx_cover_instance : channel_instance_e;
    keep soft tx_cover_instance == uncovered;
  rx_cover_instance : channel_instance_e;
    keep soft rx_cover_instance == uncovered;

  do_cover_stall (instance: channel_instance_e, location: uint (bits:6),
                  command_data_b: bool, data_stalled: bool) is {};

  do_cover_rw_sequence (instance: channel_instance_e) is {};

  do_cover_buffer (instance: channel_instance_e, virtual_channel: csco_ht_virtual_channel,
                   packet_type: csco_ht_packet_type, available_rx_buffer: uint) is {};

  do_cover_delay (instance: channel_instance_e, command_with_data: last_command_with_data_t) is {};

  do_cover_interrupt_cmd (instance: channel_instance_e) is {};

  size_of_last_tx_data_packet: uint(bits:6);

  !started_sending_last_command_with_data: bool;
  received_last_command_with_data: last_command_with_data_t;
  sent_last_command_with_data: last_command_with_data_t;

  !tx_doublewords_no: uint;
  !rx_doublewords_no: uint;

  first_rw: rw_sequence_item_t;
  second_rw: rw_sequence_item_t;
  third_rw: rw_sequence_item_t;

  ////////////////////////////////////////////////////////////////////////////////
  // Link TX Data Structures
  ////////////////////////////////////////////////////////////////////////////////
  // This is a master queue of all packets to be transmitted.
  // It should be thought of as a FIFO and the position of the packet within the queue reflect the temporal relationship with other packets.
  !tx_scheduling_queue : list of csco_ht_packet;
  // A field to hold the outstanding data packet waiting to be transmitted.
  !pending_tx_data_packet : csco_ht_packet;
  // The current packet being dealt with
  !current_tx_packet : csco_ht_packet;
  // A buffer of bits that is currently being transmitted
  !current_tx_buffer : list of bit;
  // The type of packet that is in the tx_buffer and in process of being transmitted.
  !current_tx_buffer_type : csco_ht_packet_type;

  // These are fields used to hold interrupted ctl or data packets.
  // Only one data packet can be outstanding at any single time.
  !interrupted_tx_data_buffer : list of bit;
  !interrupted_tx_data_buffer_type : csco_ht_packet_type;
  // A control packet can only be interrupted by a stall packet.
  !interrupted_tx_ctl_buffer : list of bit;
  !interrupted_tx_ctl_buffer_type : csco_ht_packet_type;
  ////////////////////////////////////////////////////////////////////////////////
  // Link RX control Information
  ////////////////////////////////////////////////////////////////////////////////
  // The number of buffers that exist at our receiver..
  max_rx_Posted_queue_buffers : uint;
  keep max_rx_Posted_queue_buffers >= 1;
  keep soft max_rx_Posted_queue_buffers == 5;
  max_rx_NonPosted_queue_buffers : uint;
  keep max_rx_NonPosted_queue_buffers >= 1;
  keep soft max_rx_NonPosted_queue_buffers == 5;
  max_rx_Response_queue_buffers : uint;
  keep max_rx_Response_queue_buffers >= 1;
  keep soft max_rx_Response_queue_buffers == 5;

  max_rx_Posted_Data_queue_buffers : uint;
  keep max_rx_Posted_Data_queue_buffers >= 1;
  keep soft max_rx_Posted_Data_queue_buffers == 5;
  max_rx_NonPosted_Data_queue_buffers : uint;
  keep max_rx_NonPosted_Data_queue_buffers >= 1;
  keep soft max_rx_NonPosted_Data_queue_buffers == 5;
  max_rx_Response_Data_queue_buffers : uint;
  keep max_rx_Response_Data_queue_buffers >= 1;
  keep soft max_rx_Response_Data_queue_buffers == 5;

  // A buffer can be in one of three states.  It is in use, it is available or it needs to be released.
  available_rx_Posted_queue_buffers : uint;
  keep available_rx_Posted_queue_buffers == 0;
  available_rx_NonPosted_queue_buffers : uint;
  keep available_rx_NonPosted_queue_buffers == 0;
  available_rx_Response_queue_buffers : uint;
  keep available_rx_Response_queue_buffers == 0;

  available_rx_Posted_Data_queue_buffers : uint;
  keep available_rx_Posted_Data_queue_buffers == 0;
  available_rx_NonPosted_Data_queue_buffers : uint;
  keep available_rx_NonPosted_Data_queue_buffers == 0;
  available_rx_Response_Data_queue_buffers : uint;
  keep available_rx_Response_Data_queue_buffers == 0;

  !available_rx_Posted_queue_buffers_bottom: bool;
  !available_rx_NonPosted_queue_buffers_bottom: bool;
  !available_rx_Response_queue_buffers_bottom: bool;
  !available_rx_Posted_Data_queue_buffers_bottom: bool;
  !available_rx_NonPosted_Data_queue_buffers_bottom: bool;
  !available_rx_Response_Data_queue_buffers_bottom: bool;

  ////////////////////////////////////////////////////////////////////////////////
  // Link RX Data Structures
  ////////////////////////////////////////////////////////////////////////////////
  // These are queues that are used to retire packets, so that the upper levels can handle them.

  // A single queue is used to retire packets so that ordering is maintained.  This allows for a simple FIFO algorithm to enforce
  // strong ordering requirements.
  !rx_retirement_queue : list of csco_ht_packet;
  // These are the two queues that are presented to the upper level, a list of packets to be dealt with at this node, and a list
  // to be forwarded to the next node.
  !rx_acceptance_queue : list of csco_ht_packet;
  !rx_forwarding_queue : list of csco_ht_packet;

  // A transmitter is required to preallocate buffer space for every response generated by its requests at some point of the
  // transmit hierarchy, but the spec does not intend this to be done with the response buffers.  However, a queue of expected
  // data will be kept at this level.
  !rx_pending_Response_queue : list (key:response_token) of csco_ht_rx_response_buffer;
  !rx_pending_Response_Data_queue : list (key:response_token) of csco_ht_rx_response_buffer;
  // A special queue to handle outstanding broadcasts.  Due to the nature of broadcasts, they will never be deleted
  // from this queue, but it will easily allow all device to handle any number of broadcast responses.
  !rx_outstanding_Broadcast_queue : list (key:response_token) of csco_ht_rx_response_buffer;
  // There can be only one outstanding data packet at a time between the virtual channels.
  !expected_rx_data : csco_ht_rx_pending_data_info;
  ////////////////////////////////////////////////////////////////////////////////
  // Link Debugging / History
  ////////////////////////////////////////////////////////////////////////////////
  // If this flag is set a message will be printed whenever a non-Nop packet is transmitted or received.
  display_channel_messages : bool;
  keep soft display_channel_messages == TRUE;
  // If this flag is set, then a message will be printed whenever a packet other than an empty Nop is transmitter or received.
  debug_channel : bool;
  keep soft debug_channel == FALSE;

  // These are lists that are used to help debug or track the behavior of the link.
  // They are enabled by the record_traffic field in the csco_ht_device_control (control_block) unit.
  // Add queues to track tx/rx histories if it is enabled.
  !tx_transaction_history : list of csco_ht_packet_record;
  !rx_transaction_history : list of csco_ht_packet_record;
  // Track the history of the selection process for the TX link.
  !tx_candidate_history : list of csco_ht_tx_candidate_list;
  ////////////////////////////////////////////////////////////////////////////////
  // Link TX Error Insertion
  ////////////////////////////////////////////////////////////////////////////////
  // A global channel flag to turn on or off error injection on this TX link
  allow_tx_errors : bool;
  keep soft allow_tx_errors == FALSE;
//  keep allow_tx_errors == TRUE;
  // Set a percentage that will determine if a ctl packet will have a illegal ctl line
  allow_illegal_ctl_packet : uint;
  keep allow_illegal_ctl_packet in range [0..100];
  ////////////////////////////////////////////////////////////////////////////////
  // Functions / Methods
  ////////////////////////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////
  // gen_nop_pkt() method to create a nop pkt...
  ////////////////////////////////////////////////////////////////////////////////
  gen_nop_pkt():Nop'command csco_ht_packet is inline {
    result = new;

    result.PostData = 0;
    result.NonPostCmd = 0;
    result.NonPostData = 0;
    result.Response = 0;
    result.ResponseData = 0;
    result.PostCmd = 0;

    result.vChan = NA;
    result.passing_rule = NA;
    result.packet_length = 4;
    result.Cmd = 0;
    result.DisCon = 0;
    result.Diag = 0;
    result.Zero = 0;
    result.Rsv = 0;

    result.post_generate();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // gen_sync_packet() method to create a nop pkt...
  ////////////////////////////////////////////////////////////////////////////////
  gen_sync_packet():csco_ht_packet is inline {
    gen result keeping {
      .command == Sync;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // insert_nop() will create a nop for transmission.  Based on certain flags it will release a variable number of
  // buffers and potentiatlly even allow starvation to occur.
  ////////////////////////////////////////////////////////////////////////////////
  insert_nop(current_tx_candidates : list of csco_ht_packet) is inline {
    // There are two modes that Nop buffer releases can operate in, either they are sent as a function of bandwidth
    // i.e. 10% of the BW should be Nop's or the target is allowed to starve, and once starvation occurs there is a
    // cumulative percentage chance that buffers will be released on subsequent cycles.
    var nop_with_releases : Nop'command csco_ht_packet;
    var nop_without_releases : Nop'command csco_ht_packet;
    var nop_weight : uint;

    var Posted_queue_buffers_to_be_released : uint;
    var NonPosted_queue_buffers_to_be_released : uint;
    var Response_queue_buffers_to_be_released : uint;

    var Posted_Data_queue_buffers_to_be_released : uint;
    var NonPosted_Data_queue_buffers_to_be_released : uint;
    var Response_Data_queue_buffers_to_be_released : uint;

    Posted_queue_buffers_to_be_released = max_rx_Posted_queue_buffers - available_rx_Posted_queue_buffers;
    NonPosted_queue_buffers_to_be_released = (max_rx_NonPosted_queue_buffers -
                                              available_rx_NonPosted_queue_buffers);
    Response_queue_buffers_to_be_released = max_rx_Response_queue_buffers - available_rx_Response_queue_buffers;

    Posted_Data_queue_buffers_to_be_released = (max_rx_Posted_Data_queue_buffers -
                                                available_rx_Posted_Data_queue_buffers);
    NonPosted_Data_queue_buffers_to_be_released = (max_rx_NonPosted_Data_queue_buffers -
                                                   available_rx_NonPosted_Data_queue_buffers);
    Response_Data_queue_buffers_to_be_released = (max_rx_Response_Data_queue_buffers -
                                                  available_rx_Response_Data_queue_buffers);

    // Determine how many buffers should be released.
    if(nop_maximum_buffer_release) {
      nop_with_releases = new;
      nop_with_releases.vChan = NA;
      nop_with_releases.passing_rule = NA;
      nop_with_releases.packet_length = 4;
      nop_with_releases.Cmd = 0;
      nop_with_releases.DisCon = 0;
      nop_with_releases.Diag = 0;
      nop_with_releases.Zero = 0;
      nop_with_releases.Rsv = 0;

      nop_with_releases.PostCmd = min(3, Posted_queue_buffers_to_be_released);
      nop_with_releases.PostData = min(3, Posted_Data_queue_buffers_to_be_released);
      nop_with_releases.NonPostCmd = min(3, NonPosted_queue_buffers_to_be_released);
      nop_with_releases.NonPostData = min(3, NonPosted_Data_queue_buffers_to_be_released);
      nop_with_releases.Response = min(3, Response_queue_buffers_to_be_released);
      nop_with_releases.ResponseData = min(3, Response_Data_queue_buffers_to_be_released);
      nop_with_releases.post_generate();
    } else {
      gen nop_with_releases keeping {
        it.PostCmd <= Posted_queue_buffers_to_be_released;
        it.PostData <= Posted_Data_queue_buffers_to_be_released;
        it.NonPostCmd <= NonPosted_queue_buffers_to_be_released;
        it.NonPostData <= NonPosted_Data_queue_buffers_to_be_released;
        it.Response <= Response_queue_buffers_to_be_released;
        it.ResponseData <= Response_Data_queue_buffers_to_be_released;
      };
    };

    // Don't allow the starvation flag to effect the initialization sequence of buffers.
    if(create_target_starvation && csco_ht_link.tx_link_initialized) {
      // If the target has been starved for the appropriate number of cycles, release some buffers.
      if(starvation_cycle_count > 0) {
        // Keep the weight less than or equal to 100%
        nop_weight = min(100, starvation_cycle_count * starvation_buffer_release_percentage);
        nop_with_releases.scheduling_index = 0;
        nop_with_releases.weight = nop_weight;
        current_tx_candidates.add(nop_with_releases);
        starvation_cycle_count += 1;
      }else {
        // Should target starvation be dependent on only the command buffers or the data buffers as well.
        if(available_rx_Posted_queue_buffers == 0 && available_rx_NonPosted_queue_buffers == 0 && available_rx_Response_queue_buffers == 0) {
          starvation_cycle_count = 1;
        };
        if(current_tx_candidates.is_empty()) {
          nop_without_releases = gen_nop_pkt();
          nop_without_releases.scheduling_index = 0;
          nop_without_releases.weight = nop_bandwidth_percent;
          current_tx_candidates.add(nop_without_releases);
        };
      };
    }else {
      // Either the flag for starvation is not set, or
      // The link is not fully initialized so Nop's need to be sent to perform the initial buffer releases.
      nop_with_releases.scheduling_index = 0;
      nop_with_releases.weight = nop_bandwidth_percent;
      current_tx_candidates.add(nop_with_releases);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // select_candidate() will take a list of candidate packets and evaluate which one should be transmitted at this time.
  // It uses predefined weights to aid in its selection and will then randomly select a packet.
  ////////////////////////////////////////////////////////////////////////////////
  // Once the candidate list has been generated, select one to be transmitted.
  select_candidate(current_tx_candidates : list of csco_ht_packet) is inline {
    var nop_candidate_weight : uint;
    var nop_candidate_index : uint;
    var nop_tx_selection : uint;
    var select_nop : bool;

    var total_candidate_weight : uint;
    var next_candidate_selected_range : uint;
    var next_candidate_index : uint;

    if(current_tx_candidates.is_empty()) {
      // If there are no candidates to be transmitted at this point, there is something wrong.
      dut_error("The candidate list is empty, there was a problem generating the list.");
    };

    // Once the list of candidate packets has been generated, determine which one to transmit.
    // Compute range.
    total_candidate_weight = 0;
    for each in current_tx_candidates {
      if(it.packet_type == Info) {
        if(nop_candidate_weight != 0) {
          dut_error("There are more than 1 Nop packets current scheduled to be transmitted.");
        };
        nop_candidate_weight = it.weight;
        nop_candidate_index = index;
      }else {
        total_candidate_weight += it.weight;
      };
    };

    // Determine if a Nop or a Ctl or Data packet will be sent.
    gen nop_tx_selection keeping {
      it in range [0..99];
      // If there are no other packets guarantee that the nop is selected.
      total_candidate_weight == 0 => it < nop_candidate_weight;
    };
    // Check this to see if we are getting the right distribution.
    if(nop_tx_selection < nop_candidate_weight) {
      select_nop = TRUE;
    }else {
      select_nop = FALSE;
    };

    if(select_nop) {
      current_tx_packet = current_tx_candidates[nop_candidate_index];
      // The Nop only exists in the candidate list, so it does not need to be removed from any other queue.
      // Update the buffer counters that are being released by the Nop.

      if current_tx_packet.PostCmd() != 0 and available_rx_Posted_queue_buffers_bottom == TRUE {
        available_rx_Posted_queue_buffers_bottom = FALSE;
        do_cover_buffer (rx_cover_instance, Posted, Ctl,
              available_rx_Posted_queue_buffers);
      };
      available_rx_Posted_queue_buffers += current_tx_packet.PostCmd();

      if current_tx_packet.PostData() != 0 and available_rx_Posted_Data_queue_buffers_bottom == TRUE {
        available_rx_Posted_Data_queue_buffers_bottom = FALSE;
        do_cover_buffer (rx_cover_instance, Posted, Data,
              available_rx_Posted_Data_queue_buffers);
      };
      available_rx_Posted_Data_queue_buffers += current_tx_packet.PostData();

      if current_tx_packet.NonPostCmd() != 0 and available_rx_NonPosted_queue_buffers_bottom == TRUE {
        available_rx_NonPosted_queue_buffers_bottom = FALSE;
        do_cover_buffer (rx_cover_instance, NonPosted, Ctl,
              available_rx_NonPosted_queue_buffers);
      };
      available_rx_NonPosted_queue_buffers += current_tx_packet.NonPostCmd();

      if current_tx_packet.NonPostData() != 0 and available_rx_NonPosted_Data_queue_buffers_bottom == TRUE {
        available_rx_NonPosted_Data_queue_buffers_bottom = FALSE;
        do_cover_buffer (rx_cover_instance, NonPosted, Data,
              available_rx_NonPosted_Data_queue_buffers);
      };
      available_rx_NonPosted_Data_queue_buffers += current_tx_packet.NonPostData();

      if current_tx_packet.Response() != 0 and available_rx_Response_queue_buffers_bottom == TRUE {
        available_rx_Response_queue_buffers_bottom = FALSE;
        do_cover_buffer (rx_cover_instance, Response, Ctl,
              available_rx_Response_queue_buffers);
      };
      available_rx_Response_queue_buffers += current_tx_packet.Response();

      if current_tx_packet.ResponseData() != 0 and available_rx_Response_Data_queue_buffers_bottom == TRUE {
        available_rx_Response_Data_queue_buffers_bottom = FALSE;
        do_cover_buffer (rx_cover_instance, Response, Data,
              available_rx_Response_Data_queue_buffers);
      };
      available_rx_Response_Data_queue_buffers += current_tx_packet.ResponseData();

      check that (available_rx_Posted_queue_buffers <= max_rx_Posted_queue_buffers) else
      dut_error("avail rx_Posted_queue_buffers (%d) is greater than max_rx_Posted_queue_buffers (%d)\n",
                available_rx_Posted_queue_buffers, max_rx_Posted_queue_buffers);

      check that (available_rx_Posted_Data_queue_buffers <= max_rx_Posted_Data_queue_buffers) else
      dut_error("avail rx_Posted_data_queue_buffers (%d) is greater than max_rx_Posted_data_queue_buffers (%d)\n",
                available_rx_Posted_Data_queue_buffers, max_rx_Posted_Data_queue_buffers);

      check that (available_rx_NonPosted_queue_buffers <= max_rx_NonPosted_queue_buffers) else
      dut_error("avail rx_NonPosted_queue_buffers (%d) is greater than max_rx_NonPosted_queue_buffers (%d)\n",
                available_rx_NonPosted_queue_buffers, max_rx_NonPosted_queue_buffers);

      check that (available_rx_NonPosted_Data_queue_buffers <= max_rx_NonPosted_Data_queue_buffers) else
      dut_error("avail rx_NonPosted_Data_queue_buffers (%d) is greater than max_rx_NonPosted_queue_buffers (%d)\n",
                available_rx_NonPosted_Data_queue_buffers, max_rx_NonPosted_Data_queue_buffers);

      check that (available_rx_Response_queue_buffers <= max_rx_Response_queue_buffers) else
      dut_error("avail rx_Response_queue_buffers (%d) is greater than max_rxReponse_queue_buffers (%d)\n",
                available_rx_Response_queue_buffers, max_rx_Response_queue_buffers);

      check that (available_rx_Response_Data_queue_buffers <= max_rx_Response_Data_queue_buffers) else
      dut_error("avail rx_Response_Data_queue_buffers (%d) is greater than max_rx_Repsonse_Data_queue_buffers (%d)\n",
                available_rx_Response_Data_queue_buffers, max_rx_Response_Data_queue_buffers);
    }else {
      // Randomly pick an number in that range.
      gen next_candidate_selected_range keeping {
        it <= total_candidate_weight;
        it > 0;
      };
      // Now determine which bucket they fit in.
      total_candidate_weight = 0;
      for each in current_tx_candidates {
        total_candidate_weight += it.weight;
        if(next_candidate_selected_range <= total_candidate_weight) {
          if(it.packet_type == Data) {
            current_tx_packet = pending_tx_data_packet;
            pending_tx_data_packet = NULL;
          }else {
            // This is a Ctl packet.
            next_candidate_index = it.scheduling_index;
            // Once the index has been determined pull that element from the list of candidates and the tx_scheduling_queue
            // and transmit it.
            current_tx_packet = tx_scheduling_queue[next_candidate_index];
            // Release the buffer from the scheduling queue and the count of available buffers.
            tx_scheduling_queue.delete(next_candidate_index);
          };
          // The packet to be transmitted has been selected.
          break;
        };
      };

      if(current_tx_packet.packet_type == Ctl) {
        if(current_tx_packet.command == Write || current_tx_packet.command == RdResponse) {
          // Prepare the data packet to be transmitted.
          if(pending_tx_data_packet != NULL) {
            dut_error("A Write or RdResponse is attempting to be tranmitted, when there is already 1 outstanding data packet.");
          }else {
            pending_tx_data_packet = current_tx_packet.retrieve_data_packet();
            size_of_last_tx_data_packet = pending_tx_data_packet.packet_length;
          };
        };

        case(current_tx_packet.vChan) {
          Posted: {
            if(available_target_Posted_queue_buffers == 0) {
              dut_error("Attempting to transmit a Posted Ctl packet with no available target buffers.");
            }else {
              available_target_Posted_queue_buffers -= 1;
              available_target_Posted_queue_buffers_bottom = TRUE;
            };
          };
          NonPosted: {
            if(current_tx_packet.command == Write || current_tx_packet.command == Flush) {
              // Place a response buffer struct on a rx scoreboard so that the receiver can rebuild the correct packets.
              var new_rx_response_buffer : csco_ht_rx_response_buffer = new;
              new_rx_response_buffer.unitID = current_tx_packet.UnitID();
              new_rx_response_buffer.SrcTag = current_tx_packet.SrcTag();
              new_rx_response_buffer.command = TgtDone;
              new_rx_response_buffer.endian_type = current_tx_packet.endian_type;
              new_rx_response_buffer.post_generate();
              rx_pending_Response_queue.add(new_rx_response_buffer);
            }else if(current_tx_packet.command == Read) {
              // Place a response buffer struct on a rx scoreboard so that the receiver can rebuild the correct packets.
              var new_rx_response_buffer : csco_ht_rx_response_buffer = new;
              new_rx_response_buffer.unitID = current_tx_packet.UnitID();
              new_rx_response_buffer.SrcTag = current_tx_packet.SrcTag();
              new_rx_response_buffer.command = RdResponse;
              new_rx_response_buffer.endian_type = current_tx_packet.endian_type;
              new_rx_response_buffer.post_generate();
              rx_pending_Response_queue.add(new_rx_response_buffer);

              new_rx_response_buffer = new;
              new_rx_response_buffer.unitID = current_tx_packet.UnitID();
              new_rx_response_buffer.SrcTag = current_tx_packet.SrcTag();
              new_rx_response_buffer.command = current_tx_packet.DataType();
              new_rx_response_buffer.endian_type = current_tx_packet.endian_type;
              new_rx_response_buffer.post_generate();
              rx_pending_Response_Data_queue.add(new_rx_response_buffer);
            }else if(current_tx_packet.command == Broadcast) {
              // Place a response buffer struct on a broadcast scoreboard so that the receiver can rebuild the correct packets.
              var new_rx_response_buffer : csco_ht_rx_response_buffer = new;
              new_rx_response_buffer.unitID = current_tx_packet.UnitID();
              new_rx_response_buffer.SrcTag = current_tx_packet.SrcTag();
              new_rx_response_buffer.command = TgtDone;
              new_rx_response_buffer.endian_type = current_tx_packet.endian_type;
              new_rx_response_buffer.post_generate();
              rx_outstanding_Broadcast_queue.add(new_rx_response_buffer);
            };
            if(available_target_NonPosted_queue_buffers == 0) {
              dut_error("Attempting to transmit a NonPosted Ctl packet with no available target buffers.");
            }else {
              available_target_NonPosted_queue_buffers -= 1;
              available_target_NonPosted_queue_buffers_bottom = TRUE;
            };
          };
          Response: {
            if(available_target_Response_queue_buffers == 0) {
              dut_error("Attempting to transmit a Response Ctl packet with no available target buffers.");
            }else {
              available_target_Response_queue_buffers -= 1;
              available_target_Response_queue_buffers_bottom = TRUE;
            };
          };
        };
      }else if(current_tx_packet.packet_type == Data) {
        case(current_tx_packet.vChan) {
          Posted: {
            if(available_target_Posted_Data_queue_buffers == 0) {
              dut_error("Attempting to transmit a Posted Data packet with no available target buffers.");
            }else {
              available_target_Posted_Data_queue_buffers -= 1;
              available_target_Posted_Data_queue_buffers_bottom = TRUE;
            };
          };
          NonPosted: {
            if(available_target_NonPosted_Data_queue_buffers == 0) {
              dut_error("Attempting to transmit a NonPosted Data packet with no available target buffers.");
            }else {
              available_target_NonPosted_Data_queue_buffers -= 1;
              available_target_NonPosted_Data_queue_buffers_bottom = TRUE;
            };
          };
          Response: {
            if(available_target_Response_Data_queue_buffers == 0) {
              dut_error("Attempting to transmit a Response Data packet with no available target buffers.");
            }else {
              available_target_Response_Data_queue_buffers -= 1;
              available_target_Response_Data_queue_buffers_bottom = TRUE;
            };
          };
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // generate_tx_buffer() will take the current_tx_packet and transform it into the current_tx_buffer that is
  // scheduled to be transmitted.  It also records the packet and prints out information if the appropriate flags are set.
  ////////////////////////////////////////////////////////////////////////////////
  generate_tx_buffer() is inline {
    current_tx_buffer_type = current_tx_packet.packet_type;

    current_tx_buffer = current_tx_packet.packed_data;

     // emit tx coverage event
    emit current_tx_packet.tx_pkt_cov_event;
    if(current_tx_packet.command == Read) {
      emit current_tx_packet.rd_tx_pkt_cov_event;
    } else if (current_tx_packet.command == Write) {
      emit current_tx_packet.wr_tx_pkt_cov_event;
    } else if (current_tx_packet.command == Broadcast) {
      emit current_tx_packet.bc_tx_pkt_cov_event;
    };

    // emit tx coverage events
    if current_tx_packet.packet_type != Data and pending_tx_data_packet != NULL {
      do_cover_stall (tx_cover_instance, 0, FALSE, FALSE);
    };

    if (current_tx_packet.command == Write or current_tx_packet.command == RdResponse)
          and pending_tx_data_packet != NULL {
      started_sending_last_command_with_data = TRUE;
      sent_last_command_with_data.data_length = size_of_last_tx_data_packet;
      sent_last_command_with_data.command = current_tx_packet.command;
      if current_tx_packet.Write'command'interrupt == TRUE {
        do_cover_interrupt_cmd (tx_cover_instance);
      };
    };


    if (tx_cover_instance != uncovered) {
      current_tx_packet.cover_instance = tx_cover_instance;
      current_tx_packet.do_cover();

      if (current_tx_packet.packet_type == Ctl) and
          ((current_tx_packet.command == Write and current_tx_packet.Write'command'interrupt == FALSE) or
            current_tx_packet.command == Read) {
        first_rw.command  = second_rw.command;
        first_rw.addr = second_rw.addr;
        first_rw.accessing_next_cache_line = second_rw.accessing_next_cache_line;

        second_rw.command  = third_rw.command;
        second_rw.addr = third_rw.addr;
        second_rw.accessing_next_cache_line = third_rw.accessing_next_cache_line;

        third_rw.command = current_tx_packet.command;
        if current_tx_packet is a Write csco_ht_packet (p) {
          third_rw.addr = p.Addr[39:5];
          if p.data_type == Byte {
            third_rw.accessing_next_cache_line = FALSE;
          } else if p.Addr[4:0]+p.Mask_Count > 31 {
            third_rw.accessing_next_cache_line = TRUE;
          };
        };
        if current_tx_packet is a Read csco_ht_packet (p) {
          third_rw.addr = p.Addr[39:5];
          if p.data_type == Byte {
            third_rw.accessing_next_cache_line = FALSE;
          } else if p.Addr[4:0]+p.Mask_Count > 31 {
            third_rw.accessing_next_cache_line = TRUE;
          };
        };
        do_cover_rw_sequence(tx_cover_instance);
      };
    };

    // Add to transmit history
    if(control_block.record_traffic) {
      // Only record Nops with no data if a special flag is set.
      if(!(current_tx_packet.command == Nop && current_tx_buffer.sum(it) == 0) || control_block.record_all_nop) {
        var new_packet_record : csco_ht_packet_record;
        gen new_packet_record keeping {
          .packet == current_tx_packet;
        };
        tx_transaction_history.add(new_packet_record);
      };
    };
    // For debug, don't print out the generation of Nop packets that have not data in them.
    // For normal display opperation, don't print out the Nop Packets.
    if((debug_channel && !(current_tx_packet.command == Nop && current_tx_buffer.sum(it) == 0))
       || (display_channel_messages && current_tx_packet.command != Nop)) {
      messagef(MEDIUM, "TX : %s", current_tx_packet.print_csco_ht_packet_string());
    };

    if(!(current_tx_packet.command == Nop && current_tx_buffer.sum(it) == 0)) {
      // If the current packet is not a stall, reset the activity timer.
      sys.csco_ht_test_control.reset_csco_ht_activity_timer();
    };

    emit csco_ht_queues_serviced;
    current_tx_packet = NULL;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // service_tx_ctl_queues() is used to select the eligible ctl packets that can be transmitted from the tx_scheduling_queue.
  // It allows for out of order packet transmission so it uses the HT ordering rules to determine which ctl packets are
  // currently legal to transmit this bit time.  It will build a list of candidates so that another method can select
  // which one to send.
  ////////////////////////////////////////////////////////////////////////////////
  service_tx_ctl_queues(current_tx_candidates : list of csco_ht_packet) is inline {
    var independent_IO_stream : bool;
    var IO_stream_index_match : list of int;
    var list_has_Fence : bool = FALSE;
    var part_of_ordered_sequence : bool = FALSE;

    // If TRUE then added the candidate to the current_tx_candidates list.
    var add_candidate : bool;
    // Pre-compute these for every candidate to determine the correct passing rules.
    var candidate_with_PassPW_can_pass_Posted_Requests : bool;
    var candidate_can_pass_Posted_Requests : bool;
    var candidate_can_pass_NonPosted_Requests : bool;
    var candidate_can_pass_Responses : bool;
    var available_target_buffer : bool;
    var available_target_data_buffer : bool;
    var candidate_is_eligible : bool;

    var candidate_vc_weight : uint;
    ////////////////////////////////////////////////////////////////////////////////
    // The tx_scheduling_queue is for all ctl packets other than Nop and Sync.
    if(!tx_scheduling_queue.is_empty()) {
      ////////////////////////////////////////////////////////////////////////////////
      for each (candidate_packet) using index (scheduling_index) in tx_scheduling_queue {
        // Start by assuming that this candidate can be added and continue to seach the current_tx_candidates list
        // until add_candidate is set to FALSE or every packet has been checked.
        add_candidate = TRUE;
        case(candidate_packet.vChan) {
          Posted: {
            candidate_with_PassPW_can_pass_Posted_Requests = Posted_Request_with_PassPW_pass_Posted_Request;
            candidate_can_pass_NonPosted_Requests = TRUE;
            candidate_can_pass_Responses = TRUE;
            candidate_vc_weight = Posted_VC_queue_weight;
            if(available_target_Posted_queue_buffers == 0) {
              available_target_buffer = FALSE;
            }else {
              available_target_buffer = TRUE;
            };
            if(candidate_packet.command == Write && available_target_Posted_Data_queue_buffers == 0) {
              available_target_data_buffer = FALSE;
            }else {
              available_target_data_buffer = TRUE;
            };
          };
          NonPosted: {
            candidate_with_PassPW_can_pass_Posted_Requests = NonPosted_Request_with_PassPW_pass_Posted_Request;
            candidate_can_pass_NonPosted_Requests = NonPosted_Request_pass_NonPosted_Request;
            candidate_can_pass_Responses = NonPosted_Request_pass_Response;
            candidate_vc_weight = NonPosted_VC_queue_weight;
            if(available_target_NonPosted_queue_buffers == 0) {
              available_target_buffer = FALSE;
            }else {
              available_target_buffer = TRUE;
            };
            if(candidate_packet.command == Write && available_target_NonPosted_Data_queue_buffers == 0) {
              available_target_data_buffer = FALSE;
            }else {
              available_target_data_buffer = TRUE;
            };
          };
          Response: {
            candidate_with_PassPW_can_pass_Posted_Requests = Posted_Request_with_PassPW_pass_Posted_Request;
            candidate_can_pass_NonPosted_Requests = TRUE;
            candidate_can_pass_Responses = Response_pass_Response;
            candidate_vc_weight = Response_VC_queue_weight;
            if(available_target_Response_queue_buffers == 0) {
              available_target_buffer = FALSE;
            }else {
              available_target_buffer = TRUE;
            };
            if(candidate_packet.command == RdResponse && available_target_Response_Data_queue_buffers == 0) {
              available_target_data_buffer = FALSE;
            }else {
              available_target_data_buffer = TRUE;
            };
          };
          default: {dut_error("Invalid candidate packet virtual channel");};
        };

        if(candidate_packet.PassPW() == FALSE
           || (candidate_packet.PassPW() == TRUE && !candidate_with_PassPW_can_pass_Posted_Requests)) {
             candidate_can_pass_Posted_Requests = FALSE;
           }else {
             candidate_can_pass_Posted_Requests = TRUE;
           };
        // Determine if the packet has available buffers and if there are associated data packets.
        if(available_target_buffer && available_target_data_buffer) {
          if((pending_tx_data_packet != NULL || !interrupted_tx_data_buffer.is_empty()) &&
             (candidate_packet.command == Write || candidate_packet.command == RdResponse)) {
               // If there is an outstanding data packet, then no control packet that has associated data can be transmitted.
               candidate_is_eligible = FALSE;
// BVE - Provide method hook here to allow back to back packets with data associated illegally.
             }else {
               candidate_is_eligible = TRUE;
             };
        }else {
          candidate_is_eligible = FALSE;
        };
        ////////////////////////////////////////////////////////////////////////////////
        if(candidate_is_eligible) {
          // First check to see if the candidate can pass other packets in the scheduling queue.
          // Start with the current entry in the scheduling queue, and compare it against all packets ahead of it.
          for j from scheduling_index-1 down to 0 do {
            var proposed_packet : csco_ht_packet;
            proposed_packet = tx_scheduling_queue[j];
            if(check_ordering_rules(proposed_packet, candidate_packet, candidate_can_pass_Posted_Requests, candidate_can_pass_NonPosted_Requests, candidate_can_pass_Responses) == FALSE) {
              add_candidate = FALSE;
              break;
            };
          };
          if(add_candidate == TRUE) {
            // Then check to see if the candidate can pass other packets in the candidate queue.
            // The first element is always eligible for transmisstion.
            for each (proposed_packet) using index (i) in current_tx_candidates {
              if(check_ordering_rules(proposed_packet, candidate_packet, candidate_can_pass_Posted_Requests, candidate_can_pass_NonPosted_Requests, candidate_can_pass_Responses) == FALSE) {
                add_candidate = FALSE;
                break;
              };
              // The current_tx_candidates list has been searched.
            };
          };
        }else {
          // There is no rx_buffer to receive this type of packet.
          add_candidate = FALSE;
        };
        ////////////////////////////////////////////////////////////////////////////////
        // Now that the candidate has been considered against all of the other packets that are currently eligible,
        // attempt to add it.
        if(add_candidate) {
          candidate_packet.scheduling_index = scheduling_index;
          // Each packet has a certain probabalistic weight that determines how often it is choosen.
          // The weight should equal: vc_weight + ordering_weight * (Max_Num_Queueable_Items - Index)
          // The vc_weight is a user definable knob and the ordering_weight (>0) will determine
          // how much the Index will affect the probability that items earlier in the queue will be serviced
          // sooner earlier in the tx_scheduling_queue.
          candidate_packet.weight = candidate_vc_weight + scheduling_timing_weight*(tx_scheduling_queue.size() - scheduling_index);
          candidate_packet.packet_type = Ctl;
          current_tx_candidates.add(candidate_packet);
        };
        // Every packet has been evaluated for transmission.
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // service_tx_queues() is used to check for any packets that need to be transmitted and select one of them.
  // It manages between outstanding ctl packets, data packets and the appropriate insertion of nop's.
  ////////////////////////////////////////////////////////////////////////////////
  service_tx_queues() is inline {
    var current_tx_candidates : list of csco_ht_packet;
    var candidate_vc_weight : uint;

    // Check the control queues for packets.
    service_tx_ctl_queues(current_tx_candidates);
    // If there is a pending data packet, put it on the candidate list.
    if(pending_tx_data_packet != NULL) {
      pending_tx_data_packet.scheduling_index = 0;
      case(pending_tx_data_packet.vChan) {
        Posted: {
          candidate_vc_weight = Posted_VC_queue_weight;
        };
        NonPosted: {
          candidate_vc_weight = NonPosted_VC_queue_weight;
        };
        Response: {
          candidate_vc_weight = Response_VC_queue_weight;
        };
        default: {dut_error("Invalid candidate packet virtual channel");};
      };
      pending_tx_data_packet.weight = candidate_vc_weight + data_scheduling_weight;
      current_tx_candidates.add(pending_tx_data_packet);
    };

    // If there is no meaningful data to be transmitted, then nop packets must be sent.
    // Add in a Nop packet to the candidate list.
    insert_nop(current_tx_candidates);

    // Add to candidate history
    if(control_block.record_traffic) {
      // Only record Nops with no data if a special flag is set.
      if(!(current_tx_candidates.size() == 1 && current_tx_candidates[0].command == Nop && current_tx_candidates[0].packed_data.sum(it) == 0)
         || control_block.record_all_nop) {
           var temp_candidate_list : csco_ht_tx_candidate_list;
           gen temp_candidate_list;
           temp_candidate_list.candidates = current_tx_candidates;
           tx_candidate_history.add(temp_candidate_list);
         };
    };

    select_candidate(current_tx_candidates);
    generate_tx_buffer();
    ////////////////////////////////////////////////////////////////////////////////
  };
  ////////////////////////////////////////////////////////////////////////////////
  // generate_stall() is a special method used to generate a nop that will release no buffers.  This packet will
  // be inserted into the middle of the transmission of a data packet and is known as a stall packet.
  ////////////////////////////////////////////////////////////////////////////////
  generate_stall() is inline {
    current_tx_packet = gen_nop_pkt();
    generate_tx_buffer();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // tx_packets() is a TCM that is the primary method used to transmit what ever is in the current_tx_buffer and to select
  // what is next to be transmitted.  It is also responsible for enabling packets to be interrupted during their transmission.
  ////////////////////////////////////////////////////////////////////////////////
  tx_packets()@csco_ht_link.csco_ht_tx_clk is {
    var chance_to_interrupt : uint;
    var tx_is_stall : bool;
    var current_tx_bit_time : csco_ht_bit_time = new;

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_device.reset @parent_device.reset;
      };
      {
        wait @csco_ht_link.csco_ht_tx_initialization_complete;
        ////////////////////////////////////////////////////////////////////////////////
        while(csco_ht_link.tx_link_initialized && control_block.tx_off(channel_identifier) == FALSE) {
          // Check to see if there is a packet to transmit, if not or if there is a interruption request a new packet.
          if(me.current_tx_buffer.is_empty()) {
            tx_doublewords_no += 1;
            if(control_block.Sync_Flood_Init == TRUE) {
              current_tx_packet = gen_sync_packet();
              messagef(MEDIUM, "sending sync packet %d", num_sync_packets_tx);
              if(num_sync_packets_tx == CSCO_HT_MIN_SYNC_SEQUENCE_CYCLES) {
                // The minimum number of packets have been transmitted, so reset the chain.
                messagef(MEDIUM, "16 sync packets transmitted");
                emit parent_device.Sync_Flood_Complete;
              }else {
                num_sync_packets_tx += 1;
              };
              generate_tx_buffer();
            }else {
              tx_is_stall = FALSE;
              service_tx_queues();
            };
          } else {
            if(csco_ht_link.csco_ht_tx_4_byte_boundary) {
              tx_doublewords_no += 1;
              // If a packet is in the process of being transmitted, check to see if it should be interrupted.
              gen chance_to_interrupt keeping {
                it in range [0..100];
              };
              if(chance_to_interrupt < interrupt_tx_packet_percentage) {
                if(current_tx_buffer_type == Ctl) {
                  check that (interrupted_tx_ctl_buffer.is_empty()) else dut_error("Another Ctl packet was already interrupted.");
                  interrupted_tx_ctl_buffer = current_tx_buffer;
                  interrupted_tx_ctl_buffer_type = current_tx_buffer_type;
                  current_tx_buffer = {};
                  tx_is_stall = TRUE;
                  emit intrp_tx_ctl_pkt;

                  do_cover_stall (tx_cover_instance, 4, TRUE,
                        interrupted_tx_data_buffer.is_empty() ? FALSE : TRUE);

                  generate_stall();
                } else if(current_tx_buffer_type == Data) {
                  check that (interrupted_tx_data_buffer.is_empty()) else dut_error("Another Data packet was already interrupted.");
                  interrupted_tx_data_buffer = current_tx_buffer;
                  interrupted_tx_data_buffer_type = current_tx_buffer_type;
                  current_tx_buffer = {};
                  tx_is_stall = FALSE;
                  emit intrp_tx_data_pkt;


                  // print tx_cover_instance;

                  do_cover_stall (tx_cover_instance,
                        size_of_last_tx_data_packet-current_tx_buffer.size()/8, FALSE, TRUE);

                  service_tx_queues();
                }else {
                  dut_error("An Info packet was going to be interrupted, which should be impossible.");
                };
              };
              // If there is no interrupt, just keep on transmitting the current packet.
            };
          };
          ////////////////////////////////////////////////////////////////////////////////
          // Transmit whatever is in the transmit buffer, a packet should be placed on the bus every cycle of the clock.
          if(!me.current_tx_buffer.is_empty()) {
            current_tx_bit_time.link_width = control_block.current_tx_width[channel_identifier];

            if((current_tx_buffer_type == Info ) && !tx_is_stall) {
              current_tx_bit_time.ctl = 1;
            } else if((current_tx_buffer_type == Info) && tx_is_stall) {
              current_tx_bit_time.ctl = 0;
            } else if(current_tx_buffer_type == Ctl) {
              var illegal_ctl_packet : uint;
              gen illegal_ctl_packet keeping {
                it in range [0..100];
              };
              // If a sync flood is under way, do not insert more errors
              if(control_block.Sync_Flood_Init == FALSE && allow_tx_errors == TRUE && (illegal_ctl_packet < allow_illegal_ctl_packet)) { // All CTL to be deasserted during a control packet
// BVE provide some flag or hook to indicate that an error was injected.
                messagef(MEDIUM, "TX : Illegal CTL line assertion created.");
                current_tx_bit_time.ctl = 0;
              }else {
                current_tx_bit_time.ctl = 1;
              };
            } else {
              current_tx_bit_time.ctl = 0;
            };

            current_tx_bit_time.clk = {};
            for i from 0 to (current_tx_bit_time.link_width.as_a(uint)/8)-1 do {
              current_tx_bit_time.clk.add(csco_ht_link.csco_ht_tx_clk_state);
            };
            for i from 0 to (current_tx_bit_time.link_width.as_a(int) - 1) {
              current_tx_bit_time.cad[i:i] = current_tx_buffer.pop0();
            };

            csco_ht_link.tx_bit_time = current_tx_bit_time;
            emit csco_ht_link.csco_ht_tx_buffer_ready;

            if(!interrupted_tx_data_buffer.is_empty() || !interrupted_tx_ctl_buffer.is_empty()) {
              // If a packet was interrupted and is being rescheduled, reset the activity timer.
              sys.csco_ht_test_control.reset_csco_ht_activity_timer();
            };

            // Before finishing transmitting the current buffer check to see if an interrupted packet needs to be resumed.
            if(me.current_tx_buffer.is_empty()) {
              if(current_tx_bit_time.ctl == 1) {
                // We just finished transmitting a Ctl packet.
                if(!interrupted_tx_data_buffer.is_empty()) {
                  current_tx_buffer = interrupted_tx_data_buffer;
                  current_tx_buffer_type = interrupted_tx_data_buffer_type;
                  interrupted_tx_data_buffer = {};
                };
                if started_sending_last_command_with_data == TRUE {
                  tx_doublewords_no = 0;
                  started_sending_last_command_with_data = FALSE;
                };
              }else {
                if current_tx_buffer_type == Data {
                  sent_last_command_with_data.data_time = tx_doublewords_no;
                  tx_doublewords_no = 0;
                  do_cover_delay (tx_cover_instance, sent_last_command_with_data);
                };
                if(!interrupted_tx_ctl_buffer.is_empty()) {
                  current_tx_buffer = interrupted_tx_ctl_buffer;
                  current_tx_buffer_type = interrupted_tx_ctl_buffer_type;
                  interrupted_tx_ctl_buffer = {};
                };
              };
            };
          } else {
            dut_error("The transmitter has an empty tx buffer, this should never happend.");
          };
          wait cycle;
          ////////////////////////////////////////////////////////////////////////////////
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // check_ordering_rules() is used to compare two packets to see if the candidate packet is able to pass the proposed packet.
  // All of the associate rules for passing are also provided, and a TRUE is returned if the candidate packet is allowed to pass.
  ////////////////////////////////////////////////////////////////////////////////
  check_ordering_rules(proposed_packet : csco_ht_packet, candidate_packet: csco_ht_packet, candidate_can_pass_Posted_Requests : bool,
                      candidate_can_pass_NonPosted_Requests : bool, candidate_can_pass_Responses : bool):bool is {
    // Check to see if the candidate_packet is in the same IO stream as any of the packets in the
    // current_tx_candidates list.
    // The IOStream is a special function that takes into account the bridge bit to determine if this
    // packet is going upstream or down.
    result = TRUE;
    if(proposed_packet.IOStream() == candidate_packet.IOStream()) {
      // Same IO stream
      // If the candidate packet is in the same IO stream and virtual channel,
      // check to see if it has a matching sequence number then it is a strongly ordered sequence.
      if(proposed_packet.vChan == candidate_packet.vChan
         && candidate_packet.vChan != Response // Response Packets do not have SeqID, so check and shortcircuit here.
         && proposed_packet.SeqID() == candidate_packet.SeqID()
         && candidate_packet.SeqID() != 0) {
           return FALSE;
         }else {
           // Otherwise refer to the ordering table in section 7.1 of the HT spec.
           if(proposed_packet.vChan == Posted && !candidate_can_pass_Posted_Requests) {
             return FALSE;
           };
           if(proposed_packet.vChan == NonPosted && !candidate_can_pass_NonPosted_Requests) {
             return FALSE;
           };
           if(proposed_packet.vChan == Response && !candidate_can_pass_Responses) {
             return FALSE;
           };
         };
    }else {
      // Independent IO streams
      if(proposed_packet.command == Fence && !candidate_can_pass_Posted_Requests) {
        return FALSE;
      };
      // Else, the candidate is eligble to pass and be reordered.
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // schedule_csco_ht_tx_packet() is the primary interface that a device has into the transmit path of a channel.
  // It takes a packet and will place it in the tx_scheduling_queue for future transmission.  If the depth of the
  // queue hits the watermark level, a message will be printed.
  ////////////////////////////////////////////////////////////////////////////////
  schedule_csco_ht_tx_packet(temp_csco_ht_packet : csco_ht_packet) is {
    // Check to see if the scheduling queue has passed it's watermark.
    if(tx_scheduling_queue.size() >= tx_scheduling_queue_watermark) {
      messagef(MEDIUM, "the tx_scheduling_queue has %d packets outstanding", tx_scheduling_queue.size());
    };
    tx_scheduling_queue.add(temp_csco_ht_packet);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // initiate_sync_flood() is a method that the reciever will call when an illegal condition occurs on the
  // bus.
  ////////////////////////////////////////////////////////////////////////////////
  initiate_sync_flood() is {
    // Eventually the initiation of a sync flood may want to error severity levels associated with it.
    if(control_block.Sync_Flood_Init == FALSE) { // Once a sync flood has been started, don't restart it.
      messagef(MEDIUM, "Initiating a Sync Flood");
      // Set the flag to start the sync flood
      control_block.Sync_Flood_Init = TRUE;
      // Immediately drop the packet in flight
      me.current_tx_buffer.clear();
      // Reset the TX clock counters
      csco_ht_link.reset_tx_clock_counters();
      emit parent_device.Sync_Flood_Init;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // decode_csco_ht_control_packet() is a simple method that will look at a cluster of 6 bits and determine what manner
  // of command field they would represent.  A ht packet type is then returned.
  ////////////////////////////////////////////////////////////////////////////////
  decode_csco_ht_control_packet(cmd_field : uint(bits:6)):csco_ht_packet is inline {
    result = new;

    result.stall = FALSE;
    if(cmd_field == 6'b000000) {
      result.command = Nop;
      result.packet_type = Info;
      result.vChan = NA;
      result.passing_rule = NA;
      result.packet_length = 4;
    } else if(cmd_field == 6'b110000) {
      result.command = RdResponse;
      result.packet_type = Ctl;
      result.vChan = Response;
      result.passing_rule = NoPassPW; // set in unpack
      result.packet_length = 4;
    } else if(cmd_field == 6'b110011) {
      result.command = TgtDone;
      result.packet_type = Ctl;
      result.vChan = Response;
      result.passing_rule = NoPassPW;
      result.packet_length = 4;
    } else if (cmd_field == 6'b000010) {
      result.command = Flush;
      result.packet_type = Ctl;
      result.vChan = NonPosted;
      result.passing_rule = NoPassPW; // set in unpack,  set to soft constraint
      result.packet_length = 4;
    } else if(cmd_field[4:3] == 2'b01) {
      // x01xxx
      result.command = Write;
      result.packet_type = Ctl;
      result.vChan = NonPosted; // set in unpack
      result.passing_rule = NoPassPW; // set in unpack
      result.packet_length = 8;
    } else if(cmd_field[5:4] == 2'b01) {
      // 01xxxx
      result.command = Read;
      result.packet_type = Ctl;
      result.vChan = NonPosted;
      result.passing_rule = NoPassPW; // set in unpack
      result.packet_length = 8;
    } else if(cmd_field[5:1] == 5'b11101) {
      // 11101x
      result.command = Broadcast;
      result.packet_type = Ctl;
      result.vChan = NonPosted; // set in unpack
      result.passing_rule = NoPassPW; // set in unpack
      result.packet_length = 8;
    } else if(cmd_field == 6'b111100) {
      result.command = Fence;
      result.packet_type = Ctl;
      result.vChan = Posted;
      result.passing_rule = NoPassPW; // set in unpack
      result.packet_length = 4;
    } else if(cmd_field == 6'b111111) {
      result.command = Sync;
      result.packet_type = Info;
      result.vChan = NA;
      result.passing_rule = NA;
      result.packet_length = 4;
    } else if((cmd_field == 6'b000001) || (cmd_field == 6'b000011) ||
              (cmd_field[5:2] == 4'b0001) || (cmd_field[5:3] == 3'b100) ||    // 0001xx and 100xxx
              (cmd_field == 6'b110001) || (cmd_field == 6'b110010) ||
              (cmd_field[5:1] == 5'b11010) || (cmd_field[5:1] == 5'b11011) || // 11010x and 11011x
              (cmd_field[5:1] == 5'b11100) || (cmd_field == 6'b111101) ||     // 11100x
              (cmd_field == 6'b111110)) {
      result.command = Reserved;
      result.packet_type = Info;
      result.vChan = NA;
      result.passing_rule = NA;
      result.packet_length = 4;
    } else {
      dut_error("Unable to decode a packet with a command field of ", cmd_field, ".");
    };
  };
////////////////////////////////////////////////////////////////////////////////
  // schedule_csco_ht_rx_packet() will take a newly assemble ht packet and place it into the retirement queue.
  // It will also check to make sure that this packet has appropriate buffer space.  Additionally it will associate
  // incomming data packets with their correct control packets and will throw errors if unexepected packets are received.
  // Unexepected packets are often unprompted response or data packets.  All of the control packets that are sent
  // are checked for side effects so that the receiver knows what to expect.
  ////////////////////////////////////////////////////////////////////////////////
  schedule_csco_ht_rx_packet(current_rx_packet : csco_ht_packet, current_rx_packet_type : csco_ht_packet_type, rx_is_stall : bool) is inline {
    var add_packet : bool = FALSE;
    var response_token : uint (bits:10);

    case(current_rx_packet_type) {
      Info: {
        case(current_rx_packet.command()) {
          Nop: {
            if(!rx_is_stall) {
              for i from 1 to current_rx_packet.PostCmd() do {
                // Check for overflow of the counter.
                if(available_target_Posted_queue_buffers + 1 > available_target_Posted_queue_buffers) {
                  if available_target_Posted_queue_buffers_bottom == TRUE {
                    available_target_Posted_queue_buffers_bottom = FALSE;
                    do_cover_buffer (tx_cover_instance, Posted, Ctl,
                          available_target_Posted_queue_buffers);
                  };
                  available_target_Posted_queue_buffers += 1;
                };
              };
              for i from 1 to current_rx_packet.PostData() do {
                // Check for overflow of the counter.
                if(available_target_Posted_Data_queue_buffers + 1 > available_target_Posted_Data_queue_buffers) {
                  if available_target_Posted_Data_queue_buffers_bottom == TRUE {
                    available_target_Posted_Data_queue_buffers_bottom = FALSE;
                    do_cover_buffer (tx_cover_instance, Posted, Data,
                          available_target_Posted_Data_queue_buffers);
                  };
                  available_target_Posted_Data_queue_buffers += 1;
                };
              };
              for i from 1 to current_rx_packet.NonPostCmd() do {
                // Check for overflow of the counter.
                if(available_target_NonPosted_queue_buffers + 1 > available_target_NonPosted_queue_buffers) {
                  if available_target_NonPosted_queue_buffers_bottom == TRUE {
                    available_target_NonPosted_queue_buffers_bottom = FALSE;
                    do_cover_buffer (tx_cover_instance, NonPosted, Ctl,
                          available_target_NonPosted_queue_buffers);
                  };
                  available_target_NonPosted_queue_buffers += 1;
                };
              };
              for i from 1 to current_rx_packet.NonPostData() do {
                // Check for overflow of the counter.
                if(available_target_NonPosted_Data_queue_buffers + 1 > available_target_NonPosted_Data_queue_buffers) {
                  if available_target_NonPosted_Data_queue_buffers_bottom == TRUE {
                    available_target_NonPosted_Data_queue_buffers_bottom = FALSE;
                    do_cover_buffer (tx_cover_instance, NonPosted, Data,
                          available_target_NonPosted_Data_queue_buffers);
                  };
                  available_target_NonPosted_Data_queue_buffers += 1;
                };
              };
              for i from 1 to current_rx_packet.Response() do {
                // Check for overflow of the counter.
                if(available_target_Response_queue_buffers + 1 > available_target_Response_queue_buffers) {
                  if available_target_Response_queue_buffers_bottom == TRUE {
                    available_target_Response_queue_buffers_bottom = FALSE;
                    do_cover_buffer (tx_cover_instance, Response, Ctl,
                          available_target_Response_queue_buffers);
                  };
                  available_target_Response_queue_buffers += 1;
                };
              };
              for i from 1 to current_rx_packet.ResponseData() do {
                // Check for overflow of the counter.
                if(available_target_Response_Data_queue_buffers + 1 > available_target_Response_Data_queue_buffers) {
                  if available_target_Response_Data_queue_buffers_bottom == TRUE {
                    available_target_Response_Data_queue_buffers_bottom = FALSE;
                    do_cover_buffer (tx_cover_instance, Response, Data,
                          available_target_Response_Data_queue_buffers);
                  };
                  available_target_Response_Data_queue_buffers += 1;
                };
              };
            };
          };
          Sync: {
            messagef(MEDIUM, "RX : A sync packet was received");
          };
          default: {
            dut_error("A Info packet that was not a Sync or a Nop was received.");
          };
        };
      };
      Ctl: {
        case(current_rx_packet.vChan) {
          Posted: {
            if(available_rx_Posted_queue_buffers == 0) {
              dut_error("We are receiveing a Posted packet even though there are no available buffers.");
            }else {
              rx_retirement_queue.add(current_rx_packet);
              available_rx_Posted_queue_buffers -= 1;
              available_rx_Posted_queue_buffers_bottom = TRUE;
// BVE - Problem with multiple outstanding Writes
              if(current_rx_packet.command == Write) {
                // Set a flag so that we know which queue to place the following data in.
                expected_rx_data = new;
                expected_rx_data.endian_type = parent_device.decode_endian_type(current_rx_packet.as_a(Write'command csco_ht_packet).Addr);
                expected_rx_data.associated_ctl_packet = current_rx_packet;
                expected_rx_data.data_type = current_rx_packet.DataType();
                expected_rx_data.data_length = current_rx_packet.DataLength();
                expected_rx_data.channel = current_rx_packet.vChan();
              }else {
                emit csco_ht_rx_packet_retired;
              };
            };
          };
          NonPosted: {
            if(available_rx_NonPosted_queue_buffers == 0) {
              dut_error("We are receiveing a NonPosted packet even though there are no available buffers.");
            }else {
              rx_retirement_queue.add(current_rx_packet);
              available_rx_NonPosted_queue_buffers -= 1;
              available_rx_NonPosted_queue_buffers_bottom = TRUE;
// BVE - Problem with multiple outstanding Writes
              if(current_rx_packet.command == Write) {
                // Set a flag so that we know which queue to place the following data in.
                expected_rx_data = new;
                expected_rx_data.endian_type = parent_device.decode_endian_type(current_rx_packet.as_a(Write'command csco_ht_packet).Addr);
                expected_rx_data.associated_ctl_packet = current_rx_packet;
                expected_rx_data.data_type = current_rx_packet.DataType();
                expected_rx_data.data_length = current_rx_packet.DataLength();
                expected_rx_data.channel = current_rx_packet.vChan();
              }else {
                emit csco_ht_rx_packet_retired;
              };
            };
          };
          Response: {
            if(available_rx_Response_queue_buffers == 0) {
              dut_error("We are receiveing a Response packet even though there are no available buffers.");
            }else {
              // First check to see if the packet is expected.
              if(current_rx_packet.BridgeBit() == 1) {
                // This is a response from a Host Bridge so the UnitID is the source of the request.
                // Hence the UnitID is used to compute the response token.
                response_token = %{current_rx_packet.UnitID(), current_rx_packet.SrcTag()};
              }else {
                // Otherwise this is a response from an IO Device and the UnitID is the target of the request.
                // Thus the UnitID of 0 should be used to compute the response token to align with the requestors UnitID.
                response_token = %{5'b00000, current_rx_packet.SrcTag()};
              };
              if(!rx_pending_Response_queue.key_exists(response_token) && !rx_outstanding_Broadcast_queue.key_exists(response_token)) {
                dut_error("A Response packet is being received even though none are scheduled to be received.");
              }else {
                // get the pending response pkt from the queue to set the endian type for it's
                // associated data
                var pending_rx_response: csco_ht_rx_response_buffer = rx_pending_Response_queue.key(response_token);

                rx_retirement_queue.add(current_rx_packet);
                // Buffers are not pre-allocated at the HT link level, so buffer managment must occur.
                available_rx_Response_queue_buffers -= 1;
                available_rx_Response_queue_buffers_bottom = TRUE;

                if(current_rx_packet.command == RdResponse) {
                  // Set a flag so that we know which queue to place the following data in.
                  // Extract the expected type of data packet from the info that the tx side creates.
                  expected_rx_data = new;
                  expected_rx_data.endian_type = pending_rx_response.endian_type;
                  expected_rx_data.associated_ctl_packet = current_rx_packet;
                  expected_rx_data.data_type = rx_pending_Response_Data_queue.key(response_token).command;
                  expected_rx_data.data_length = current_rx_packet.DataLength();
                  expected_rx_data.channel = current_rx_packet.vChan();
                }else {
                  emit csco_ht_rx_packet_retired;
                };
                // Set it up to screen for responses from Broadcasts and never remove those from the broadcast queue.
                if(rx_pending_Response_queue.key_exists(response_token)) {
                  rx_pending_Response_queue.delete(rx_pending_Response_queue.key_index(response_token));
                };
              };
            };
          };
          default: {
            dut_error("Failure to classify a received packet");
          };
        };
      };
      Data: {
        case(expected_rx_data.channel) {
          Posted: {
            if(available_rx_Posted_Data_queue_buffers == 0) {
              dut_error("We are receiveing a Posted Data packet even though there are no available buffers.");
            }else {
              expected_rx_data.associated_ctl_packet.attach_data_packet(current_rx_packet);
              emit csco_ht_rx_packet_retired;
              available_rx_Posted_Data_queue_buffers -= 1;
              available_rx_Posted_Data_queue_buffers_bottom = TRUE;
              expected_rx_data = NULL;
            };
          };
          NonPosted: {
            if(available_rx_NonPosted_Data_queue_buffers == 0) {
              dut_error("We are receiveing a NonPosted Data packet even though there are no available buffers.");
            }else {
              expected_rx_data.associated_ctl_packet.attach_data_packet(current_rx_packet);
              emit csco_ht_rx_packet_retired;
              available_rx_NonPosted_Data_queue_buffers -= 1;
              available_rx_NonPosted_Data_queue_buffers_bottom = TRUE;
              expected_rx_data = NULL;
            };
          };
          Response: {
            if(available_rx_Response_Data_queue_buffers == 0) {
              dut_error("We are receiveing a Response Data packet even though there are no available buffers.");
            }else {
              // First check to see if the packet is expected.
              if(expected_rx_data.associated_ctl_packet.BridgeBit() == 1) {
                // This is a response from a Host Bridge so the UnitID is the source of the request.
                // Hence the UnitID is used to compute the response token.
                response_token = %{expected_rx_data.associated_ctl_packet.UnitID(), expected_rx_data.associated_ctl_packet.SrcTag()};
              }else {
                // Otherwise this is a response from an IO Device and the UnitID is the target of the request.
                // Thus the UnitID of 0 should be used to compute the response token to align with the requestors UnitID.
                response_token = %{5'b00000, expected_rx_data.associated_ctl_packet.SrcTag()};
              };
              if(!rx_pending_Response_Data_queue.key_exists(response_token)) {
                dut_error("A Response Data packet is being received even though none are scheduled to be received.");
              }else {
                expected_rx_data.associated_ctl_packet.attach_data_packet(current_rx_packet);
                emit csco_ht_rx_packet_retired;
                available_rx_Response_Data_queue_buffers -= 1;
                available_rx_Response_Data_queue_buffers_bottom = TRUE;
                expected_rx_data = NULL;
                rx_pending_Response_Data_queue.delete(rx_pending_Response_Data_queue.key_index(response_token));
              };
            };
          };
          default: {
            dut_error("Attempting to retire a data packet that does not belong to a valid virtual channel");
          };
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // rx_packets() is the primary TCM for the receive path.  It will collect bit times from the link interface and
  // assemble them into packets.  Once that is done it will place them into the retirement queue so that the channel can
  // decide if they are to be accepted by this device or forwarded to the next one on the link.
  ////////////////////////////////////////////////////////////////////////////////
  rx_packets()@csco_ht_link.csco_ht_rx_clk is {
    var current_rx_bit_time : csco_ht_bit_time;
    var current_rx_control_encoding : csco_ht_packet_encoding;

    // A given receiever can only be receiving one packet of a single type.

    var rx_is_stall : bool;
    // The info packet is a special case of the control type.  It is point to point and can always be accepted.
    // A stall packet is an info packet with no valid data in it, so it must be treated specially.

    // Only one data packet can be outstanding at any single time.
    var interrupted_rx_data_packet : csco_ht_packet;
    var interrupted_rx_data_type : csco_ht_packet_type;
    // A control packet can only be interrupted by a stall packet.
    var interrupted_rx_ctl_packet : csco_ht_packet;
    var interrupted_rx_ctl_type : csco_ht_packet_type;

    // These fields are used to track the current packet that is being received
    var current_rx_packet_type : csco_ht_packet_type;
    var current_rx_packet : csco_ht_packet;
    var rx_packet_in_progress : bool;
    var rx_packet_completed : bool;

    // This is the initial value that will be seen on the bus and is used to frame the first ctl packet.
    var previous_rx_bit_time : csco_ht_bit_time = new;
    previous_rx_bit_time.link_width = control_block.read_current_rx_width(channel_identifier);
    previous_rx_bit_time.ctl = 0;
    for i from 0 to (previous_rx_bit_time.link_width.as_a(uint)/8)-1 do {
      previous_rx_bit_time.clk.add(0);
    };
    previous_rx_bit_time.cad = ((1 << previous_rx_bit_time.link_width.as_a(int)) - 1);
    var tmp_data : list of bit;

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_device.reset @parent_device.reset;
      };
      {
        wait @csco_ht_link.csco_ht_rx_initialization_complete;
        ////////////////////////////////////////////////////////////////////////////////
        while(csco_ht_link.rx_link_initialized && control_block.tx_off(channel_identifier) == FALSE) {
          // Since the ht bus is a source synchronous bus, the receiver knows that it needs to read data when it sees a clock edge
          sync @csco_ht_link.csco_ht_rx_buffer_ready;
          current_rx_bit_time = csco_ht_link.rx_bit_time;
          if(interrupted_rx_data_packet != NULL || interrupted_rx_ctl_packet != NULL) {
            // If a packet was interrupted and is being rescheduled, reset the activity timer.
            sys.csco_ht_test_control.reset_csco_ht_activity_timer();
          };
          ////////////////////////////////////////////////////////////////////////////////
          // Determine what type of packet is being recieved and if a packet is being interrupted.
          ////////////////////////////////////////////////////////////////////////////////
          if(csco_ht_link.csco_ht_rx_4_byte_boundary) {
            rx_doublewords_no += 1;
            // On a 4 byte boundary several things could happen: Receive a new ctl or data packet; Interrupt a ctl or data packet.
            if(current_rx_bit_time.ctl != previous_rx_bit_time.ctl) {
              // The value of the ctl line is changing.
              if(rx_packet_in_progress) {
                // A packet is being interrupted
                if(current_rx_bit_time.ctl == 1) {
                  // A data packet is being interrupted.
                  if(interrupted_rx_data_packet != NULL) {
                    dut_error("Error receiving an interrupt ctl packet.");
                  };
                  interrupted_rx_data_packet = current_rx_packet;
                  interrupted_rx_data_type = current_rx_packet_type;
                  rx_is_stall = FALSE;

                  do_cover_stall (rx_cover_instance,
                        current_rx_packet.packet_length - current_rx_packet.packed_data.size()/8,
                        FALSE, FALSE);

                }else {
                  // A ctl packet is being interrupted.
                  if(interrupted_rx_ctl_packet != NULL) {
                    dut_error("Error receiving an interrupt stall packet.");
                  };
                  interrupted_rx_ctl_packet = current_rx_packet;
                  interrupted_rx_ctl_type = current_rx_packet_type;
                  rx_is_stall = TRUE;

                  do_cover_stall (rx_cover_instance, 4, TRUE,
                        (expected_rx_data == NULL && interrupted_rx_data_packet == NULL) ? FALSE : TRUE);

                };
                // Allow the receiver to start a new packet.
                rx_packet_in_progress = FALSE;
              } else {
                rx_is_stall = FALSE;
                // Check to see is an interrupted packet is being resumed.
                if(current_rx_bit_time.ctl == 1) {
                  // Incomming ctl info
                  if(interrupted_rx_ctl_packet != NULL) {
                    // Resume rx of existing ctl packet.
                    current_rx_packet = interrupted_rx_ctl_packet;
                    current_rx_packet_type = interrupted_rx_ctl_type;
                    rx_packet_in_progress = TRUE;
                    interrupted_rx_ctl_packet = NULL;
                  };
                }else {
                  // Incomming data info
                  if(interrupted_rx_data_packet != NULL) {
                    // Resume rx of existing data packet.
                    current_rx_packet = interrupted_rx_data_packet;
                    current_rx_packet_type = interrupted_rx_data_type;
                    rx_packet_in_progress = TRUE;
                    interrupted_rx_data_packet = NULL;
                  };
                };
              };
            };
          }else {
            if(current_rx_bit_time.ctl != previous_rx_bit_time.ctl) {
              sys.csco_ht_errors.report_protocol_error(ILLEGAL_CTL_CHANGE,
                                                   "An illegal ctl line change occured.",
                                                   control_block, channel_identifier);
              // Eventually the initiation of a sync flood may want to error severity levels associated with it.
              // Generate a sync flood
              initiate_sync_flood();
            };
          };
          ////////////////////////////////////////////////////////////////////////////////
          // Receive the packet and collect the information
          ////////////////////////////////////////////////////////////////////////////////
          if(!rx_packet_in_progress) {
            // A packet is not currently being captured.
            if(current_rx_bit_time.ctl == 1) {
              // A control packet is being received.
              current_rx_packet = decode_csco_ht_control_packet(current_rx_bit_time.cad[5:0]);
              current_rx_control_encoding = current_rx_packet.command;

              case(current_rx_control_encoding) {
                Nop: {
                  // Info packet, use for point-to-point communication and is not propogated throught the HT chain.
                  current_rx_packet_type = Info;
                };
                Sync: {
                  // The link is being initialized, or an error has occurred.
                  current_rx_packet_type = Info;
                  // Generate a sync flood
                  initiate_sync_flood();
                };
                Reserved: {
                  // A packet with a reserved command field was received.  Throw an error.
                  sys.csco_ht_errors.report_protocol_error(RSV_COMMAND_FIELD,
                                                       "A packet with a reserved command field was received.",
                                                       control_block,
                                                       channel_identifier);
                };
                default: {
                  // A Flush, Write, Read, RdResponse, TgtDone, Broadcast or Fence was received.
                  current_rx_packet_type = Ctl;
                };
              };

              tmp_data = %{current_rx_bit_time.cad[current_rx_bit_time.link_width.as_a(uint)-1:0]};
              current_rx_packet.packed_data = tmp_data.copy();
              if(current_rx_packet.command == Write) {
                current_rx_packet.as_a(Write'command csco_ht_packet).assoc_data = NULL;
              } else if(current_rx_packet.command == RdResponse) {
                current_rx_packet.as_a(RdResponse'command csco_ht_packet).assoc_data = NULL;

              };

              rx_packet_in_progress = TRUE;

              if (expected_rx_data != NULL) and interrupted_rx_data_packet == NULL {
                // a new CTL packet is received right after a Write/RdResponse command, and before receiving
                // any associated data packets
                do_cover_stall (rx_cover_instance, 0, FALSE, FALSE);
              };

            }else {
              // A data packet or a stall packet is being received.
              if(rx_is_stall) {
                // rx a stall
                current_rx_packet = gen_nop_pkt();

                tmp_data = %{current_rx_bit_time.cad[current_rx_bit_time.link_width.as_a(uint)-1:0]};
                current_rx_packet.packed_data = tmp_data.copy();
                rx_packet_in_progress = TRUE;
                current_rx_packet_type = Info;
              }else {
                // rx data
                if(expected_rx_data != NULL) {
                  // The receiver is expecting data
                  current_rx_packet = new;
                  current_rx_packet.packet_type = Data;
                  current_rx_packet.endian_type = expected_rx_data.endian_type;
                  current_rx_packet.command = expected_rx_data.data_type;
                  current_rx_packet.packet_length = expected_rx_data.data_length;
                  current_rx_packet.vChan = expected_rx_data.channel;
                  current_rx_packet.passing_rule = NA;

                  tmp_data = %{current_rx_bit_time.cad[current_rx_bit_time.link_width.as_a(uint)-1:0]};
                  current_rx_packet.packed_data = tmp_data.copy();
                  rx_packet_in_progress = TRUE;
                  current_rx_packet_type = Data;
                }else {
                  sys.csco_ht_errors.report_protocol_error(ILLEGAL_DATA_PACKET,
                                                       "The receiver received data when none was expected.",
                                                       control_block, channel_identifier);
                  // Eventually the initiation of a sync flood may want to error severity levels associated with it.
                  // Generate a sync flood
                  initiate_sync_flood();

                  // Create a new packet buffer to receive the bad data so that the model does not go off into the weeds
                  current_rx_packet = new;
                  current_rx_packet.packet_type = Data;
                  current_rx_packet.passing_rule = NA;

                  tmp_data = %{current_rx_bit_time.cad[current_rx_bit_time.link_width.as_a(uint)-1:0]};
                  current_rx_packet.packed_data = tmp_data.copy();
                  rx_packet_in_progress = TRUE;
                  current_rx_packet_type = Data;
                };
              };
            };
            // Check to see if the entire packet has been received.
            if(current_rx_packet.packed_data.size()/8 == current_rx_packet.packet_length) {
              rx_packet_completed = TRUE;
            };
            ////////////////////////////////////////////////////////////////////////////////
          }else {
            // The link is currently in the middle of receiving a packet.
            if(current_rx_packet.packed_data.size()/8 < current_rx_packet.packet_length) {
              current_rx_packet.packed_data.add(%{current_rx_bit_time.cad[current_rx_bit_time.link_width.as_a(uint) - 1:0]});

              // Check to see if the entire packet has been received.
              if(current_rx_packet.packed_data.size()/8 == current_rx_packet.packet_length) {
                rx_packet_completed = TRUE;
              };
            }else {
              sys.csco_ht_errors.report_protocol_error(ILLEGAL_PACKET_SIZE, "Receiveing a packet that exceeds permited lengths.", control_block, channel_identifier);
            };
          };
          ////////////////////////////////////////////////////////////////////////////////
          // Retire the packet
          ////////////////////////////////////////////////////////////////////////////////
          if(rx_packet_completed) {
            current_rx_packet.unpack_packet();

            // emit coverage event
            emit current_rx_packet.rx_pkt_cov_event;
            if (rx_cover_instance != uncovered) {
              current_rx_packet.cover_instance = rx_cover_instance;
              current_rx_packet.do_cover();
            };

            // Add to receive history
            if(control_block.record_traffic) {
              // Only record Nops with no data if a special flag is set.
              if(!(current_rx_packet.command == Nop && current_rx_packet.packed_data.sum(it) == 0) || control_block.record_all_nop) {
                var new_packet_record : csco_ht_packet_record;
                gen new_packet_record keeping {
                  .packet == current_rx_packet;
                };
                rx_transaction_history.add(new_packet_record);
              };
            };

            // For debug, don't print out the generation of Nop packets that have not data in them.
            // For normal display opperation, don't print out the Nop Packets.
            if((debug_channel && !(current_rx_packet.command == Nop && current_rx_packet.packed_data.sum(it) == 0))
               || (display_channel_messages && current_rx_packet.command != Nop)) {
                 messagef(MEDIUM, "RX : %s", current_rx_packet.print_csco_ht_packet_string());
               };

            if(!(current_rx_packet.command == Nop && current_rx_packet.packed_data.sum(it) == 0)) {
              // If the current packet is not a stall, reset the activity timer.
              sys.csco_ht_test_control.reset_csco_ht_activity_timer();
            };
            // Add the packet to the appropriate receive queue.
            schedule_csco_ht_rx_packet(current_rx_packet, current_rx_packet_type, rx_is_stall);
            // Finish retiring the packet.

            if (current_rx_packet.command == Write or current_rx_packet.command == RdResponse)
                  and expected_rx_data != NULL {
              rx_doublewords_no = 0;
              received_last_command_with_data.command = current_rx_packet.command;
            };

            if current_rx_packet_type == Data {
              received_last_command_with_data.data_time = rx_doublewords_no;
              rx_doublewords_no = 0;
              received_last_command_with_data.data_length = current_rx_packet.packet_length;
              do_cover_delay (rx_cover_instance, received_last_command_with_data);
            };

            rx_packet_completed = FALSE;
            current_rx_packet = NULL;
            rx_packet_in_progress = FALSE;
          };
          ////////////////////////////////////////////////////////////////////////////////
          // Record the bit time for use with the next iteration.
          previous_rx_bit_time = current_rx_bit_time;
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // io_device_packet_routing() is used by IO devices to pull packets from the retirement queue and sort them into
  // packets that are accepted by this node or forwarded to the next node.  Note: Any packet that is not for this device
  // is placed in the forwarding queue and it is the responsability of the device to send back error responses if it is an
  // end of chain device.  This method provides the primary interface for the receive path for this channel to the device.
  ////////////////////////////////////////////////////////////////////////////////
  io_device_packet_routing()@csco_ht_link.csco_ht_rx_clk is {
    var retired_rx_packet : csco_ht_packet;

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_device.reset @parent_device.reset;
      };
      {
        wait true(control_block.physical_link_initialized(channel_identifier) == TRUE);
        ////////////////////////////////////////////////////////////////////////////////
        while(control_block.physical_link_initialized(channel_identifier)) {
          sync @csco_ht_rx_packet_retired;

          // This currently uses a very simple fifo scheme that will retire all packets in order, and do as many as available at a time.
          while(!rx_retirement_queue.is_empty()) {
            retired_rx_packet = rx_retirement_queue.pop0();

            case(retired_rx_packet.command) {
              Flush: {
                // A Flush packet is only allowed to flow in the upstream direction.  Thus the only receiver that
                // should see this is the Primary_Downstream_Interface.
                check that (control_block.read_link_orientation(channel_identifier) == TX_Downstream_RX_Upstream_Interface) else dut_error("Illegal Flush packet direction.");

                // A flush's whole point is to ensure that writes have been observed at host memory, so simply forward this to the TX path.
                rx_forwarding_queue.add(retired_rx_packet);
              };
              Write: {
                if(retired_rx_packet.retrieve_data_packet() == NULL) {
                  // The data for this Write is not ready yet.
                  rx_retirement_queue.push0(retired_rx_packet);
                  break;
                };

                if((retired_rx_packet.UnitID() == 0 && retired_rx_packet.CompatBit() == 0 && control_block.device_header.compare_address(retired_rx_packet.Address()))
                   || (retired_rx_packet.UnitID() == 0 && retired_rx_packet.CompatBit() == 1 && (control_block.southbridge || control_block.southbridge_path))) {
                     // This is a request from the host bridge for this device.
                     rx_acceptance_queue.add(retired_rx_packet);
                   }else {
                     rx_forwarding_queue.add(retired_rx_packet);
                   };
              };
              Read: {
                if((retired_rx_packet.UnitID() == 0 && retired_rx_packet.CompatBit() == 0 && control_block.device_header.compare_address(retired_rx_packet.Address()))
                   || (retired_rx_packet.UnitID() == 0 && retired_rx_packet.CompatBit() == 1 && (control_block.southbridge || control_block.southbridge_path))) {
                     // This is a request from the host bridge for this device.
                     rx_acceptance_queue.add(retired_rx_packet);
                   }else {
                     rx_forwarding_queue.add(retired_rx_packet);
                   };
              };
              RdResponse: {
                if(retired_rx_packet.retrieve_data_packet() == NULL) {
                  // The data for this RdResponse is not ready yet.
                  rx_retirement_queue.push0(retired_rx_packet);
                  break;
                };
                if(retired_rx_packet.BridgeBit() == 1 && control_block.compare_current_UnitID(retired_rx_packet.UnitID())) {
                  // This is a response from the host bridge and it is for this node.
                  rx_acceptance_queue.add(retired_rx_packet);
                }else {
                  rx_forwarding_queue.add(retired_rx_packet);
                };
              };
              TgtDone: {
                if(retired_rx_packet.BridgeBit() == 1 && control_block.compare_current_UnitID(retired_rx_packet.UnitID())) {
                  // This is a response from the host bridge and it is for this node.
                  rx_acceptance_queue.add(retired_rx_packet);
                }else {
                  rx_forwarding_queue.add(retired_rx_packet);
                };
              };
              Broadcast: {
                // A Broadcast packet is only supposed to flow downstream, from Host to IO devices.  Thus, the only receiver that
                // should see this is the Primary_Upstream_Interface.
                check that (control_block.read_link_orientation(channel_identifier) == TX_Upstream_RX_Downstream_Interface) else dut_error("Illegal Broadcast packet direction.");

                // A broadcast is to be accepted to be handled at this node, and sent on to the next.
                rx_acceptance_queue.add(retired_rx_packet);
                rx_forwarding_queue.add(retired_rx_packet);
              };
              Fence: {
                // A Fence packet is only allowed to flow in the upstream direction.  Thus the only receiver that
                // should see this is the Primary_Downstream_Interface.
                check that (control_block.read_link_orientation(channel_identifier) == TX_Downstream_RX_Upstream_Interface) else dut_error("Illegal Fence packet direction.");

                // A fence's whole point is to control ordering behavior at the transmitter, so simply forward this to the TX path.
                rx_forwarding_queue.add(retired_rx_packet);
              };
              default: {
                // Nop, Sync, Reserved, DoubleWord_Read, DoubleWord_Write, Byte_Read, Byte_Write
                dut_error("A packet was put into the retirement queue that does not belong there.");
              };
            };
          };
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // host_bridge_packet_routing() is used by Host Bridge devices to pull packets from the retirement queue and sort them into
  // packets that are accepted by this node or forwarded to the next node.  Note: Any packet that is not for this device
  // is placed in the forwarding queue and it is the responsability of the device to send back error responses if it is an
  // end of chain device.  This method provides the primary interface for the receive path for this channel to the device.
  ////////////////////////////////////////////////////////////////////////////////
  host_bridge_packet_routing()@csco_ht_link.csco_ht_rx_clk is {
    var retired_rx_packet : csco_ht_packet;

    // There are some conditions that need to be added to this to allow for double ended chains and host bridges to
    // exist at both ends.

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_device.reset @parent_device.reset;
      };
      {
        wait true(control_block.physical_link_initialized(channel_identifier) == TRUE);
        ////////////////////////////////////////////////////////////////////////////////
        while(control_block.physical_link_initialized(channel_identifier)) {
          sync @csco_ht_rx_packet_retired;

          // This currently uses a very simple fifo scheme that will retire all packets in order, and do as many as available at a time.
          while(!rx_retirement_queue.is_empty()) {
            retired_rx_packet = rx_retirement_queue.pop0();

            case(retired_rx_packet.command) {
              Flush: {
                // A flush is used to push writes into host memory.
                rx_acceptance_queue.add(retired_rx_packet);
              };
              Write: {
                if(retired_rx_packet.retrieve_data_packet() == NULL) {
                  // The data for this Write is not ready yet.
                  rx_retirement_queue.push0(retired_rx_packet);
                  break;
                };

                if((retired_rx_packet.UnitID() == 0 && control_block.device_header.compare_address(retired_rx_packet.Address()))
                   || retired_rx_packet.Address()[39:24] == 16'hFDF8 // This would indicate an interrupt request.
                   || retired_rx_packet.UnitID() != 0) {
                     // This is a request from another host bridge or a request from some downstream device.
                     rx_acceptance_queue.add(retired_rx_packet);
                   }else {
                     rx_forwarding_queue.add(retired_rx_packet);
                   };
              };
              Read: {
                if((retired_rx_packet.UnitID() == 0 && control_block.device_header.compare_address(retired_rx_packet.Address()))
                   || retired_rx_packet.UnitID() != 0) {
                     // This is a request from another host bridge or a request from some downstream device.
                     rx_acceptance_queue.add(retired_rx_packet);
                   }else {
                     rx_forwarding_queue.add(retired_rx_packet);
                   };
              };
              RdResponse: {
                if(retired_rx_packet.retrieve_data_packet() == NULL) {
                  // The data for this RdResponse is not ready yet.
                  rx_retirement_queue.push0(retired_rx_packet);
                  break;
                };

                if(retired_rx_packet.BridgeBit() == 0) {
                  // This is a response from the fabric for the host bridge.
                  rx_acceptance_queue.add(retired_rx_packet);
                }else {
                  rx_forwarding_queue.add(retired_rx_packet);
                };
              };
              TgtDone: {
                if(retired_rx_packet.BridgeBit() == 0) {
                  // This is a response from the fabric for the host bridge.
                  rx_acceptance_queue.add(retired_rx_packet);
                }else {
                  rx_forwarding_queue.add(retired_rx_packet);
                };
              };
              Broadcast: {
                // Broadcasts are dropped.
              };
              Fence: {
                // Use the fence to finish pushing packets forward in the host ordering
                rx_acceptance_queue.add(retired_rx_packet);
              };
              default: {
                // Nop, Sync, Reserved, DoubleWord_Read, DoubleWord_Write, Byte_Read, Byte_Write
                dut_error("A packet was put into the retirement queue that does not belong there.");
              };
            };
          };
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // reset() is a method used to initialize any internal fields that handle current
  // transmitting traffic.  Essentially this is a method to shutdown then entire channel
  // and prepare it to reinitialize the HT chain.
  ////////////////////////////////////////////////////////////////////////////////
  reset() is {
    ////////////////////////////////////////////////////////////////////////////////
    // Reinitialize all of the data pointers, counters, etc.
    ////////////////////////////////////////////////////////////////////////////////
    initiate_sync_sequence = FALSE;

    num_sync_packets_tx = 0;

    available_target_Posted_queue_buffers = 0;
    available_target_NonPosted_queue_buffers = 0;
    available_target_Response_queue_buffers = 0;
    available_target_Posted_Data_queue_buffers = 0;
    available_target_NonPosted_Data_queue_buffers = 0;
    available_target_Response_Data_queue_buffers = 0;

    tx_scheduling_queue = {};
    pending_tx_data_packet = NULL;
    current_tx_packet = NULL;
    current_tx_buffer = {};
    current_tx_buffer_type = Info;

    interrupted_tx_data_buffer = {};
    interrupted_tx_data_buffer_type = Info;
    interrupted_tx_ctl_buffer = {};
    interrupted_tx_ctl_buffer_type = Info;

    tx_transaction_history = {};
    tx_candidate_history = {};

    available_rx_Posted_queue_buffers = 0;
    available_rx_NonPosted_queue_buffers = 0;
    available_rx_Response_queue_buffers = 0;
    available_rx_Posted_Data_queue_buffers = 0;
    available_rx_NonPosted_Data_queue_buffers = 0;
    available_rx_Response_Data_queue_buffers = 0;

    rx_retirement_queue = {};
    rx_acceptance_queue = {};
    rx_forwarding_queue = {};

    rx_pending_Response_queue = {};
    rx_pending_Response_Data_queue = {};
    rx_outstanding_Broadcast_queue = {};
    expected_rx_data = NULL;

    rx_transaction_history = {};

    csco_ht_link.reset();
    start csco_ht_link.pre_reset_tx_path_initialization();
  };
//   ////////////////////////////////////////////////////////////////////////////////
//   // pre_reset_tx_initialization() is a TCM that will call the link level method to send CLK, 0 CTL and 1's on the CAD
//   // lines when the device first comes up.  This is done before the reset sequence is initiated.
//   ////////////////////////////////////////////////////////////////////////////////
//   pre_reset_tx_initialization()@csco_ht_link.csco_ht_tx_clk is {
//     // On the transmit side, this behavior is the same for a cold or warm reset.
//     start csco_ht_link.pre_reset_tx_path_initialization();
//   };
  ////////////////////////////////////////////////////////////////////////////////
  // initialize_tx_channel() will take the transmitter through the proper link level resets.
  ////////////////////////////////////////////////////////////////////////////////
  initialize_tx_channel(reset_type : csco_ht_reset_type)@csco_ht_link.csco_ht_tx_clk is {
    var init_tx_bit_time : csco_ht_bit_time;
    var rx_buffers_need_to_be_released : bool;
    var nop : csco_ht_packet;
    var init_tx_candidates : list of csco_ht_packet;

    // Use local copies of these variables so that the device can be reset.
    var temp_initiate_sync_sequence_delay : uint = initiate_sync_sequence_delay;
    var temp_sync_sequence_length : uint = sync_sequence_length;
//     ////////////////////////////////////////////////////////////////////////////////
//     // Reinitialize all of the data pointers, counters, etc.
//     ////////////////////////////////////////////////////////////////////////////////
//     num_sync_packets_tx = 0;

//     available_target_Posted_queue_buffers = 0;
//     available_target_NonPosted_queue_buffers = 0;
//     available_target_Response_queue_buffers = 0;
//     available_target_Posted_Data_queue_buffers = 0;
//     available_target_NonPosted_Data_queue_buffers = 0;
//     available_target_Response_Data_queue_buffers = 0;

//     tx_scheduling_queue = {};
//     pending_tx_data_packet = NULL;
//     current_tx_packet = NULL;
//     current_tx_buffer = {};
//     current_tx_buffer_type = Info;

//     interrupted_tx_data_buffer = {};
//     interrupted_tx_data_buffer_type = Info;
//     interrupted_tx_ctl_buffer = {};
//     interrupted_tx_ctl_buffer_type = Info;

//     tx_transaction_history = {};
//     tx_candidate_history = {};
    ////////////////////////////////////////////////////////////////////////////////

    start csco_ht_link.initialize_path();
    start csco_ht_link.initialize_tx_path();
    start tx_packets();

    // Once the PWROK signal is asserted, and the RESET signal is deasserted, sample the CAD lines.
    wait true(parent_device.PWROK == TRUE && parent_device.RESET_L == FALSE);
    // After reset is unasserted, wait some length of time and then switch the value of ctl across a posedge clk.
    // Wait at least some user specified number of cycles before transitioning ctl.
    // And ensure that this transition occurs on a posedge of the clk.
    while(temp_initiate_sync_sequence_delay != 0 || (temp_initiate_sync_sequence_delay == 0 && csco_ht_link.csco_ht_tx_clk_state == 0)) {
      init_tx_bit_time = generate_init_tx_bit_time(0, 1);
      csco_ht_link.tx_bit_time = init_tx_bit_time;
      emit csco_ht_link.csco_ht_tx_buffer_ready;
      if(temp_initiate_sync_sequence_delay != 0) {
        temp_initiate_sync_sequence_delay -= 1;
      };
      wait cycle;
    };
    // Once the target has asserted its ctl signal start the sync_sequence.  It must be >= 16 bit times.
    while(!initiate_sync_sequence
          || (initiate_sync_sequence && (temp_sync_sequence_length != 0 || (temp_sync_sequence_length == 0 && csco_ht_link.csco_ht_tx_clk_state == 0)))) {
            init_tx_bit_time = generate_init_tx_bit_time(1, 1);
            csco_ht_link.tx_bit_time = init_tx_bit_time;
            emit csco_ht_link.csco_ht_tx_buffer_ready;
            if(initiate_sync_sequence && temp_sync_sequence_length != 0) {
              temp_sync_sequence_length -= 1;
            };
            wait cycle;
          };
    for i from 1 to CSCO_HT_NUM_TX_RX_POINTER_INIT_CYCLES do {
      init_tx_bit_time = generate_init_tx_bit_time(0, 0);
      csco_ht_link.tx_bit_time = init_tx_bit_time;
      emit csco_ht_link.csco_ht_tx_buffer_ready;
      wait cycle;
    };
    // Frame the incomming packets.
    for i from 1 to CSCO_HT_NUM_PACKET_FRAMING_CYCLES do {
      init_tx_bit_time = generate_init_tx_bit_time(0, 1);
      csco_ht_link.tx_bit_time = init_tx_bit_time;
      emit csco_ht_link.csco_ht_tx_buffer_ready;
      wait cycle;
    };
    // Link level initialization is complete.
    // Allow the normal transmit routine to initialize the targets buffers.
    emit csco_ht_link.csco_ht_tx_initialization_complete;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // initialize_rx_channel() will take the receiver through the proper link level sequence.
  ////////////////////////////////////////////////////////////////////////////////
  initialize_rx_channel(reset_type : csco_ht_reset_type)@csco_ht_link.csco_ht_rx_clk is {
    var init_rx_bit_time : csco_ht_bit_time;
    var sync_sequence_count : uint;
    sync_sequence_count = 0;
    var tx_rx_pointer_init_sequence_count : uint;
    tx_rx_pointer_init_sequence_count = 0;
    var packet_framing_count : uint;
    packet_framing_count = 0;
    var sync_sequence_complete : bool;
    sync_sequence_complete = FALSE;
    var tx_rx_pointer_init_complete : bool;
    tx_rx_pointer_init_complete = FALSE;
    var packet_framing_complete : bool;
    packet_framing_complete = FALSE;
    var width: uint;

//     ////////////////////////////////////////////////////////////////////////////////
//     // Reinitialize all of the data pointers, counters, etc.
//     ////////////////////////////////////////////////////////////////////////////////
//     available_rx_Posted_queue_buffers = 0;
//     available_rx_NonPosted_queue_buffers = 0;
//     available_rx_Response_queue_buffers = 0;
//     available_rx_Posted_Data_queue_buffers = 0;
//     available_rx_NonPosted_Data_queue_buffers = 0;
//     available_rx_Response_Data_queue_buffers = 0;

//     rx_retirement_queue = {};
//     rx_acceptance_queue = {};
//     rx_forwarding_queue = {};

//     rx_pending_Response_queue = {};
//     rx_pending_Response_Data_queue = {};
//     rx_outstanding_Broadcast_queue = {};
//     expected_rx_data = NULL;

//     rx_transaction_history = {};
    ////////////////////////////////////////////////////////////////////////////////

    start csco_ht_link.initialize_rx_path();
    start rx_packets();

    if(control_block.read_device_function() == Host_Bridge) {
      start host_bridge_packet_routing();
    }else {
      start io_device_packet_routing();
    };

    // Once the PWROK signal is asserted, and the RESET signal is deasserted, sample the CAD lines.
    wait true(parent_device.PWROK == TRUE && parent_device.RESET_L == FALSE);

    // Sample the incomming CAD bits to determine the width of the incomming transmitter
    sync @csco_ht_link.csco_ht_rx_buffer_ready;
    init_rx_bit_time = csco_ht_link.rx_bit_time;
    // After a cold reset all link widths with come up as 8 bit links.  (Since we are ignoring 2 & 4 bits possibilities.)
    // In order to achieve a link width greater than 8 bits, the host bridge has to read all of the CSR's and manually
    // configure the widths, then reset the entire bus.
    case (init_rx_bit_time.cad[init_rx_bit_time.link_width.as_a(uint)-1:0]) {
      8'hFF: {
        // 8 bits were received on this link.
      };
      16'hFFFF: {
        // 16 bits were received on this link.
      };
      32'hFFFFFFFF: {
        // 32 bits were received on this link.
      };
      default: {
        dut_error("The received has detected that this link is attempting to negotiate a width < 8 bits.");
      };
    };

    // Poll the link to look for the transition of ctl with a posedge clk that indicates the start of the 1st sync packet.
    // Once the transition of ctl occurs, go an monitor the incoming sync packets.
    while(!initiate_sync_sequence) {
      sync @csco_ht_link.csco_ht_rx_buffer_ready;
      init_rx_bit_time = csco_ht_link.rx_bit_time;
      case {
        init_rx_bit_time.clk[0] == 0 && init_rx_bit_time.ctl == 1: {
          sys.csco_ht_errors.report_init_error(INIT_ILLEGAL_CTL_CHANGE, "The transmitter transitioned the ctl signal across a clk negedge.", control_block, channel_identifier);
        };
        init_rx_bit_time.clk[0] == 1 && init_rx_bit_time.ctl == 1: {
          // Start of the first sync packet.
          initiate_sync_sequence = TRUE;
          sync_sequence_count += 1;
        };
        default: {
          // keep waiting for the sync sequence.
        };
      };
      wait cycle;
    };
    // Monitor and ensure that the correct number of packets are received to complete the sync sequence
    while(!sync_sequence_complete) {
      sync @csco_ht_link.csco_ht_rx_buffer_ready;
      init_rx_bit_time = csco_ht_link.rx_bit_time;
      width = init_rx_bit_time.link_width.as_a(uint);

      case {
          ((init_rx_bit_time.ctl == 1) &&
           (init_rx_bit_time.cad[width-1:0] == ((1 << width) - 1))): {
          sync_sequence_count += 1;
        };
          (init_rx_bit_time.ctl == 0 && init_rx_bit_time.cad == 0): {
          // This is the first packet of the tx/rx pointer initialization sequence.
          tx_rx_pointer_init_sequence_count += 1;
          if(sync_sequence_count >= CSCO_HT_MIN_SYNC_SEQUENCE_CYCLES) {
            sync_sequence_complete = TRUE;
          }else {
            var error_string : string;
            error_string = appendf("The minimum number of bit times in the sync sequence was violated.  Only %d cycles were seen.", sync_sequence_count);
            sys.csco_ht_errors.report_init_error(INIT_MIN_SYNC_SEQ, error_string, control_block, channel_identifier);
          };
        };
        default: {
          var error_string : string;
          error_string = appendf("Unexpected bit pattern received during the sync sequence.  Only %d cycles were seen.", sync_sequence_count);
          sys.csco_ht_errors.report_init_error(INITIALIZATION_ERROR, error_string, control_block, channel_identifier);
        };
      };
      wait cycle;
    };
    // Monitor and ensure that the correct number of packets are received to complete the tx/rx pointer init sequence.
    // Both API & AMD have their core wait in this state until the TX on their RX port complete the transmission of the 16
    // 1/1 packets (sync packets) and begin the transmission of 0/0 (512 window) before they begin their 512 window.  This
    // ensures that the other side comes out of the sync stage.
    while(!tx_rx_pointer_init_complete) {
      sync @csco_ht_link.csco_ht_rx_buffer_ready;
      init_rx_bit_time = csco_ht_link.rx_bit_time;
      width = init_rx_bit_time.link_width.as_a(uint);

      case {
          init_rx_bit_time.ctl == 0 && init_rx_bit_time.cad == 0: {
          tx_rx_pointer_init_sequence_count += 1;
        };
          init_rx_bit_time.ctl == 0 && init_rx_bit_time.cad[width-1:0] == ((1 << width) - 1): {
          packet_framing_count += 1;
          // This was added to accomodate the current API HT core which violates the 512 cycle count.
          if(tx_rx_pointer_init_sequence_count == CSCO_HT_NUM_TX_RX_POINTER_INIT_CYCLES) {
            tx_rx_pointer_init_complete = TRUE;
          }else if(tx_rx_pointer_init_sequence_count > CSCO_HT_NUM_TX_RX_POINTER_INIT_CYCLES) {
            var error_string : string;
            error_string = appendf("The maximum number of bit times in the tx/rx pointer init sequence was violated.  Only %d cycles were seen.", tx_rx_pointer_init_sequence_count);
            sys.csco_ht_errors.report_init_error(INIT_MAX_LOAD_PTR_SEQ, error_string, control_block, channel_identifier);
          }else {
            var error_string : string;
            error_string = appendf("The minimum number of bit times in the tx/rx pointer init sequence was violated.  Only %d cycles were seen.", tx_rx_pointer_init_sequence_count);
            sys.csco_ht_errors.report_init_error(INIT_MIN_LOAD_PTR_SEQ, error_string, control_block, channel_identifier);
          };
        };
        default: {
          var error_string : string;
          error_string = appendf("Unexpected bit pattern received during tx/rx pointer initialization sequence.  Only %d cycles were seen.", tx_rx_pointer_init_sequence_count);
          sys.csco_ht_errors.report_init_error(INITIALIZATION_ERROR, error_string, control_block, channel_identifier);
        };
      };
      wait cycle;
    };
    // Monitor the transmission used to frame the first incomming packet.
    while(!packet_framing_complete) {
      sync @csco_ht_link.csco_ht_rx_buffer_ready;
      init_rx_bit_time = csco_ht_link.rx_bit_time;
      width = init_rx_bit_time.link_width.as_a(uint);

      case {
        init_rx_bit_time.ctl == 0 && init_rx_bit_time.cad[width-1:0] == ((1 << width) - 1): {
          packet_framing_count += 1;
        };
        init_rx_bit_time.ctl == 1: {
          // Wait to terminate the initialization monitoring until it is determined that a CTL packet immediately followed
          // the packet framing sequence.

          // Once the framing sequence is completed, check to make sure that the correct number of bit times occured.
          if(packet_framing_count > CSCO_HT_NUM_PACKET_FRAMING_CYCLES) {
            var error_string : string;
            error_string = appendf("The maximum number of bit times in the packet framing sequence was violated.  %d cycles were seen.", packet_framing_count);
            sys.csco_ht_errors.report_init_error(INIT_MAX_FRAMING_SEQ, error_string, control_block, channel_identifier);
          }else if(packet_framing_count < CSCO_HT_NUM_PACKET_FRAMING_CYCLES) {
            var error_string : string;
            error_string = appendf("The minimum number of bit times in the packet framing sequence was violated.  Only %d cycles were seen.", packet_framing_count);
            sys.csco_ht_errors.report_init_error(INIT_MIN_FRAMING_SEQ, error_string, control_block, channel_identifier);
          }else {
            // The correct number of packet framing cycles were observed.
          };
          // Complete the initalization monitor.
          packet_framing_complete = TRUE;
        };
        default: {
          var error_string : string;
          error_string = appendf("Unexpected bit pattern received during packet framing sequence.  Only %d cycles were seen.", packet_framing_count);
          sys.csco_ht_errors.report_init_error(INITIALIZATION_ERROR, error_string, control_block, channel_identifier);
        };
      };
      wait cycle;
      if(packet_framing_count == CSCO_HT_NUM_PACKET_FRAMING_CYCLES) {
        // After the correct number of framing cycles are seen, emit the event to indicate that the initialization
        // is complete.  Then go on to check to make sure that the framing pattern is not violated.
        emit csco_ht_link.csco_ht_rx_initialization_complete;
      };
    };
    // End of receive path initialization.
  };
  ////////////////////////////////////////////////////////////////////////////////
  // generate_init_tx_bit_time() will generate an csco_ht_bit_time for the transmit link with
  // all of the cad bits set to the same value.
  ////////////////////////////////////////////////////////////////////////////////
  generate_init_tx_bit_time(ctl_state : bit, cad_state : bit):csco_ht_bit_time is {
    result = new;

    result.link_width = control_block.current_tx_width[channel_identifier];
    result.ctl = ctl_state;
    for i from 0 to (result.link_width.as_a(uint)/8)-1 do {
        result.clk.add(csco_ht_link.csco_ht_tx_clk_state);
    };
    result.cad = (cad_state == 1) ? ((cad_state << result.link_width.as_a(uint)) - 1) : 0;
  };
  /////////////////////////////////////////////////////////////////////////////
  // print_channel_info() is a basic method for formating informational
  // message to be printed.
  //
  // spadix - print_channel_info() was replaced by messagef(), short_name()
  //          and a message logger
  /////////////////////////////////////////////////////////////////////////////

  logger : message_logger;
    keep soft logger.tags == {NORMAL};

  short_name(): string is {
    if(control_block.read_link_orientation(channel_identifier) == TX_Upstream_RX_Downstream_Interface) {
      result = "Upstream Link";
    } else {
      result = "Downstream Link";
    };
  };

  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// A basic struct used to hold a list of packets that were candidates for transmission at a given time and
// record the associated simulator time.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_tx_candidate_list {
  !candidates : list of csco_ht_packet;
  time : uint;
  keep me.time == sys.time;
};
////////////////////////////////////////////////////////////////////////////////
// A structure used to know what response will expected once an appropriate ctl packet is transmitted.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_rx_response_buffer {
  unitID : uint (bits:5);
  SrcTag : uint (bits:5);
  command : csco_ht_packet_encoding;
  endian_type: csco_ht_endian_type;
  !response_token : uint (bits:10);

  post_generate() is also {
    response_token = %{unitID, SrcTag};
  };
};
////////////////////////////////////////////////////////////////////////////////
// A structure used to track what the next incoming data packet will be, and what ctl packet it should be associated with.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_rx_pending_data_info {
  associated_ctl_packet : csco_ht_packet;
  endian_type: csco_ht_endian_type;
  data_type : csco_ht_packet_encoding;
  data_length : uint; // Number of bytes in the packet.
  channel : csco_ht_virtual_channel;
};
////////////////////////////////////////////////////////////////////////////////
'>
