////////////////////////////////////////////////////////////////////////////
// Title         : HT Device Control
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////
// File          : csco_ht_device_control.e
// Author        : Brian Van Essen  <vanessen@cisco.com>
// Created       : 25.01.2001
// Last modified : 01.05.2001
////////////////////////////////////////////////////////////////////////////
// Description :
// Each device in an HT chain has one csco_ht_device_control_block associated with it.
// This provides a central resource for all ht control information.
// Within a control block the index of a list indicates the associated link interface.
////////////////////////////////////////////////////////////////////////////
// Modification history :
// 25.01.2001 : created
////////////////////////////////////////////////////////////////////////////

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****

<'
////////////////////////////////////////////////////////////////////////////

struct csco_ht_device_control_block {
  device_function : csco_ht_device_encoding;               // Designates if this is a Host Bridge or an IO device
  device_reference_name : string;                          // The name of the E instance of the controlling Device

  num_device_csco_ht_interfaces : uint;                    // The number of ht interfaces that this device has.
  keep num_device_csco_ht_interfaces in [1..2];            // Currently restrict this to a max of 2 interfaces.

  // For an EOC device or a Tunnel device, link orientation is based uponthe Master Host Bit in the Capability Register
  // For a Host device the link orientation is fixed.
  link_orientation : list of csco_ht_link_encoding;        // indicates the direction of the tx and rx path for this interface.
  keep me.link_orientation.size() == num_device_csco_ht_interfaces;
  // Since the Master Host bit is set to 0, the default orientation for this device is as follows.
  keep for each in me.link_orientation {
    (me.device_function == Tunnel_IO_Device or me.device_function == EOC_IO_Device) && index == 0 => it == TX_Upstream_RX_Downstream_Interface;
    me.device_function == Host_Bridge && index == 0 => it == TX_Downstream_RX_Upstream_Interface;
    index == 1 => it == TX_Downstream_RX_Upstream_Interface;
  };
  // A flag used by the Host Bridge to indicate when the HT chain has been fully initialized.
  chain_initialization_complete : bool;
  keep soft chain_initialization_complete == FALSE;
  ////////////////////////////////////////////////////////////////////////////////
  // These are a series of fields which hold the current values of the device parameters which have been extracted
  // from the CSRs
  ////////////////////////////////////////////////////////////////////////////////
  current_UnitID : list of uint (bits:5);
  keep soft me.current_UnitID.size() == 1;
  keep soft me.current_UnitID[0] == 0;
  // A keyed list used to maintain a list of outstanding SrcTags associated with each unitID.
  !current_tx_SrcTag : list (key:UnitID) of csco_ht_unitID_srctag_pair;

  current_tx_width : list of csco_ht_link_size;
  keep me.current_tx_width.size() == num_device_csco_ht_interfaces;
  keep for each in current_tx_width {
    it == Eight_Bits;
  };
  current_rx_width : list of csco_ht_link_size;
  keep me.current_rx_width.size() == num_device_csco_ht_interfaces;
  keep for each in current_rx_width {
    it == Eight_Bits;
  };
  // This is a uint that is a multiplier for the csco_ht_fundamental_clock event to create a clock of the proper frequency.
  current_tx_clock_freq : list of csco_ht_clock_multiplier;
  keep me.current_tx_clock_freq.size() == num_device_csco_ht_interfaces;
  keep for each in me.current_tx_clock_freq {
    sys.csco_ht_use_direct_csco_ht_clock == FALSE => it == csco_ht_200MHz;
    sys.csco_ht_use_direct_csco_ht_clock == TRUE => it == csco_ht_DirectDrive;
  };

  // Configuration Space Registers
  device_header : csco_ht_device_header;
  capability_register : csco_ht_capability_header;

  // This bit indicates which link a csr read or write came in on.  It is used to keep the Master Host bit in the capabilities
  // register up to date.  When ever a device gets a csr request it must update this bit.
  csr_inbound_link : bit;
  keep csr_inbound_link == 0;

  // Flags to indicate if this device is on the path to the southbridge or is the southbridge.  Currently these are
  // unused.
  southbridge : bool;
  keep soft southbridge == FALSE;
  southbridge_path : bool;
  keep soft southbridge_path == FALSE;

  // Flag to record traffic at the channel level;
  record_traffic : bool;
  keep soft me.record_traffic == FALSE;
  record_all_nop : bool;
  keep soft me.record_all_nop == FALSE;

  // A sync pattern must be generated on all TX lines for at least 16 bit times for 8, 16, 32 bit links.
  Sync_Flood_Init : bool;
  keep Sync_Flood_Init == FALSE;
//  Sync_Flood_Complete : bool;
//  keep Sync_Flood_Complete == FALSE;
//  event Sync_Flood_Init;
  ////////////////////////////////////////////////////////////////////////////////
  // Public methods needed to extract data from the control block.
  ////////////////////////////////////////////////////////////////////////////////
  // Perform a hardware write to the read-only Initialization Complete bit.
  ////////////////////////////////////////////////////////////////////////////////
  link_initialization_complete(link_id : uint) is {
    // If the Link Failure bit is set, don't set the link initialized.
    if(link_id == 0) {
      //KMM if(capability_register.regs.key(Link_Control0).Link_Control0'reg_type'Link_Failure != 1) {
      //KMM  capability_register.regs.key(Link_Control0).Link_Control0'reg_type'Initialization_Complete = 1;
      //KMM  capability_register.regs.key(Link_Control0).update_packed_data();

      if capability_register.regs.key(Link_Control0) is a Link_Control0'reg_type csco_ht_capability_reg(tmpreg) {
        if(tmpreg.Link_Failure != 1) {
                tmpreg.Initialization_Complete = 1;
                tmpreg.update_packed_data();
        }else {
                dut_error("Attempting to initialize a link which has the Link Failure Bit Set.");
        };
      };
    }else {
      if capability_register.regs.key(Link_Control1) is a Link_Control1'reg_type csco_ht_capability_reg(tmpreg) {
        if(tmpreg.Link_Failure != 1) {
                tmpreg.Initialization_Complete = 1;
                tmpreg.update_packed_data();
        }else {
                dut_error("Attempting to initialize a link which has the Link Failure Bit Set.");
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Indicates when the physical link is up and running.
  ////////////////////////////////////////////////////////////////////////////////
  physical_link_initialized(link_id : uint):bool is {
    if(capability_register.links_initialized()[link_id] == 1 && tx_off(link_id) == FALSE) {
      result = TRUE;
    }else {
      result = FALSE;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Indicates that the config registers for the device have also been initialized.
  // Since there is no real way to determine when the configuration of a given device
  // is done, so we will assume that once the UnitID is set the device is ready to go.
  ////////////////////////////////////////////////////////////////////////////////
  logical_link_initialized(link_id : uint):bool is {
    var BARs_initialized : bool = FALSE;
    var UnitID_initialized : bool = FALSE;
    // Until the BAR's and the UnitID are set, the device is not considered initialized.
    if(physical_link_initialized(link_id) == TRUE) {
      for each (bar) in device_header.Base_Address_Register {
        if((index % 2 == 0 && bar.data[31:3] != 29'h000_0000) || index % 2 == 1 && bar.data[31:0] != 32'h0000_0000) {
          BARs_initialized = TRUE;
        };
      };

      if(device_function != Host_Bridge) {
        if(current_UnitID[0] != 5'b00000) {
          UnitID_initialized = TRUE;
        };
      }else {
        if(chain_initialization_complete == TRUE) {
          UnitID_initialized = TRUE;
        };
      };

      if(BARs_initialized && UnitID_initialized) {
        result = TRUE;
      }else {
        result = FALSE;
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Generate a new SrcTag for a given UnitID
  ////////////////////////////////////////////////////////////////////////////////
  new_SrcTag(ID : uint (bits:5)):uint (bits:5) is {
    gen result keeping {
      it not in current_tx_SrcTag.key(ID).tag_list;
    };
    current_tx_SrcTag.key(ID).tag_list.add(result);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Retire a SrcTag associated with a particular UnitID from use.
  ////////////////////////////////////////////////////////////////////////////////
  retire_SrcTag(ID : uint (bits:5), Tag : uint (bits:5)) is {
    if(current_tx_SrcTag.key(ID).tag_list.key_exists(Tag) == TRUE) {
      current_tx_SrcTag.key(ID).tag_list.delete(current_tx_SrcTag.key(ID).tag_list.key_index(Tag));
    } else {
      dut_error("Trying to retire a SrcTag that is not in use.\n");
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Check to see if there are 32 outstanding NonPosted transactions for this UnitID.
  ////////////////////////////////////////////////////////////////////////////////
  free_SrcTag(ID : uint (bits:5)):bool is {
    result = (current_tx_SrcTag.key(ID).tag_list.size() != 32);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Check to see if the link is currently marked as the End Of Chain
  ////////////////////////////////////////////////////////////////////////////////
  EOC(link_id : uint):bool is {
    result = (capability_register.links_eoc_setting()[link_id] == 1);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Check to see if the link's transmitter is marked as off.
  ////////////////////////////////////////////////////////////////////////////////
  tx_off(link_id : uint):bool is {
    result = (capability_register.links_active()[link_id] == 1);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return the value of the device function field.
  ////////////////////////////////////////////////////////////////////////////////
  read_device_function():csco_ht_device_encoding is {
    result = device_function;
  };

  ////////////////////////////////////////////////////////////////////////////////
  // Return the string of the device function field.
  ////////////////////////////////////////////////////////////////////////////////
  print_device_function():string is {
    if(device_function == Host_Bridge) {
      result = "Host_Bridge";
    } else if (device_function == Tunnel_IO_Device) {
      result = "Tunnel_IO_Dev";
    } else {
      result = "EOC_IO_Dev";
    };
  };

  ////////////////////////////////////////////////////////////////////////////////
  // Return the link orientation for a particular link.
  ////////////////////////////////////////////////////////////////////////////////
  read_link_orientation(link_index : uint):csco_ht_link_encoding is {
    result = link_orientation[link_index];
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Check to see if this device controls the current UnitID.
  ////////////////////////////////////////////////////////////////////////////////
  compare_current_UnitID(UnitID : uint(bits:5)):bool is {
    var match_found : bool;
    match_found = FALSE;

    for each in current_UnitID {
      if(UnitID == it) {
        match_found = TRUE;
      };
    };
    result = match_found;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return the tx width of the current link.
  ////////////////////////////////////////////////////////////////////////////////
  read_current_tx_width(link_index : uint):csco_ht_link_size is {
    result = current_tx_width[link_index];
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return the maximum possible tx width of the current link.
  ////////////////////////////////////////////////////////////////////////////////
  read_max_tx_width(link_index : uint):csco_ht_link_size is {
    var temp_list : list of csco_ht_link_size;

    temp_list = capability_register.read_max_tx_link_widths();

    result = temp_list[link_index];
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return the rx width of the current link.
  ////////////////////////////////////////////////////////////////////////////////
  read_current_rx_width(link_index : uint):csco_ht_link_size is {
    result = current_rx_width[link_index];
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return the current clock frequency of the tx link.
  ////////////////////////////////////////////////////////////////////////////////
  read_current_tx_clock_freq(link_index : uint):uint is {
    result = current_tx_clock_freq[link_index].as_a(uint);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return the value of the greatest clock frequency for all tx links of this device.
  ////////////////////////////////////////////////////////////////////////////////
  read_max_current_tx_clock_freq():uint is {
    var max_clk : uint;
    // Initialize the value to the first element in the list.
    max_clk = current_tx_clock_freq[0].as_a(uint);
    for each in current_tx_clock_freq {
      if(it.as_a(uint) < max_clk) {
        max_clk = it.as_a(uint);
      };
    };
    result = max_clk;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Check to see if an address exists in the range given by the csr BAR registers.
  ////////////////////////////////////////////////////////////////////////////////
  decode_address(addr : uint (bits:40)):csco_ht_memory_range is {
    case {
      // 00_0000_0000h -> FC_FFFF_FFFFh - DRAM / Memory-Mapped I/O
      addr[39:32] <= 8'hFC: {
        result = Memory_Mapped_IO;
      };
      // FD_F800_0000h -> FD_F8FF_FFFFh - Interrupts
      addr[39:24] == 16'hFDF8: {
        result = Interrupts;
      };
      // FD_FC00_0000h -> FD_FDFF_FFFFh - PCI I/O
      addr[39:24] >= 16'hFDFC && addr[39:24] <= 16'hFDFD: {
        result = PCI_IO;
      };
      // FD_FE00_0000h -> FD_FFFF_FFFFh - PCI Configuration
      addr[39:24] >= 16'hFDFE && addr[39:24] <= 16'hFDFF: {
        result = PCI_Configuration;
      };
      // FD_0000_0000h -> FD_F7FF_FFFFh - Reserved
      // FD_F800_0000h -> FD_F8FF_FFFFh - Interrupts
      // FD_F900_0000h -> FD_F90F_FFFFh - IACK
      // FD_F910_0000h -> FD_F91F_FFFFh - System Management
      // FD_F920_0000h -> FD_FBFF_FFFFh - Reserved
      // FE_0000_0000h -> FF_FFFF_FFFFh - Reserved
      default: {
        // If the address request falls into this range just forward it along the chain.
        result = Reserved;
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Perform a read to the CSR space.  To indicate a read error return a empty list.
  // By definition in the HT spec, CSR accesses are always/only 1 doubleword (32 bits) at a time.
  ////////////////////////////////////////////////////////////////////////////////
  csr_read(addr : uint (bits:40), byte_mask : uint (bits:4)):list of bit is {
    if(addr[7:0] >= 0 && addr[7:0] <= 8'h3C) { // This address exists in the device header.
      result = device_header.read(addr[7:0], byte_mask);
    }else if(addr[7:0] >= CSCO_HT_CAPABILITIES_OFFSET
             && ((addr[7:0] <= CSCO_HT_CAPABILITIES_OFFSET + 8'h0C && (device_function == Tunnel_IO_Device || device_function == EOC_IO_Device))
                 || (addr[7:0] <= CSCO_HT_CAPABILITIES_OFFSET + 8'h08 && device_function == Host_Bridge))) {
                   // Make sure that the CSCO_HT_CAPABILITIES_OFFSET is subtracted so that the offset into the capabilities reg is correct.
                   result = capability_register.read(addr[7:0] - CSCO_HT_CAPABILITIES_OFFSET, byte_mask);
                 }else {
                   // To indicate an error, return an null list.
                   result = {};
                 };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Perform a write to the CSR space.  An error is indicated by returning TRUE.
  ////////////////////////////////////////////////////////////////////////////////
  csr_write(addr : uint (bits:40), byte_mask : uint (bits:4), data : uint):bool is {
    if(addr[7:0] >= 0 && addr[7:0] <= 8'h3C) { // This address exists in the device header.
      device_header.write(addr[7:0], byte_mask, data);
      result = FALSE;
    }else if(addr[7:0] >= CSCO_HT_CAPABILITIES_OFFSET
             && ((addr[7:0] <= CSCO_HT_CAPABILITIES_OFFSET + 8'h0C && (device_function == Tunnel_IO_Device || device_function == EOC_IO_Device))
                 || (addr[7:0] <= CSCO_HT_CAPABILITIES_OFFSET + 8'h08 && device_function == Host_Bridge))) {
                   // Make sure that the CSCO_HT_CAPABILITIES_OFFSET is subtracted so that the offset into the capabilities reg is correct.
                   capability_register.write(addr[7:0] - CSCO_HT_CAPABILITIES_OFFSET, byte_mask, data);
                   result = FALSE;
                 }else {
                   // To indicate an error, return TRUE;
                   result = TRUE;
                 };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // This is a method that is used to set the csr's to some value other than the default.
  ////////////////////////////////////////////////////////////////////////////////
  pre_load_csr() is empty;
  ////////////////////////////////////////////////////////////////////////////////
  // Extract the current values of the CSRs and push the data into the local fields.
  ////////////////////////////////////////////////////////////////////////////////
  update_extracted_csr_fields() is {
    // Update the extracted fields.
    var prev_base_UnitID : uint (bits:5) = current_UnitID[0];
    current_UnitID = capability_register.read_UnitID();
    if(prev_base_UnitID != current_UnitID[0] || current_tx_SrcTag.is_empty()) {
      // If the Base UnitID has changed, delete all of the outstanding SrcTags. Note, if this is done while there are outstanding
      // SrcTags an error will be generated.  That should be okay, as this shouldn't happen.
      current_tx_SrcTag.clear();
      var temp_csco_ht_unitID_srctag_pair : csco_ht_unitID_srctag_pair;
      for each (ID) in current_UnitID {
        gen temp_csco_ht_unitID_srctag_pair keeping {
          .UnitID == ID;
        };
        current_tx_SrcTag.add(temp_csco_ht_unitID_srctag_pair);
      };
    };

    if(device_function == Tunnel_IO_Device) { // Only a Tunnel Device can have its link orientation change.
      for each in me.link_orientation { // The Master host is off of the link indicated by the Master Host Bit.
        if(index == capability_register.master_host_link()) {
          me.link_orientation[index] = TX_Upstream_RX_Downstream_Interface;
        }else {
          me.link_orientation[index] = TX_Downstream_RX_Upstream_Interface;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Reset all of the local fields and the CSRs to their default value.
  ////////////////////////////////////////////////////////////////////////////////
  reset(reset_type : csco_ht_reset_type) is {
    Sync_Flood_Init = FALSE;
//    Sync_Flood_Complete = FALSE;
    // Reset all of the csr's
    device_header.reset(reset_type);
    capability_register.reset(reset_type);
    // Provide and empty function to allow for easy preload of the csr's.
    pre_load_csr();
    // Update the extracted fields.
    update_extracted_csr_fields();
    // Link width and Freq can only be changed on a reset.
    current_tx_width = capability_register.read_tx_link_widths();
    current_rx_width = capability_register.read_rx_link_widths();
    current_tx_clock_freq = capability_register.read_tx_clock_freq();
  };
};
////////////////////////////////////////////////////////////////////////////////
// A structure used to keep track of a UnitID and all of the outstanding SrcTags associated with that
// UnitID.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_unitID_srctag_pair {
  UnitID : uint (bits:5);
  // A list of all currently used SrcTag's
  !tag_list : list (key:it) of uint (bits:5);
};
////////////////////////////////////////////////////////////////////////////////
// csco_ht_basic_reg is a basic structure used to create the CSRs for HT.  It includes a list of packed
// data and a list of packed write mask data.  This allows the CSRs to be updated by
// indexing into the list of packed data and then unpacking that information back into
// the expanded fields.
// This structure also provides methods to pack the register into an HT transaction
// and perform the appropriate read or write.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_basic_reg {
  parent_device : csco_ht_device;
  keep parent_device == get_enclosing_unit(csco_ht_device);
  // Size of the register in bits.
  reg_size : csco_ht_reg_size;
  temp_reg_size : uint;
  keep temp_reg_size == reg_size.as_a(uint);
  keep temp_reg_size >= 8;
  keep temp_reg_size <= 256;

  reg_name : string;
  keep soft reg_name == "";

  !data : list of bit;

  // reset value of the register
  !reset_value: list of bit;

  write_mask : list of bit;
  keep write_mask.size() == reg_size.as_a(uint);

  // soft constrain the functional write mask to indicate
  // all the bit fields that are writeable and wont hang the chip
  func_write_mask : list of bit;
  keep func_write_mask.size() == reg_size.as_a(uint);

  //soft constrain the set mask to indicate
  //all of the bit fields that should be set to a 1
  set_mask : list of bit;
  keep set_mask.size() == reg_size.as_a(uint);
  keep for each in set_mask {
    soft it == 1'b0;
  };


  // soft constrain the read mask to indicate
  // all the bit fields that valids (ie: ignore persist bits)
  read_mask : list of bit;
  keep read_mask.size() == reg_size.as_a(uint);

  // add flags it register has been read or written...
  read_flag: bool;
  keep read_flag == FALSE;

  write_flag: bool;
  keep write_flag == FALSE;

  !expected_data : list of bit; // A place to hold the data that is expected back from a read_ht.
  ignore_data_compare : bool; // A flag used to indicate that the data should not be compared against the expected data.
  keep soft ignore_data_compare == FALSE;

  // A flag used to indicate that the data should be compared taking into account the read mask.
  read_mask_compare : bool;
  keep soft read_mask_compare == FALSE;

  // flag to indicate whether the register address should be access directly via
  // HT or indirectly through some backdoor mechanism
  indirect_access: bool;
  keep soft indirect_access == TRUE;

  // Indicated where a register is positioned in a 32 bit range.
  // Since all HT access are based on DoubleWords (32 bits) this info is necessary.
  byte_offset : uint (bits:2);
  keep temp_reg_size == 16 => byte_offset in range [0, 2];
  keep temp_reg_size == 24 => byte_offset in range [0, 1];
  keep temp_reg_size >= 32 => byte_offset == 0; // If the register is larger than 32 bits, force a 0 byte offset.

  // List of parameters to constrain how register access occur.
  passing_rule : csco_ht_passing_rules;
  keep passing_rule != NA;
  response_passing_rule : csco_ht_passing_rules;
  keep response_passing_rule != NA;
  BW_latency_req : csco_ht_BW_latency_req;
  mem_access_type : csco_ht_mem_access_type;
  SeqID : uint (bits:4);
  UnitID : uint (bits:5);
  Compat : bit;
  read_data_type : csco_ht_data_type;
  // Regs less than 32 bits must use Byte access, 32 bit regs can use either Byte or DoubleWord access,
  // and regs > 32 bits must use Doubleword access.
  keep temp_reg_size < 32 => read_data_type == Byte;
  keep temp_reg_size > 32 => read_data_type == DoubleWord;
  read_mask_count : uint (bits:4);
  keep read_data_type == Byte && reg_size == Eight => read_mask_count == 1'b1 << byte_offset;
  keep read_data_type == Byte && reg_size == Sixteen => read_mask_count == 2'b11 << byte_offset;
  keep read_data_type == Byte && reg_size == TwentyFour => read_mask_count == 3'b111 << byte_offset;
  keep read_data_type == Byte && reg_size == ThirtyTwo => read_mask_count == 4'hF;
  keep read_data_type == DoubleWord => read_mask_count == value(reg_size.as_a(uint)/32-1);
  write_vChan : csco_ht_virtual_channel;
  keep write_vChan in [NonPosted, Posted];
  write_data_type : csco_ht_data_type;
  // For register that are smaller than 1 doubleword force a byte write.
  keep temp_reg_size < 32 => write_data_type == Byte;
  write_byte_mask : uint;
  keep temp_reg_size >= 32 => write_byte_mask == value((32'h0000_0001 << reg_size.as_a(uint)/8)-1);
  keep temp_reg_size < 32 => write_byte_mask[31:4] == 28'h0000_000;
  keep temp_reg_size == 8 => write_byte_mask == 1'b1 << byte_offset;
  keep temp_reg_size == 16 => write_byte_mask == 2'b11 << byte_offset;
  keep temp_reg_size == 24 => write_byte_mask == 3'b111 << byte_offset;
  Addr : uint (bits:40);
  keep Addr[1:0] == 2'b0;

  // These are required by a bug in 3.3.2 with generation of temporary fields and inheritance.
  !temp_data : list of bit;
  !last_read_request : csco_ht_packet;
  !last_read_response : csco_ht_packet;
  !last_read_data : csco_ht_packet;
  !last_write_request : csco_ht_packet;
  !last_write_response : csco_ht_packet;
  !last_write_data : list of bit;
  ////////////////////////////////////////////////////////////////////////////////
  // After generation, pack the data.
  ////////////////////////////////////////////////////////////////////////////////
  post_generate() is also {
    data = pack(packing.low, me);
    reset_value = data.copy();
    check that (data.size() == reg_size.as_a(uint)) else
    dut_error("data size != reg size \n");

    // Update the expected data only after initial generation.  After that, it is up to the user to determine when
    // the field should be updated.
    update_expected_data();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // If the expanded fields are manually updated, allow a method to update the packed data.
  ////////////////////////////////////////////////////////////////////////////////
  update_packed_data () is {
    data = pack(packing.low, me);
    check that (data.size() == reg_size.as_a(uint)) else
    dut_error("data size != reg size\n");
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Push the packed data back into the expanded fields and then into the device control
  // blocks local fields.
  ////////////////////////////////////////////////////////////////////////////////
  update_fields () is {
    var tmp : csco_ht_basic_reg;
    tmp = me;
    unpack(packing.low, tmp.data, tmp);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Copy the value of the data field into the expected data field.
  ////////////////////////////////////////////////////////////////////////////////
  update_expected_data() is {
    expected_data = data.copy();
    check that (expected_data.size() == reg_size.as_a(uint)) else
    dut_error("expected data size != reg_size\n");
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Compare the data in the expected data field to that in the data field.
  ////////////////////////////////////////////////////////////////////////////////
  compare_data() is {
    if(ignore_data_compare == FALSE) {
      for i from 0 to reg_size.as_a(uint)-1 do {
        if(me.data[i] != me.expected_data[i]) {
          var error_msg : string;
          if(reg_name != "") {
            error_msg = appendf("%s Read Error occurred : Data Mismatch : Addr=0x%x Expected Data=0x%x Received Data=0x%x", reg_name, Addr, expected_data[reg_size.as_a(uint)-1:0], data[reg_size.as_a(uint)-1:0]);
          }else {
            error_msg = appendf("HT_BASIC_REG Read Error occurred : Data Mismatch : Addr=0x%x Expected Data=0x%x Received Data=0x%x", Addr, expected_data[reg_size.as_a(uint)-1:0], data[reg_size.as_a(uint)-1:0]);
          };
          sys.csco_ht_errors.report_transaction_error(UNMATCHED_READ_DATA, error_msg, parent_device.control_block, last_read_request, last_read_response);
          print();
          return;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Compare the data in the expected data field to that in the data field.
  ////////////////////////////////////////////////////////////////////////////////
  compare_rd_msk_data() is {
    if(ignore_data_compare == FALSE) {
      for i from 0 to reg_size.as_a(uint)-1 do {
        if((me.data[i] != me.expected_data[i]) && (read_mask[i] == 1'b1)){
          var error_msg : string;
          if(reg_name != "") {
            error_msg = appendf("%s Read Error occurred : Data Mismatch : Addr=0x%x Expected Data=0x%x Received Data=0x%x", reg_name, Addr, expected_data[reg_size.as_a(uint)-1:0], data[reg_size.as_a(uint)-1:0]);
          }else {
            error_msg = appendf("HT_BASIC_REG Read Error occurred : Data Mismatch : Addr=0x%x Expected Data=0x%x Received Data=0x%x", Addr, expected_data[reg_size.as_a(uint)-1:0], data[reg_size.as_a(uint)-1:0]);
          };
          sys.csco_ht_errors.report_transaction_error(UNMATCHED_READ_DATA, error_msg, parent_device.control_block, last_read_request, last_read_response);
          print();
          return;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Provide a method to reset the CSR
  ////////////////////////////////////////////////////////////////////////////////
  reset(reset_type : csco_ht_reset_type) is empty;
  ////////////////////////////////////////////////////////////////////////////////
  // RdSized and WrSized configuration accesses of greater than one doubleword are not supported.
  // For the address just use the doubleword offset
  ////////////////////////////////////////////////////////////////////////////////
  // Read the packed data and fill in bits that are not included in the read mask with 0's.
  ////////////////////////////////////////////////////////////////////////////////
  read_local(byte_mask : list of bit):list of bit is {
    // Perform a DUT check to ensure that the method is getting called correctly.
    check that (byte_mask.size() == data.size()/8) else
    dut_error("byte mask size != data size/8\n");

    // Setup result so that it is some random list of bits.
    gen temp_data keeping {
      .size() == data.size();
    };
    // This will not work due to a bug in 3.3.2 with generation and inheritance.
    //     gen result keeping {
    //       .size() == data.size();
    //     };
    for i from 0 to byte_mask.size()-1 do {
      if(byte_mask[i] == 1) {
        //        result[(i+1)*8-1:(i)*8] = data[(i+1)*8-1:(i)*8];
        temp_data[(i+1)*8-1:(i)*8] = data[(i+1)*8-1:(i)*8];
      };
    };
    result = temp_data;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Write the data into the packed data and then push it out to the expanded fields.
  ////////////////////////////////////////////////////////////////////////////////
  write_local(byte_mask : list of bit, new_data : list of bit) is {
    // Perform a DUT check to ensure that the method is getting called correctly.
    check that (byte_mask.size() == me.data.size()/8) else
    dut_error("byte mask size != data size/8\n");

    for i from 0 to byte_mask.size()-1 do {
      if(byte_mask[i] == 1) {
        for j from 0 to 7 do {
          if(write_mask[i*8+j:i*8+j] == 1) {
            me.data[i*8+j] = new_data[i*8+j];
          };
        };
      };
    };
    for i from 0 to byte_mask.size()-1 do {
      if(byte_mask[i] == 1) {
        for j from 0 to 7 do {
          if(set_mask[i*8+j:i*8+j] == 1) {
            me.data[i*8+j] = 1;
          };
        };
      };
    };
    update_fields();
    update_expected_data();
  };

  ////////////////////////////////////////////////////////////////////////////////
  // Write the data into the packed data and then push it out to the expanded fields.
  // This shadow write will ignore the write mask field for the register. The reason
  // for this is that with the read_ht method, you will want to go out, read the contents
  // of the register and update a local "shadow" copy.  To do this the write_mask
  // must be ignored.  Additionally, the expected data will not be updated.
  // Note: This is a back door write, please use it carefully.
  ////////////////////////////////////////////////////////////////////////////////
  write_shadow(byte_mask : list of bit, new_data : list of bit) is {
    // Perform a DUT check to ensure that the method is getting called correctly.
    check that (byte_mask.size() == me.data.size()/8) else
    dut_error("byte mask size != data size/8 \n");

    for i from 0 to byte_mask.size()-1 do {
      if(byte_mask[i] == 1) {
        for j from 0 to 7 do {
          me.data[i*8+j] = new_data[i*8+j];
        };
      };
    };
    update_fields();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // this method writes a '1' to every bit in the register based on the functional
  // write mask (func_wr_mask) and then reads the register to check the expected
  // write value. if the bit is not set in the func_wr_mask, that bit
  // will not be written.
  ////////////////////////////////////////////////////////////////////////////////
  walk_ones_ht(channel : csco_ht_channel_interface):csco_ht_error_code @parent_device.csco_ht_device_fundamental_clock is {
    var zero_data: list of bit;
    var wr_bit: uint(bits: 32);
    var ret_value : csco_ht_error_code;
    var read_ret_value : csco_ht_error_code;
    var wr_mask: uint(bits: 32) = func_write_mask[:]; // assign func write mask to a uint

    // Go through all the bits, writing ones to the bits you can write to in the reg.
    // shift 1'b1 by the bit number that will be written, then '&' it with the func_write_mask
    // to get the value to be written
    for {wr_bit = 1; wr_bit != 0; wr_bit <<= 1} {
      if((wr_bit & wr_mask) != 0) {
        write_local(%{4'hF}, %{(wr_bit & wr_mask)});
        ret_value = write_ht(channel);

        wait [2] * cycle;
        read_ret_value = read_ht(channel);
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // In order to make pre-determined register access more readable, the reg_name
  // field is appended to the packet packet_string so that it will be printed when
  // the packet is transmitted.
  ////////////////////////////////////////////////////////////////////////////////
  append_csco_ht_reg_name(packet : csco_ht_packet) is {
    if(reg_name != "") {
      packet.packet_string = appendf("%s : Reg Name=%s", packet.packet_string, reg_name);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // direct_read_ht - empty method that's extendable by the user to allow
  // a backdoor way of reading a register's address location
  ////////////////////////////////////////////////////////////////////////////////
  direct_read_ht() is empty;
  ////////////////////////////////////////////////////////////////////////////////
  // direct_write_ht - empty method that's extendable by the user to allow
  // a backdoor way of writing a register's address location
  ////////////////////////////////////////////////////////////////////////////////
  direct_write_ht() is empty;
  ////////////////////////////////////////////////////////////////////////////////
  // read_ht is a method that will take the register data and associate meta information
  // and create a read transaction on the HT bus.  It will then schedule the packet for
  // transmission, wait for and collect the response.  Finally, it will compare the returned
  // data against a copy of expected data to see if the target register performed properly.
  // Error flags are returned if the response was NXA or ERROR.
  ////////////////////////////////////////////////////////////////////////////////
  read_ht(channel : csco_ht_channel_interface):csco_ht_error_code @parent_device.csco_ht_device_clock is {
    var token : uint (bits:10);
    var endian_type: csco_ht_endian_type = parent_device.decode_endian_type(Addr);
    var srctag : uint (bits:5) = parent_device.gen_tx_SrcTag(0);

    if(indirect_access) {
      gen me.last_read_request keeping {

//KMM        .command == Read;
//KMM        .endian_type == endian_type;
//KMM        .passing_rule == me.passing_rule;
//KMM        .Read'command'response_passing_rule == me.response_passing_rule;
//KMM        .Read'command'data_type == me.read_data_type;
//KMM        .Read'command'BW_latency_req == me.BW_latency_req;
//KMM        .Read'command'access_type == me.mem_access_type;
//KMM        .Read'command'SeqID == me.SeqID;
//KMM        .Read'command'UnitID == me.UnitID;
//KMM        .Read'command'SrcTag_Rsv == srctag;
//KMM        .Read'command'Compat == me.Compat;
//KMM        .Read'command'Mask_Count == read_mask_count;
//KMM        .Read'command'Addr == me.Addr;

        .endian_type == endian_type;
        .passing_rule == me.passing_rule;
        it is a Read'command csco_ht_packet(tmp) and
        tmp.response_passing_rule == me.response_passing_rule and
        tmp.data_type == me.read_data_type and
        tmp.BW_latency_req == me.BW_latency_req and
        tmp.access_type == me.mem_access_type and
        tmp.SeqID == me.SeqID and
        tmp.UnitID == me.UnitID and
        tmp.SrcTag_Rsv == srctag and
        tmp.Compat == me.Compat and
        tmp.Mask_Count == read_mask_count and
        tmp.Addr == me.Addr;
      };

      // Provide an empty method to allow the user to append text to the end of the csco_ht_packet's packet_string.
      append_csco_ht_reg_name(last_read_request);

      // Schedule the packet for transmission
      // call ht scoreboard if scoreboard is not disabled, otherwise call the external scoreboard hook
      if(parent_device.scoreboard.score_level != csco_ht_no_scoring) {
        parent_device.scoreboard.post(last_read_request);
        parent_device.schedule_packet(channel, last_read_request);
      } else {
        parent_device.score_tx_pkt(last_read_request);
      };

      token = last_read_request.generate_token();

      // Mark the transaction so that it will not be collected by the csco_ht_device accept_packet method.
      parent_device.scoreboard.mark_config_transaction(token);

      // Wait until the transaction has completed
      wait true(parent_device.scoreboard.transaction_complete(token));

      // Take returned data and update register
      last_read_response = parent_device.scoreboard.get_completed_response(token);
      // If there was an error of any type, don't use the data.
      if(last_read_response.NXA() == 1) {
        result = READ_NXA;
      } else if(last_read_response.ErrorFlag() == 1) {
        result = READ_ERROR;
      } else {
        result = NO_ERROR;
        last_read_data = last_read_response.retrieve_data_packet();
        // Determine if it is a byte or doubleword transaction
        var temp_mask : list of bit;
        var temp_data : list of bit;
        if(last_read_request.DataType() == Byte_Read) {
          // It was a Byte read, so get the read mask
          temp_mask = %{read_mask_count[reg_size.as_a(uint)/8-1+byte_offset:0+byte_offset]};

//KMM     temp_data = %{last_read_data.Byte_Read'command'Data[reg_size.as_a(uint)+(byte_offset*8)-1:0+(byte_offset*8)]};
          temp_data = %{last_read_data.as_a(Byte_Read'command csco_ht_packet).Data[reg_size.as_a(uint)+(byte_offset*8)-1:0+(byte_offset*8)]};
        }else {
          // It was a doubleword read.
          temp_mask = %{4'hF};
//KMM          temp_data = %{last_read_data.DoubleWord_Read'command'Data[reg_size.as_a(uint)+(byte_offset*8)-1:0+(byte_offset*8)]};
          temp_data = %{last_read_data.as_a(DoubleWord_Read'command csco_ht_packet).Data[reg_size.as_a(uint)+(byte_offset*8)-1:0+(byte_offset*8)]};

        };
        write_shadow(temp_mask, temp_data);

        if(read_mask_compare) {
          compare_rd_msk_data();
        } else {
          compare_data();
        };
      };
      // Retire SrcTag
      parent_device.control_block.retire_SrcTag(token[9:5], token[4:0]);
      // Retire transaction
      parent_device.scoreboard.retire_transaction(token);
    } else {
      // call method to directly read the register address location
      direct_read_ht();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Provide a Non-Blocking Read
  ////////////////////////////////////////////////////////////////////////////////
  nonblocking_read_ht(channel : csco_ht_channel_interface) @parent_device.csco_ht_device_clock is {
    var token : uint (bits:10);
    var endian_type: csco_ht_endian_type = parent_device.decode_endian_type(Addr);
    var srctag : uint (bits:5) = parent_device.gen_tx_SrcTag(0);

    gen me.last_read_request keeping {
//KMM      .command == Read;
//KMM      .endian_type == endian_type;
//KMM      .passing_rule == me.passing_rule;
//KMM      .Read'command'response_passing_rule == me.response_passing_rule;
//KMM      .Read'command'data_type == me.read_data_type;
//KMM      .Read'command'BW_latency_req == me.BW_latency_req;
//KMM      .Read'command'access_type == me.mem_access_type;
//KMM      .Read'command'SeqID == me.SeqID;
//KMM      .Read'command'UnitID == me.UnitID;
//KMM      .Read'command'SrcTag_Rsv == srctag;
//KMM      .Read'command'Compat == me.Compat;
//KMM      .Read'command'Mask_Count == read_mask_count;
//KMM      .Read'command'Addr == me.Addr;

      .endian_type == endian_type;
      .passing_rule == me.passing_rule;
      it is a Read'command csco_ht_packet(tmp) and
      tmp.response_passing_rule == me.response_passing_rule and
      tmp.data_type == me.read_data_type
      and tmp.BW_latency_req == me.BW_latency_req
      and tmp.access_type == me.mem_access_type
      and tmp.SeqID == me.SeqID
      and tmp.UnitID == me.UnitID
      and tmp.SrcTag_Rsv == srctag
      and tmp.Compat == me.Compat
      and tmp.Mask_Count == read_mask_count
      and tmp.Addr == me.Addr;
    };

    // Provide an empty method to allow the user to append text to the end of the csco_ht_packet's packet_string.
    append_csco_ht_reg_name(last_read_request);

    // Schedule the packet for transmission
    // call ht scoreboard if scoreboard is not disabled, otherwise call the external scoreboard hook
    if(parent_device.scoreboard.score_level != csco_ht_no_scoring) {
      parent_device.scoreboard.post(last_read_request);
      parent_device.schedule_packet(channel, last_read_request);
    } else {
      parent_device.score_tx_pkt(last_read_request);
    };

    token = last_read_request.generate_token();

    // Mark the transaction so the csco_ht_device accept_packet method with do a
    // compare with the reg data
    parent_device.scoreboard.mark_reg_transaction(token);

    // add the register to the list of register read requests
    parent_device.read_reg_req.add(me);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Check a Non-Blocking Read
  // method called by csco_ht_device accept_packet to compare the reg data
  ////////////////////////////////////////////////////////////////////////////////
  check_nonblocking_read_ht(token: uint(bits: 10)) is {
    // Take returned data and update register
    last_read_response = parent_device.scoreboard.get_completed_response(token);

    // If there was an error of any type, don't use the data.
    if((last_read_response.NXA() != 1) || (last_read_response.ErrorFlag() != 1)) {
        last_read_data = last_read_response.retrieve_data_packet();
      // Determine if it is a byte or doubleword transaction
      var temp_mask : list of bit;
      var temp_data : list of bit;
      if(last_read_request.DataType() == Byte_Read) {
        // It was a Byte read, so get the read mask
        temp_mask = %{read_mask_count[reg_size.as_a(uint)/8-1+byte_offset:0+byte_offset]};
//KMM        temp_data = %{last_read_data.Byte_Read'command'Data[reg_size.as_a(uint)+(byte_offset*8)-1:0+(byte_offset*8)]};
        temp_data = %{last_read_data.as_a(Byte_Read'command csco_ht_packet).Data[reg_size.as_a(uint)+(byte_offset*8)-1:0+(byte_offset*8)]};
      }else {
        // It was a doubleword read.
        temp_mask = %{4'hF};
//KMM        temp_data = %{last_read_data.DoubleWord_Read'command'Data[reg_size.as_a(uint)+(byte_offset*8)-1:0+(byte_offset*8)]};
        temp_data = %{last_read_data.as_a(DoubleWord_Read'command csco_ht_packet).Data[reg_size.as_a(uint)+(byte_offset*8)-1:0+(byte_offset*8)]};
      };
      write_shadow(temp_mask, temp_data);

      if(read_mask_compare) {
        compare_rd_msk_data();
      } else {
        compare_data();
      };
    };
  };

  ////////////////////////////////////////////////////////////////////////////////
  // write_ht is a method that will take the register data and associate meta information
  // and create a write transaction on the HT bus.  It will then schedule the packet for
  // transmission and if it was a NonPosted transaction, it will wait for and collect the response.
  // Error flags are returned if the response was NXA or ERROR.
  ////////////////////////////////////////////////////////////////////////////////
  write_ht(channel : csco_ht_channel_interface):csco_ht_error_code @parent_device.csco_ht_device_fundamental_clock is {
    var token : uint (bits:10);
    var srctag : uint (bits:5);
    var endian_type: csco_ht_endian_type = parent_device.decode_endian_type(Addr);

    if(indirect_access) {
      // Since the typical behavior of the register model is to update the physical fields and
      // then write the register to the network, call this method to ensure that the packed data
      // reflects the physical fields.
      update_packed_data();

      if(write_vChan == NonPosted) {
        // Only generate a SrcTag if the transaction will be non posted.
        srctag = parent_device.gen_tx_SrcTag(0);
      };

      // HT transactions have to be an even number of 32 bit blocks.
      if(reg_size.as_a(uint) % 32 != 0) {
        gen last_write_data keeping {
          // Generate a list of data that is a multiple of 4 bytes, and round up.
          .size() == (reg_size.as_a(uint)/32 + 1) * 32;
        };
      }else {
        gen last_write_data keeping {
          .size() == reg_size.as_a(uint);
        };
      };

      // Copy the data into the temp field, starting after said byte offset.
      last_write_data[(byte_offset*8)+reg_size.as_a(uint)-1:byte_offset*8] = data[reg_size.as_a(uint)-1:0];

      gen last_write_request keeping {
//KMM        .command == Write;
//KMM        .endian_type == endian_type;
//KMM        .vChan == me.write_vChan;
//KMM        .passing_rule == me.passing_rule;
//KMM        .Write'command'data_type == me.write_data_type;
//KMM        .Write'command'BW_latency_req == me.BW_latency_req;
//KMM        .Write'command'access_type == me.mem_access_type;
//KMM        .Write'command'SeqID == me.SeqID;
//KMM        .Write'command'UnitID == me.UnitID;
//KMM        .vChan == NonPosted => .Write'command'SrcTag_Rsv == srctag;
//KMM        .Write'command'Compat == me.Compat;
//KMM        .Write'command'data_type == Byte => .Write'command'Mask_Count == last_write_data.size()/32;
//KMM        .Write'command'data_type == DoubleWord => .Write'command'Mask_Count == value(last_write_data.size()/32-1);
//KMM        .Write'command'Addr == me.Addr;
//KMM        .Write'command'data_type == Byte => .Write'command'assoc_data.Byte_Write'command'temp_mask == write_byte_mask[3:0];
//KMM        .Write'command'data_type == Byte => .Write'command'assoc_data.Byte_Write'command'Mask == write_byte_mask;
//KMM        .Write'command'data_type == Byte => .Write'command'assoc_data.Byte_Write'command'Data == last_write_data;
//KMM        .Write'command'data_type == DoubleWord => .Write'command'assoc_data.DoubleWord_Write'command'Data == last_write_data;

        .endian_type == endian_type;
        .vChan == me.write_vChan;
        .passing_rule == me.passing_rule;
        it is a Write'command csco_ht_packet(tmp) and
        tmp.data_type == me.write_data_type and
        tmp.BW_latency_req == me.BW_latency_req and
        tmp.access_type == me.mem_access_type and
        tmp.SeqID == me.SeqID and
        tmp.UnitID == me.UnitID ;

        it is a Write'command csco_ht_packet(tmp) and .vChan == NonPosted => tmp.SrcTag_Rsv == srctag;

        it is a Write'command csco_ht_packet(tmp) and tmp.Compat == me.Compat;

        it is a Write'command csco_ht_packet(tmp) and tmp.data_type == Byte => tmp.Mask_Count == last_write_data.size()/32 ;
        it is a Write'command csco_ht_packet(tmp) and tmp.data_type == DoubleWord => tmp.Mask_Count == value(last_write_data.size()/32-1);
        it is a Write'command csco_ht_packet(tmp) and tmp.Addr == me.Addr;

        it is a Write'command csco_ht_packet(tmp) and
        tmp.assoc_data is a Byte_Write'command csco_ht_packet(tmp_assoc_data) and
        tmp.data_type == Byte => tmp_assoc_data.temp_mask == write_byte_mask[3:0] and
        tmp_assoc_data.Mask == write_byte_mask and
        tmp_assoc_data.Data == last_write_data;

        it is a Write'command csco_ht_packet(tmp) and
        tmp.assoc_data is a DoubleWord_Write'command csco_ht_packet(tmp_assoc_data) and
        tmp.data_type == DoubleWord => tmp_assoc_data.Data == last_write_data;




      };

      // Provide an empty method to allow the user to append text to the end of the csco_ht_packet's packet_string.
      append_csco_ht_reg_name(last_write_request);

      // Schedule the packet for transmission
      // call ht scoreboard if scoreboard is not disabled, otherwise call the external scoreboard hook
      if(parent_device.scoreboard.score_level != csco_ht_no_scoring) {
        parent_device.scoreboard.post(last_write_request);
        parent_device.schedule_packet(channel, last_write_request);
      } else {
        parent_device.score_tx_pkt(last_write_request);
      };

      // If the write is NonPosted, wait for the response and clean up, otherwise, just exit.
      if(write_vChan == NonPosted) {
        token = last_write_request.generate_token();
        // Mark the transaction so that it will not be collected by the csco_ht_device accept_packet method.
        parent_device.scoreboard.mark_config_transaction(token);

        // Wait until the transaction has completed
        wait true(parent_device.scoreboard.transaction_complete(token));

        // Take returned data and check for errors.
        last_write_response = parent_device.scoreboard.get_completed_response(token);
        // If there was an error of any type, don't use the data.
        if(last_write_response.NXA() == 1) {
          result = READ_NXA;
        }else if(last_write_response.ErrorFlag() == 1) {
          result = READ_ERROR;
        }else {
          result = NO_ERROR;
        };
        // Retire SrcTag
        parent_device.control_block.retire_SrcTag(token[9:5], token[4:0]);
        // Retire transaction
        parent_device.scoreboard.retire_transaction(token);
      }else {
        result = NO_ERROR;
      };
    } else {
      // call method to directly write the register address location
      direct_write_ht();
      result = NO_ERROR;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Define empty print method to be extended later
  ////////////////////////////////////////////////////////////////////////////////
  print() is { };
  ////////////////////////////////////////////////////////////////////////////////
};

////////////////////////////////////////////////////////////////////////////////
// csco_ht_csr_reg's are a child of the standard csco_ht_basic_reg.  They provide specific hooks and
// definitions that are specific to registers in the HT CSR space.
// csco_ht_csr_reg's must be at least 1 byte, and at most 4 bytes to allow proper configuration
// cycle access.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_csr_reg like csco_ht_basic_reg {
  // Offset within the appropriate header.
  doubleword_offset : byte;
  keep doubleword_offset[1:0] == 2'b00; // Ensure that this offset refers to a 4 byte block.
  persistance_mask : list of bit;
  keep persistance_mask.size() == reg_size.as_a(uint);

  target_UnitID : uint (bits:5);
  keep soft target_UnitID == 0;
  target_function : uint (bits:3);
  keep soft target_function == 0;
  // A structure used to maintain the default values of the register for use during a reset.
  !reset_data : list of bit;
  post_generate() is also {
    reset_data = data.copy()
  };

  csr_access_type : csco_ht_csr_access_type;
  keep csr_access_type == Type0;

  // List of parameters to constrain how CSR access occur.
  keep write_vChan == NonPosted;
  keep passing_rule == NoPassPW;
  keep response_passing_rule == NoPassPW;
  keep BW_latency_req == Normal;
  keep mem_access_type == NonCoherent;
  keep SeqID == 0; // Always used non-sequenced requests.
  keep UnitID == 0; // Only the host bridge can send CSR requests.
  keep Compat == 0; // Don't set the compat bit;

  // Each time an HT access is performed, regenerate the Addr field so that any changes to UnitID (etc.) are picked up.
  read_ht(channel : csco_ht_channel_interface):csco_ht_error_code @parent_device.csco_ht_device_clock is first {
    gen Addr keeping {
      it == %{csr_access_type.as_a(uint), 8'h00, target_UnitID, target_function, doubleword_offset};
      it[1:0] == 2'b0;
    };
  };
  nonblocking_read_ht(channel : csco_ht_channel_interface) @parent_device.csco_ht_device_clock is first {
    gen Addr keeping {
      it == %{csr_access_type.as_a(uint), 8'h00, target_UnitID, target_function, doubleword_offset};
      it[1:0] == 2'b0;
    };
  };
  write_ht(channel : csco_ht_channel_interface):csco_ht_error_code @parent_device.csco_ht_device_fundamental_clock is first {
    gen Addr keeping {
      it == %{csr_access_type.as_a(uint), 8'h00, target_UnitID, target_function, doubleword_offset};
      it[1:0] == 2'b0;
    };
  };
  update_fields () is also {
    parent_device.control_block.update_extracted_csr_fields();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Reset the register and push the reset_data into the data fields.  Check the persistance mask and reset type.
  ////////////////////////////////////////////////////////////////////////////////
  reset(reset_type : csco_ht_reset_type) is {
    for each in data {
      // If the bit in the persistance mask is 0 then the data can be updated on a warm reset, otherwise only a cold reset will update the data.
      if(persistance_mask[index] == 0 || reset_type == Cold) {
        data[index] = reset_data[index];
      };
    };
    check that (data.size() == reg_size.as_a(uint)) else dut_error("Register size constraints exceeded.");
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// csco_ht_capability_header will actually instantiate and provide access to all of the
// csco_ht_capability_registers.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_capability_header {
  // Customize the Capability register based on the device function.
  device_function : csco_ht_device_encoding;
  keep soft device_function == get_enclosing_unit(csco_ht_device).device_function;
  block_offset : byte;
  keep soft block_offset == CSCO_HT_CAPABILITIES_OFFSET;

  !regs : list (key:reg_type) of csco_ht_capability_reg;

  temp_regs : list of csco_ht_capability_reg;
  keep for each in temp_regs {
    // Set up the offset for the capability header.
    it.block_offset == block_offset;

    device_function == Host_Bridge => it.interface_type == Secondary;
    device_function == Tunnel_IO_Device or device_function == EOC_IO_Device => it.interface_type == Primary;

    // If the capability header is for an EOC device, make sure that the second control register
    // is hardwired, and the Link Failure bit is set.
    device_function == EOC_IO_Device && it.reg_type == Link_Control1 => it.write_mask == %{16'h0000};

    // If the capability header is for an EOC device, make sure that the second config register
    // is hardwired.
    device_function == EOC_IO_Device && it.reg_type == Link_Config1 => it.write_mask == %{16'h0000};

    // If the capability header is for an EOC device, make sure that the Freq1 register is hardwired.
    device_function == EOC_IO_Device && it.reg_type == Link_Freq => it.write_mask == %{8'h0F};

//KMM    device_function == EOC_IO_Device && it.reg_type == Link_Control1 => it.Link_Control1'reg_type'Link_Failure == 1'b1;

    it is a Link_Control1'reg_type csco_ht_capability_reg (tmp) and device_function == EOC_IO_Device  => tmp.Link_Failure == 1'b1;

  };
  keep temp_regs.is_all_iterations(.reg_type);
  ////////////////////////////////////////////////////////////////////////////////
  post_generate() is also {
    regs.add(temp_regs);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Provide read access to the registers.
  ////////////////////////////////////////////////////////////////////////////////
  read(addr : byte, mask : uint (bits:4)):list of bit is {
    case(addr) {
      8'h00: {
        result.add(regs.key(Capability_ID).read_local(%{mask[0:0]}));
        result.add(regs.key(Capabilities_Pointer).read_local(%{mask[1:1]}));
        result.add(regs.key(Command).read_local(%{mask[3:2]}));
      };
      8'h04: {
        result.add(regs.key(Link_Control0).read_local(%{mask[1:0]}));
        result.add(regs.key(Link_Config0).read_local(%{mask[3:2]}));
      };
      8'h08: {
        if(regs.key(Command).interface_type == Primary) {
          result.add(regs.key(Link_Control1).read_local(%{mask[1:0]}));
          result.add(regs.key(Link_Config1).read_local(%{mask[3:2]}));
        }else { // regs.key(Command).interface_type == Secondary
          result.add(regs.key(HT_Rev_ID).read_local(%{mask[0:0]}));
          result.add(regs.key(Link_Freq).read_local(%{mask[1:1]}));
          result.add(regs.key(Reserved).read_local(%{mask[3:2]}));
        };
      };
      8'h0C: {
        check that (regs.key(Command).interface_type == Primary) else dut_error("Illegal read to HT Capability Registers");
        result.add(regs.key(HT_Rev_ID).read_local(%{mask[0:0]}));
        result.add(regs.key(Link_Freq).read_local(%{mask[1:1]}));
        result.add(regs.key(Reserved).read_local(%{mask[3:2]}));
      };
    };
   check that (result.size() == 32) else
   dut_error("result size != 32\n");

  };
  ////////////////////////////////////////////////////////////////////////////////
  // Provide write access to the registers.
  ////////////////////////////////////////////////////////////////////////////////
  write(addr: byte, mask : uint (bits:4), data : uint) is {
    case(addr) {
      8'h00: {
        regs.key(Capability_ID).write_local(%{mask[0:0]}, %{data[7:0]});
        regs.key(Capabilities_Pointer).write_local(%{mask[1:1]}, %{data[15:8]});
        regs.key(Command).write_local(%{mask[3:2]}, %{data[31:16]});
//KMM        if(regs.key(Command).interface_type == Primary) {
//KMM          // When the Command Register is written, the current UnitID must be copied back to the target_UnitID.
//KMM          update_target_UnitID((regs.key(Command)).Primary'interface_type'Command'reg_type'Base_UnitID);
//KMM        };
        if(regs.key(Command) is a Primary'interface_type Command'reg_type csco_ht_capability_reg(tmp) ) {
          // When the Command Register is written, the current UnitID must be copied back to the target_UnitID.
          update_target_UnitID(tmp.Base_UnitID);
        };
      };
      8'h04: {
        regs.key(Link_Control0).write_local(%{mask[1:0]}, %{data[15:0]});
        regs.key(Link_Config0).write_local(%{mask[3:2]}, %{data[31:16]});
      };
      8'h08: {
        if(regs.key(Command).interface_type == Primary) {
          regs.key(Link_Control1).write_local(%{mask[1:0]}, %{data[15:0]});
          regs.key(Link_Config1).write_local(%{mask[3:2]}, %{data[31:16]});
        }else { // regs.key(Command).interface_type == Secondary
          regs.key(HT_Rev_ID).write_local(%{mask[0:0]}, %{data[7:0]});
          regs.key(Link_Freq).write_local(%{mask[1:1]}, %{data[15:8]});
          regs.key(Reserved).write_local(%{mask[3:2]}, %{data[31:16]});
        };
      };
      8'h0C: {
        check that (regs.key(Command).interface_type == Primary) else dut_error("Illegal write to HT Capability Registers");
        regs.key(HT_Rev_ID).write_local(%{mask[0:0]}, %{data[7:0]});
        regs.key(Link_Freq).write_local(%{mask[1:1]}, %{data[15:8]});
        regs.key(Reserved).write_local(%{mask[3:2]}, %{data[31:16]});
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  update_target_UnitID(new_UnitID : uint (bits:5)) is {
    for each in regs do {
      it.target_UnitID = new_UnitID;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return the current link / direction to the master host.
  ////////////////////////////////////////////////////////////////////////////////
  master_host_link():bit is {
    if(device_function != Host_Bridge) {
//KMM result = regs.key(Command).Primary'interface_type'Command'reg_type'Master_Host;
      result = regs.key(Command).as_a(Primary'interface_type Command'reg_type csco_ht_capability_reg).Master_Host;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  read_UnitID():list of uint (bits:5) is {
    var temp_list : list of uint (bits:5);
    if(device_function == Host_Bridge) {
      temp_list.add(0);
    }else {
      gen temp_list keeping {
//KMM        it.size() == regs.key(Command).Primary'interface_type'Command'reg_type'Unit_Count;
        it.size() == regs.key(Command).as_a(Primary'interface_type Command'reg_type csco_ht_capability_reg).Unit_Count;
      };
      for each in temp_list {
//KMM        temp_list[index] =  regs.key(Command).Primary'interface_type'Command'reg_type'Base_UnitID + index;
        temp_list[index] =  regs.key(Command).as_a(Primary'interface_type Command'reg_type csco_ht_capability_reg).Base_UnitID + index;
      };
    };
    result = temp_list;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return a list of the current transmitter widths for all links.
  ////////////////////////////////////////////////////////////////////////////////
  decode_link_width(link_width : uint (bits:3)):csco_ht_link_size is {
    case(link_width) {
      3'b000: {
        return Eight_Bits;
      };
      3'b001: {
        return Sixteen_Bits;
      };
      3'b011: {
        return ThirtyTwo_Bits;
      };
      default: {
        dut_error("Illegal link size encoding.");
      };
    };
  };
  read_tx_link_widths():list of csco_ht_link_size is {
    var temp_list : list of csco_ht_link_size;
    if(regs.key_exists(Link_Config0)) {
     //KMM  temp_list.add(decode_link_width(regs.key(Link_Config0).Link_Config0'reg_type'LinkWidthOut));
      temp_list.add(decode_link_width(regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).LinkWidthOut));
    };
    if(regs.key_exists(Link_Config1)) {
//KMM      temp_list.add(decode_link_width(regs.key(Link_Config1).Link_Config1'reg_type'LinkWidthOut));
      temp_list.add(decode_link_width(regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).LinkWidthOut));
    };
    result = temp_list;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return a list of the maximum transmitter widths for all links.
  ////////////////////////////////////////////////////////////////////////////////
  read_max_tx_link_widths():list of csco_ht_link_size is {
    var temp_list : list of csco_ht_link_size;
    if(regs.key_exists(Link_Config0)) {
      //KMM temp_list.add(decode_link_width(regs.key(Link_Config0).Link_Config0'reg_type'MaxLinkWidthOut));
      temp_list.add(decode_link_width(regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).MaxLinkWidthOut));
    };
    if(regs.key_exists(Link_Config1)) {
      //KMM temp_list.add(decode_link_width(regs.key(Link_Config1).Link_Config1'reg_type'MaxLinkWidthOut));
      temp_list.add(decode_link_width(regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).MaxLinkWidthOut));
    };
    result = temp_list;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return a list of the current receiver widths for all links.
  ////////////////////////////////////////////////////////////////////////////////
  read_rx_link_widths():list of csco_ht_link_size is {
    var temp_list : list of csco_ht_link_size;
    if(regs.key_exists(Link_Config0)) {
//KMM      temp_list.add(decode_link_width(regs.key(Link_Config0).Link_Config0'reg_type'LinkWidthIn));
      temp_list.add(decode_link_width(regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).LinkWidthIn));
    };
    if(regs.key_exists(Link_Config1)) {
//KMM      temp_list.add(decode_link_width(regs.key(Link_Config1).Link_Config1'reg_type'LinkWidthIn));
      temp_list.add(decode_link_width(regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).LinkWidthIn));
    };
    result = temp_list;
  };
  ////////////////////////////////////////////////////////////////////////////////
  decode_clock_freq(freq : uint (bits:4)):csco_ht_clock_multiplier is {
    if(sys.csco_ht_use_direct_csco_ht_clock == TRUE) {
      // If the inbound clock is a direct drive, override the CSR's value and put
      // in a standard multiple.
      return csco_ht_DirectDrive;
    }else {
      case(freq) {
        4'b0000: {
          return csco_ht_200MHz;
        };
        4'b0001: {
          return csco_ht_300MHz;
        };
        4'b0010: {
          return csco_ht_400MHz;
        };
        4'b0011: {
          return csco_ht_500MHz;
        };
        4'b0100: {
          return csco_ht_600MHz;
        };
        4'b0101: {
          return csco_ht_800MHz;
        };
        4'b0110: {
          return csco_ht_1000MHz;
        };
        default: {
          dut_error("Reserved link freq encoding.");
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return a list of the current tx clock frequencies on all links.
  ////////////////////////////////////////////////////////////////////////////////
  read_tx_clock_freq():list of csco_ht_clock_multiplier is {
    var temp_list : list of csco_ht_clock_multiplier;
//KMM    temp_list.add(decode_clock_freq(regs.key(Link_Freq).Link_Freq'reg_type'Freq0));
    temp_list.add(decode_clock_freq(regs.key(Link_Freq).as_a(Link_Freq'reg_type csco_ht_capability_reg ).Freq0));
    if(regs.key(Link_Freq).interface_type == Primary) {
//KMM      temp_list.add(decode_clock_freq(regs.key(Link_Freq).Primary'interface_type'Link_Freq'reg_type'Freq1));
      temp_list.add(decode_clock_freq(regs.key(Link_Freq).as_a(Primary'interface_type Link_Freq'reg_type csco_ht_capability_reg).Freq1));
    };
    result = temp_list;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return a list of bits indicating the state of the Initialization Complete bit for all links.
  ////////////////////////////////////////////////////////////////////////////////
  links_initialized():list of bit is {
    if(regs.key_exists(Link_Control0)) {
//KMM      result.add(regs.key(Link_Control0).Link_Control0'reg_type'Initialization_Complete);
      result.add(regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).Initialization_Complete);
    };
    if(regs.key_exists(Link_Control1)) {
//KMM      result.add(regs.key(Link_Control1).Link_Control1'reg_type'Initialization_Complete);
      result.add(regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).Initialization_Complete);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return a list of bits indicating the state of the CRC_Error bits for all links.
  ////////////////////////////////////////////////////////////////////////////////
  links_crc_errors():list of uint (bits:4) is {
    if(regs.key_exists(Link_Control0)) {
      result.add(regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).CRC_Error);
    };
    if(regs.key_exists(Link_Control1)) {
      result.add(regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).CRC_Error);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return a list of bits indicating the state of the End Of Chain bit for all links.
  ////////////////////////////////////////////////////////////////////////////////
  links_eoc_setting():list of bit is {
    if(regs.key_exists(Link_Control0)) {
//KMM      result.add(regs.key(Link_Control0).Link_Control0'reg_type'End_of_Chain);
      result.add(regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).End_of_Chain);
    };
    if(regs.key_exists(Link_Control1)) {
//KMM      result.add(regs.key(Link_Control1).Link_Control1'reg_type'End_of_Chain);
      result.add(regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).End_of_Chain);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return a list of bits indicating the state of the Transmitter Off bit for all links.
  ////////////////////////////////////////////////////////////////////////////////
  links_active():list of bit is {
    if(regs.key_exists(Link_Control0)) {
//KMM      result.add(regs.key(Link_Control0).Link_Control0'reg_type'ransmitter_Off);
      result.add(regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).Transmitter_Off);
    };
    if(regs.key_exists(Link_Control1)) {
//KMM      result.add(regs.key(Link_Control1).Link_Control1'reg_type'Transmitter_Off);
      result.add(regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).Transmitter_Off);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Reset the capability register.
  ////////////////////////////////////////////////////////////////////////////////
  reset(reset_type : csco_ht_reset_type) is {
    for each in regs {
      it.reset(reset_type);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// csco_ht_capability_reg defines the specific capability registers that exist in the
// ht csr space.  It provides write mask, initialization values, etc.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_capability_reg like csco_ht_csr_reg {
  interface_type : csco_ht_capability_type; // Denotes if this is a Primary or Secondary interface block.
  reg_type : csco_ht_capability_reg_type;
  keep interface_type == Secondary => reg_type != Link_Control1;
  keep interface_type == Secondary => reg_type != Link_Config1;
  block_offset : byte;
  keep soft block_offset == CSCO_HT_CAPABILITIES_OFFSET;
  // Assign the register type to a name field so that packets can be identified.
  keep reg_name == reg_type.as_a(string);

  // Set the initial value of the target address.  Remember that this has to change when the UnitID is changed.
  keep Addr == %{csr_access_type.as_a(uint), 8'h00, target_UnitID, target_function, doubleword_offset};
  ////////////////////////////////////////////////////////////////////////////////
  when Capability_ID'reg_type csco_ht_capability_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b00;
    keep doubleword_offset == block_offset + 8'h00;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %ID : byte;
    keep ID == 8'h08;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Capabilities_Pointer'reg_type csco_ht_capability_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b01;
    keep doubleword_offset == block_offset + 8'h00;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %capability_pointer : byte;
    keep soft capability_pointer == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Command'reg_type csco_ht_capability_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b10;
    keep doubleword_offset == block_offset + 8'h00;
    keep persistance_mask == %{16'h0000};
    keep interface_type == Primary => soft write_mask == %{16'h081F};
    keep interface_type == Secondary => soft write_mask == %{16'h0003};

    when Primary'interface_type csco_ht_capability_reg {
      %Base_UnitID : uint (bits:5);
      keep soft Base_UnitID == 5'b00000;
      // Unit Count indicates how many UnitID's there are in this device.  Typically 1
      %Unit_Count : uint (bits:5);
      keep soft Unit_Count == 5'b00001;
      %Master_Host : bit;
      keep soft Master_Host == 1'b0;
      %Default_Direction : bit;
      keep soft Default_Direction == 1'b0;
      %Reserved : bit;
      keep soft Reserved == 1'b0;

      write_local(byte_mask : list of bit, new_data : list of bit) is first {
      // Any time the command register is written,
        // the Master_Host bit is written with the link number that the write came from.
        Master_Host = parent_device.control_block.csr_inbound_link;
        // Since the physical field is being modified, that change has to be pushed out to the packed data.
        update_packed_data();
      };
    };
    when Secondary'interface_type csco_ht_capability_reg {
      %Warm_Reset : bit;
      keep soft Warm_Reset == 1'b1;
      %Double_Ended : bit;
      keep soft Double_Ended == 1'b0;
      %Reserved : uint (bits:11);
      keep soft Reserved == 11'h000;
    };
    %Capability_Type : uint (bits:3);
    keep interface_type == Primary => Capability_Type == 3'b000;
    keep interface_type == Secondary => Capability_Type == 3'b001;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Link_Control0'reg_type csco_ht_capability_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b00;
    keep doubleword_offset == block_offset + 8'h04;
    keep persistance_mask == %{16'h0F10};
    keep soft write_mask == %{16'hFFDE}; // Override the default write mask if a link is inactive
    %Rsv_1 : bit;
    keep Rsv_1 == 1'b0;
    %CRC_Flood_Enable : bit;
    keep soft CRC_Flood_Enable == 1'b0;
    %CRC_Start_Test : bit;
    keep soft CRC_Start_Test == 1'b0;
    %CRC_Force_Error : bit;
    keep soft CRC_Force_Error == 1'b0;
    %Link_Failure : bit;
    keep soft Link_Failure == 1'b0;
    %Initialization_Complete : bit;
    keep soft Initialization_Complete == 1'b0;
    %End_of_Chain : bit; // Cannot be cleared by software
    keep soft End_of_Chain == 1'b0;
    %Transmitter_Off : bit; // Cannot be cleared by software
    keep soft Transmitter_Off == 1'b0;
    %CRC_Error : uint (bits:4);
    keep soft CRC_Error == 4'h0;
    %Rsv_12_15 : uint (bits:4);
    keep Rsv_12_15 == 4'h0;

    write_local(byte_mask : list of bit, new_data : list of bit) is first {
      // Certain bits can be written to a 1 by software but cannot be written to a 0 by software.  i.e. they cannot be cleared.
      // End of Chain, Transmitter Off
      if(byte_mask[0] == 1'b1) { // The first byte of the register is being written
        // Potential write to End Of Chain bit
        if(End_of_Chain == 1'b1) {
          // Change incomming data bit to a 1 so that write has no effect.
          new_data[6] = 1;
        };
        // Potential write to Transmitter Off bit
        if(Transmitter_Off == 1'b1) {
          // Change incomming data bit to a 1 so that write has no effect.
          new_data[7:7] = 1;
        };
      };
      // CRC Error bits are cleared by writing a 1 to them.
      if(byte_mask[1] == 1'b1) { // The second byte of the register is being written
        // This is potentially a write to the CRC Error bits.
        for i from 0 to 3 do {
          if(CRC_Error[i:i] == 1'b1) { // CRC Error Bit i is set
            // Invert the value in the data to be written so that a 1 writes a 0 and a 0 writes a 1.
            new_data[8+i] = ~new_data[8+i];
          };
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Link_Control1'reg_type csco_ht_capability_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b00;
    keep doubleword_offset == block_offset + 8'h08;
    keep persistance_mask == %{16'h0F10};
    keep soft write_mask == %{16'hFFDE}; // Override the default write mask if a link is inactive
    %Rsv_1 : bit;
    keep Rsv_1 == 1'b0;
    %CRC_Flood_Enable : bit;
    keep soft CRC_Flood_Enable == 1'b0;
    %CRC_Start_Test : bit;
    keep soft CRC_Start_Test == 1'b0;
    %CRC_Force_Error : bit;
    keep soft CRC_Force_Error == 1'b0;
    %Link_Failure : bit;
    keep soft Link_Failure == 1'b0;
    %Initialization_Complete : bit;
    keep soft Initialization_Complete == 1'b0;
    %End_of_Chain : bit; // Cannot be cleared by software
    keep soft End_of_Chain == 1'b0;
    %Transmitter_Off : bit; // Cannot be cleared by software
    keep soft Transmitter_Off == 1'b0;
    %CRC_Error : uint (bits:4);
    keep soft CRC_Error == 4'h0;
    %Rsv_12_15 : uint (bits:4);
    keep Rsv_12_15 == 4'h0;

    write_local(byte_mask : list of bit, new_data : list of bit) is first {
      // Certain bits can be written to a 1 by software but cannot be written to a 0 by software.  i.e. they cannot be cleared.
      // End of Chain, Transmitter Off
      if(byte_mask[0] == 1'b1) { // The first byte of the register is being written
        // Potential write to End Of Chain bit
        if(End_of_Chain == 1'b1) {
          // Change incomming data bit to a 1 so that write has no effect.
          new_data[6] = 1;
        };
        // Potential write to Transmitter Off bit
        if(Transmitter_Off == 1'b1) {
          // Change incomming data bit to a 1 so that write has no effect.
          new_data[7:7] = 1;
        };
      };
      // CRC Error bits are cleared by writing a 1 to them.
      if(byte_mask[1] == 1'b1) { // The second byte of the register is being written
        // This is potentially a write to the CRC Error bits.
        for i from 0 to 3 do {
          if(CRC_Error[i:i] == 1'b1) { // CRC Error Bit i is set
            // Invert the value in the data to be written so that a 1 writes a 0 and a 0 writes a 1.
            new_data[8+i] = ~new_data[8+i];
          };
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Link_Config0'reg_type csco_ht_capability_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b10;
    keep doubleword_offset == block_offset + 8'h04;
    keep persistance_mask == %{16'hFF00};
    keep soft write_mask == %{16'hFF00};
    %MaxLinkWidthIn : uint (bits:3);
    keep soft MaxLinkWidthIn == 3'b000;
    %DwFcIn : bit;
    keep soft DwFcIn == 1'b0;
    %MaxLinkWidthOut : uint (bits:3);
    keep soft MaxLinkWidthOut == 3'b000;
    %DwFcOut :bit;
    keep soft DwFcOut == 1'b0;
    %LinkWidthIn : uint (bits:3);
    keep soft LinkWidthIn == 3'b000;
    %DwFcInEn : bit;
    keep soft DwFcInEn == 1'b0;
    %LinkWidthOut : uint (bits:3);
    keep soft LinkWidthOut == 3'b000;
    %DwFcOutEn : bit;
    keep soft DwFcOutEn == 1'b0;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Link_Config1'reg_type csco_ht_capability_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b10;
    keep doubleword_offset == block_offset + 8'h08;
    keep persistance_mask == %{16'hFF00};
    keep soft write_mask == %{16'hFF00};
    %MaxLinkWidthIn : uint (bits:3);
    keep soft MaxLinkWidthIn == 3'b000;
    %DwFcIn : bit;
    keep soft DwFcIn == 1'b0;
    %MaxLinkWidthOut : uint (bits:3);
    keep soft MaxLinkWidthOut == 3'b000;
    %DwFcOut :bit;
    keep soft DwFcOut == 1'b0;
    %LinkWidthIn : uint (bits:3);
    keep soft LinkWidthIn == 3'b000;
    %DwFcInEn : bit;
    keep soft DwFcInEn == 1'b0;
    %LinkWidthOut : uint (bits:3);
    keep soft LinkWidthOut == 3'b000;
    %DwFcOutEn : bit;
    keep soft DwFcOutEn == 1'b0;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when HT_Rev_ID'reg_type csco_ht_capability_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b00;
    keep interface_type == Primary => doubleword_offset == block_offset + 8'h0C;
    keep interface_type == Secondary => doubleword_offset == block_offset + 8'h08;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %MinorRev : uint (bits:5);
    keep soft MinorRev == 5'b00000;
    %MajorRev : uint (bits:3);
    keep soft MajorRev == 3'b001;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Link_Freq'reg_type csco_ht_capability_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b01;
    keep interface_type == Primary => doubleword_offset == block_offset + 8'h0C;
    keep interface_type == Secondary => doubleword_offset == block_offset + 8'h08;
    keep persistance_mask == %{8'hFF};
    keep interface_type == Primary => soft write_mask == %{8'hFF};
    keep interface_type == Secondary => soft write_mask == %{8'h0F};
    %Freq0 : uint (bits:4);
    keep soft Freq0 == 4'h0;
    when Primary'interface_type csco_ht_capability_reg {
      %Freq1 : uint (bits:4);
      keep soft Freq1 == 4'h0;
    };
    when Secondary'interface_type csco_ht_capability_reg {
      %Rsv : uint (bits:4);
      keep Rsv == 4'h0;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Reserved'reg_type csco_ht_capability_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b10;
    keep interface_type == Primary => doubleword_offset == block_offset + 8'h0C;
    keep interface_type == Secondary => doubleword_offset == block_offset + 8'h08;
    keep persistance_mask == %{16'h00};
    keep soft write_mask == %{16'h00};
    %Rsv : uint (bits:16);
    keep Rsv == 16'h0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// csco_ht_device_header actually instantiates and provides access to the csco_ht_device_reg's.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_device_header {
  control_block : csco_ht_device_control_block;
  keep control_block == get_enclosing_unit(csco_ht_device).control_block;
  !regs : list (key:reg_type) of csco_ht_device_reg;

  temp_regs : list of csco_ht_device_reg;
  keep for each in temp_regs {
    // 32 bit BARs
//KMM    it.reg_type == BAR0 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_32};
//KMM    it.reg_type == BAR0 => it.BAR0'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
    it is a BAR0'reg_type csco_ht_device_reg(tmp0) => tmp0.write_mask ==  %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_32} and tmp0.bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
//KMM   it.reg_type == BAR1 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_UNUSED};
//KMM    it.reg_type == BAR1 => it.BAR1'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
    it is a BAR1'reg_type csco_ht_device_reg(tmp1) => tmp1.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_UNUSED} and tmp1.bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
//KMM    it.reg_type == BAR2 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_32};
//KMM    it.reg_type == BAR2 => it.BAR2'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
    it is a BAR2'reg_type csco_ht_device_reg(tmp2) => tmp2.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_32} and tmp2.bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
 //KMM   it.reg_type == BAR3 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_UNUSED};
//KMM    it.reg_type == BAR3 => it.BAR3'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
    it is a BAR3'reg_type csco_ht_device_reg(tmp3) => tmp3.write_mask  == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_UNUSED} and tmp3.bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
//KMM  it.reg_type == BAR4 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_32};
//KMM    it.reg_type == BAR4 => it.BAR4'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
    it is a BAR4'reg_type csco_ht_device_reg(tmp4) => tmp4.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_32} and tmp4.bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
//KMM    it.reg_type == BAR5 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_UNUSED};
//KMM    it.reg_type == BAR5 => it.BAR5'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
    it is a BAR5'reg_type csco_ht_device_reg(tmp5) => tmp5.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_UNUSED} and  tmp5.bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 32 bit memory region.
    // 64 bit BARs
//     it.reg_type == BAR0 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_LOW};
//     it.reg_type == BAR0 => it.BAR0'reg_type'bar == %{32'h04}; // Bits[2:1] are set to 2'b10 indicating that this is a 64 bit memory region.
//     it.reg_type == BAR1 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HIGH};
//     it.reg_type == BAR1 => it.BAR1'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 64 bit memory region.
//     it.reg_type == BAR2 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_LOW};
//     it.reg_type == BAR2 => it.BAR2'reg_type'bar == %{32'h04}; // Bits[2:1] are set to 2'b10 indicating that this is a 64 bit memory region.
//     it.reg_type == BAR3 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HIGH};
//     it.reg_type == BAR3 => it.BAR3'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 64 bit memory region.
//     it.reg_type == BAR4 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_LOW};
//     it.reg_type == BAR4 => it.BAR4'reg_type'bar == %{32'h04}; // Bits[2:1] are set to 2'b10 indicating that this is a 64 bit memory region.
//     it.reg_type == BAR5 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HIGH};
//     it.reg_type == BAR5 => it.BAR5'reg_type'bar == %{32'h00}; // Bits[2:1] are set to 2'b00 indicating that this is a 64 bit memory region.
    // 40 bit BARs
//     it.reg_type == BAR0 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HT};
//     it.reg_type == BAR0 => it.BAR0'reg_type'bar == %{32'h06}; // Bits[2:1] are set to 2'b11 indicating that this is a 40 bit memory region.
//     it.reg_type == BAR1 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HT};
//     it.reg_type == BAR1 => it.BAR1'reg_type'bar == %{32'h06}; // Bits[2:1] are set to 2'b11 indicating that this is a 40 bit memory region.
//     it.reg_type == BAR2 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HT};
//     it.reg_type == BAR2 => it.BAR2'reg_type'bar == %{32'h06}; // Bits[2:1] are set to 2'b11 indicating that this is a 40 bit memory region.
//     it.reg_type == BAR3 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HT};
//     it.reg_type == BAR3 => it.BAR3'reg_type'bar == %{32'h06}; // Bits[2:1] are set to 2'b11 indicating that this is a 40 bit memory region.
//     it.reg_type == BAR4 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HT};
//     it.reg_type == BAR4 => it.BAR4'reg_type'bar == %{32'h06}; // Bits[2:1] are set to 2'b11 indicating that this is a 40 bit memory region.
//     it.reg_type == BAR5 => it.write_mask == %{CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HT};
//     it.reg_type == BAR5 => it.BAR5'reg_type'bar == %{32'h06}; // Bits[2:1] are set to 2'b11 indicating that this is a 40 bit memory region.
    // Indicates the location of the HT Capabilities Registers in the address map.
    it is a Capabilities_Pointer'reg_type csco_ht_device_reg(tmp6) => tmp6.Offset == CSCO_HT_CAPABILITIES_OFFSET;
  };
  keep temp_regs.is_all_iterations(.reg_type);
  !Base_Address_Register : list of csco_ht_device_reg;
  ////////////////////////////////////////////////////////////////////////////////
  post_generate() is also {
    regs.add(temp_regs);
    var bar_name : string;
    for i from 0 to 5 do {
      bar_name = appendf("BAR%d", i);
      Base_Address_Register.add(regs.key(bar_name.as_a(csco_ht_device_reg_type)));
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Provide read access to the registers.
  ////////////////////////////////////////////////////////////////////////////////
  read(addr : byte, mask : uint (bits:4)):list of bit is {
    case(addr) {
      8'h00: {
        result.add(regs.key(VendorID).read_local(%{mask[1:0]}));
        result.add(regs.key(DeviceID).read_local(%{mask[3:2]}));
      };
      8'h04: {
        result.add(regs.key(Command).read_local(%{mask[1:0]}));
        result.add(regs.key(Status).read_local(%{mask[3:2]}));
      };
      8'h08: {
        result.add(regs.key(RevisionID).read_local(%{mask[0:0]}));
        result.add(regs.key(ClassCode).read_local(%{mask[3:1]}));
      };
      8'h0C: {
        result.add(regs.key(Cache_Line_Size).read_local(%{mask[0:0]}));
        result.add(regs.key(Latency_Timer).read_local(%{mask[1:1]}));
        result.add(regs.key(Header_Type).read_local(%{mask[2:2]}));
        result.add(regs.key(BIST).read_local(%{mask[3:3]}));
      };
      8'h10: {
        result.add(regs.key(BAR0).read_local(%{mask[3:0]}));
      };
      8'h14: {
        result.add(regs.key(BAR1).read_local(%{mask[3:0]}));
      };
      8'h18: {
        result.add(regs.key(BAR2).read_local(%{mask[3:0]}));
      };
      8'h1C: {
        result.add(regs.key(BAR3).read_local(%{mask[3:0]}));
      };
      8'h20: {
        result.add(regs.key(BAR4).read_local(%{mask[3:0]}));
      };
      8'h24: {
        result.add(regs.key(BAR5).read_local(%{mask[3:0]}));
      };
      8'h28: {
        result.add(regs.key(Cardbus_CIS_Pointer).read_local(%{mask[3:0]}));
      };
      8'h2C: {
        result.add(regs.key(SubSystem_Vendor_ID).read_local(%{mask[1:0]}));
        result.add(regs.key(SubSystem_ID).read_local(%{mask[3:2]}));
      };
      8'h30: {
        result.add(regs.key(Expansion_ROM_Base_Address).read_local(%{mask[3:0]}));
      };
      8'h34: {
        result.add(regs.key(Capabilities_Pointer).read_local(%{mask[0:0]}));
        result.add(regs.key(Reserved_34h).read_local(%{mask[3:1]}));
      };
      8'h38: {
        result.add(regs.key(Reserved_38h).read_local(%{mask[0:0]}));
      };
      8'h3C: {
        result.add(regs.key(Interrupt_Line).read_local(%{mask[0:0]}));
        result.add(regs.key(Interrupt_Pin).read_local(%{mask[1:1]}));
        result.add(regs.key(Min_Gnt).read_local(%{mask[2:2]}));
        result.add(regs.key(Max_Lat).read_local(%{mask[3:3]}));
      };
    };
    check that (result.size() == 32) else
    dut_error("result size != 32\n");
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Provide write access to the registers.
  ////////////////////////////////////////////////////////////////////////////////
  write(addr : byte, mask : uint (bits:4), data : uint) is {
    case(addr) {
      8'h00: {
        regs.key(VendorID).write_local(%{mask[1:0]}, %{data[15:0]});
        regs.key(DeviceID).write_local(%{mask[3:2]}, %{data[31:16]});
      };
      8'h04: {
        regs.key(Command).write_local(%{mask[1:0]}, %{data[15:0]});
        regs.key(Status).write_local(%{mask[3:2]}, %{data[31:16]});
      };
      8'h08: {
        regs.key(RevisionID).write_local(%{mask[0:0]}, %{data[7:0]});
        regs.key(ClassCode).write_local(%{mask[3:1]}, %{data[31:8]});
      };
      8'h0C: {
        regs.key(Cache_Line_Size).write_local(%{mask[0:0]}, %{data[7:0]});
        regs.key(Latency_Timer).write_local(%{mask[1:1]}, %{data[15:8]});
        regs.key(Header_Type).write_local(%{mask[2:2]}, %{data[23:16]});
        regs.key(BIST).write_local(%{mask[3:3]}, %{data[31:24]});
      };
      8'h10: {
        regs.key(BAR0).write_local(%{mask[3:0]}, %{data[31:0]});
      };
      8'h14: {
        regs.key(BAR1).write_local(%{mask[3:0]}, %{data[31:0]});
      };
      8'h18: {
        regs.key(BAR2).write_local(%{mask[3:0]}, %{data[31:0]});
      };
      8'h1C: {
        regs.key(BAR3).write_local(%{mask[3:0]}, %{data[31:0]});
      };
      8'h20: {
        regs.key(BAR4).write_local(%{mask[3:0]}, %{data[31:0]});
      };
      8'h24: {
        regs.key(BAR5).write_local(%{mask[3:0]}, %{data[31:0]});
      };
      8'h28: {
        regs.key(Cardbus_CIS_Pointer).write_local(%{mask[3:0]}, %{data[31:0]});
      };
      8'h2C: {
        regs.key(SubSystem_Vendor_ID).write_local(%{mask[1:0]}, %{data[15:0]});
        regs.key(SubSystem_ID).write_local(%{mask[3:2]}, %{data[31:16]});
      };
      8'h30: {
        regs.key(Expansion_ROM_Base_Address).write_local(%{mask[3:0]}, %{data[31:0]});
      };
      8'h34: {
        regs.key(Capabilities_Pointer).write_local(%{mask[0:0]}, %{data[7:0]});
        regs.key(Reserved_34h).write_local(%{mask[3:1]}, %{data[31:8]});
      };
      8'h38: {
        regs.key(Reserved_38h).write_local(%{mask[0:0]}, %{data[31:0]});
      };
      8'h3C: {
        regs.key(Interrupt_Line).write_local(%{mask[0:0]}, %{data[7:0]});
        regs.key(Interrupt_Pin).write_local(%{mask[1:1]}, %{data[15:8]});
        regs.key(Min_Gnt).write_local(%{mask[2:2]}, %{data[23:16]});
        regs.key(Max_Lat).write_local(%{mask[3:3]}, %{data[31:24]});
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  update_target_UnitID(new_UnitID : uint (bits:5)) is {
    for each in regs do {
      it.target_UnitID = new_UnitID;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Compare the current address to the BAR registers to determine if this address belongs to this device.  If so return TRUE,
  // otherwise return FALSE, indicating to send this request to another device.
  ////////////////////////////////////////////////////////////////////////////////
  compare_address(addr : uint (bits:40)):bool is {
   var range_index : uint;
    // If the address is in the PCI I/O range
    // The PCI I/O range is only a 25 bit space, but bits 31:26 of the BAR must be 0 for a match.
    if(addr[39:24] >= 16'hFDFC && addr[39:24] <= 16'hFDFD) {
      for each (BAR) in Base_Address_Register {
        if(BAR.data[0:0] == 1) { // This is an IO address decoder.
          range_index = 25;
          while(BAR.write_mask[range_index:range_index] == 1) {
            range_index -= 1;
          };
          if(BAR.data[31:26] == 6'b000000 && BAR.data[25:range_index+1] == addr[25:range_index+1]) {
            return TRUE;
          };
        };
      };
    }else if(addr[39:32] <= 8'hFC) { // Memory-Mapped I/O
      var BAR_index : uint;
      var BAR_addr : uint (bits:64);
      var BAR_mask : uint (bits:64);
      BAR_index = 0;
      repeat {
        if(Base_Address_Register[BAR_index].data[2:0] == 3'b100) { // A 64 bit Memory-Mapped region.
          BAR_addr = %{Base_Address_Register[BAR_index+1].data[31:0], Base_Address_Register[BAR_index].data[31:0]};
          // Use the write mask of the BAR register to determine how large the addressible region of the BAR is.
          range_index = 39;
          BAR_mask = %{Base_Address_Register[BAR_index+1].write_mask[31:0], Base_Address_Register[BAR_index].write_mask[31:0]};
          if(BAR_mask[63:0] != 64'h0000_0000_0000_0000) {  // If the BAR_mask is all 0's then the BAR is not in use.
            while(BAR_mask[range_index:range_index] == 1) {
              range_index -= 1;
            };
            if(BAR_addr[63:40] == 24'h000000 && BAR_addr[39:range_index+1] == addr[39:range_index+1]) {
              return TRUE;
            };
          };
          BAR_index += 2;
        }else if(Base_Address_Register[BAR_index].data[2:0] == 3'b110) { // A 40 bit HT Memory-Mapped region.
          // In the PCI Spec, bits [2:1] of the BAR are reserved.  In this case we are using them to indicate that the
          // address should be shifted left 8 bits to represent a 40 bit address.
          BAR_addr = %{Base_Address_Register[BAR_index].data[31:0]};
          // Use the write mask of the BAR register to determine how large the addressible region of the BAR is.
          range_index = 31;
          BAR_mask = %{Base_Address_Register[BAR_index].write_mask[31:0]};
          if(BAR_mask[31:0] != 32'h0000_0000) {  // If the BAR_mask is all 0's then the BAR is not in use.
            while(BAR_mask[range_index:range_index] == 1) {
              range_index -= 1;
            };
            if(BAR_addr[31:range_index+1] == addr[39:range_index+1+8]) { // Effectively shift the bar left by 8 bits.
              return TRUE;
            };
          };
          BAR_index += 1;
        }else { // A 32 bit Memory-Mapped region
          BAR_addr = %{Base_Address_Register[BAR_index].data[31:0]};
          // Use the write mask of the BAR register to determine how large the addressible region of the BAR is.
          range_index = 31;
          BAR_mask = %{Base_Address_Register[BAR_index].write_mask[31:0]};
          if(BAR_mask[31:0] != 32'h0000_0000) {  // If the BAR_mask is all 0's then the BAR is not in use.
            while(BAR_mask[range_index:range_index] == 1) {
              range_index -= 1;
            };
            if(BAR_addr[31:range_index+1] == addr[31:range_index+1]) {
              return TRUE;
            };
          };
          BAR_index += 1;
        };
      } until BAR_index > 5;
      return FALSE;
    }else if(addr[39:24] >= 16'hFDFE && addr[39:24] <= 16'hFDFF && control_block.compare_current_UnitID(addr[15:11])) { // Type 0 PCI Configuration Request to this UnitID
      return TRUE;
    }else {
      return FALSE;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Reset the Device Header.
  ////////////////////////////////////////////////////////////////////////////////
  reset(reset_type : csco_ht_reset_type) is {
    for each in regs {
      it.reset(reset_type);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// csco_ht_device_reg defines the specific device registers that exist in the ht csr space.
// It provides write mas, initialization values, etc.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_device_reg like csco_ht_csr_reg {
  reg_type : csco_ht_device_reg_type;
  // Assign the register type to a name field so that packets can be identified.
  keep reg_name == reg_type.as_a(string);

  // Set the initial value of the target address.  Remember that this has to change when the UnitID is changed.
  keep Addr == %{csr_access_type.as_a(uint), 8'h00, target_UnitID, target_function, doubleword_offset};
  ////////////////////////////////////////////////////////////////////////////////
  when VendorID'reg_type csco_ht_device_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h00;
    keep persistance_mask == %{16'h0000};
    keep soft write_mask == %{16'h0000};
    %ID : uint (bits:16);
    keep soft ID == 16'h0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when DeviceID'reg_type csco_ht_device_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b10;
    keep doubleword_offset == 8'h00;
    keep persistance_mask == %{16'h0000};
    keep soft write_mask == %{16'h0000};
    %ID : uint (bits:16);
    keep soft ID == 16'h0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Command'reg_type csco_ht_device_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h04;
    keep persistance_mask == %{16'h0000};
    keep soft write_mask == %{16'h0107};
    %IO_Space_Enable : bit;
    keep soft IO_Space_Enable == 1'b0;
    %Memory_Space_Enable : bit;
    keep soft Memory_Space_Enable == 1'b0;
    %Bus_Master_Enable : bit;
    keep soft Bus_Master_Enable == 1'b0;
    %Rsv_3_7 : uint (bits:5);
    keep Rsv_3_7 == 5'b00000;
    %SERR_Enable : bit;
    keep soft SERR_Enable == 1'b0;
    %Rsv_9_15 : uint (bits:7);
    keep Rsv_9_15 == 7'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Status'reg_type csco_ht_device_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b10;
    keep doubleword_offset == 8'h04;
    keep persistance_mask == %{16'h7800};
    keep soft write_mask == %{16'h7800};
    %Rsv_0_3 : uint (bits:4);
    keep Rsv_0_3 == 4'h0;
    %Capabilities_List : bit;
    keep Capabilities_List == 1'b1;
    %Rsv_5_10 : uint (bits:6);
    keep Rsv_5_10 == 6'h00;
    %Signaled_Target_Abort : bit;
    keep soft Signaled_Target_Abort == 1'b0;
    %Received_Target_Abort : bit;
    keep soft Received_Target_Abort == 1'b0;
    %Received_Master_Abort : bit;
    keep soft Received_Master_Abort == 1'b0;
    %Signaled_System_Error : bit;
    keep soft Signaled_System_Error == 1'b0;
    %Rsv_15 : bit;
    keep Rsv_15 == 1'b0;

    write_local(byte_mask : list of bit, new_data : list of bit) is first {
      // Certain bits can only be cleared by writing a 1 to them once they are set to a 1.
      // Signaled Target Abort
      // Received Target Abort
      // Received Master Abort
      // Signaled System Error
      if(byte_mask[1] == 1'b1) { // The second byte is being written.
        // Potential Write to Signaled Target Abort
        if(Signaled_Target_Abort == 1'b1) {
        // Invert the value in the data to be written so that a 1 writes a 0 and a 0 writes a 1.
          new_data[11] = ~new_data[11];
        };
        // Potential Write to Received Target Abort
        if(Received_Target_Abort == 1'b1) {
          // Invert the value in the data to be written so that a 1 writes a 0 and a 0 writes a 1.
          new_data[12] = ~new_data[12];
        };
        // Potential Write to Received Master Abort
        if(Received_Master_Abort == 1'b1) {
          // Invert the value in the data to be written so that a 1 writes a 0 and a 0 writes a 1.
          new_data[13] = ~new_data[13];
        };
        // Potential Write to Signaled System Error
        if(Signaled_System_Error == 1'b1) {
          // Invert the value in the data to be written so that a 1 writes a 0 and a 0 writes a 1.
          new_data[14] = ~new_data[14];
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when RevisionID'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h08;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %ID : uint (bits:8);
    keep soft ID == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when ClassCode'reg_type csco_ht_device_reg {
    keep reg_size == TwentyFour;
    keep byte_offset == 2'b01;
    keep doubleword_offset == 8'h08;
    keep persistance_mask == %{24'h00_0000};
    keep soft write_mask == %{24'h00_0000};
    %Code : uint (bits:24);
    keep soft Code == 24'h00_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Cache_Line_Size'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h0C;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %Cache_Line : uint (bits:8);
    keep soft Cache_Line == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Latency_Timer'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b01;
    keep doubleword_offset == 8'h0C;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %Timer : uint (bits:8);
    keep soft Timer == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Header_Type'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b10;
    keep doubleword_offset == 8'h0C;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %Header : uint (bits:8);
    keep soft Header == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when BIST'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b11;
    keep doubleword_offset == 8'h0C;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %BIST : uint (bits:8);
    keep soft BIST == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when BAR0'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h10;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %bar : uint;
    keep soft bar == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when BAR1'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h14;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %bar : uint;
    keep soft bar == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when BAR2'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h18;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %bar : uint;
    keep soft bar == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when BAR3'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h1C;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %bar : uint;
    keep soft bar == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when BAR4'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h20;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %bar : uint;
    keep soft bar == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when BAR5'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h24;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %bar : uint;
    keep soft bar == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Cardbus_CIS_Pointer'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h28;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %Pointer : uint;
    keep soft Pointer == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when SubSystem_Vendor_ID'reg_type csco_ht_device_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h2C;
    keep persistance_mask == %{16'h0000};
    keep soft write_mask == %{16'h0000};
    %ID : uint (bits:16);
    keep soft ID == 16'h0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when SubSystem_ID'reg_type csco_ht_device_reg {
    keep reg_size == Sixteen;
    keep byte_offset == 2'b10;
    keep doubleword_offset == 8'h2C;
    keep persistance_mask == %{16'h0000};
    keep soft write_mask == %{16'h0000};
    %ID : uint (bits:16);
    keep soft ID == 16'h0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Expansion_ROM_Base_Address'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h30;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %ROM_Addr : uint;
    keep soft ROM_Addr == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Capabilities_Pointer'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h34;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %Offset : uint (bits:8);
    keep soft Offset == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Reserved_34h'reg_type csco_ht_device_reg {
    keep reg_size == TwentyFour;
    keep byte_offset == 2'b01;
    keep doubleword_offset == 8'h34;
    keep persistance_mask == %{24'h00_0000};
    keep soft write_mask == %{24'h00_0000};
    %Rsv : uint (bits:24);
    keep soft Rsv == 24'h00_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Reserved_38h'reg_type csco_ht_device_reg {
    keep reg_size == ThirtyTwo;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h38;
    keep persistance_mask == %{32'h0000_0000};
    keep soft write_mask == %{32'h0000_0000};
    %Rsv : uint;
    keep soft Rsv == 32'h0000_0000;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Interrupt_Line'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b00;
    keep doubleword_offset == 8'h3C;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %Line : uint (bits:8);
    keep soft Line == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Interrupt_Pin'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b01;
    keep doubleword_offset == 8'h3C;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %Pin : uint (bits:8);
    keep soft Pin == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Min_Gnt'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b10;
    keep doubleword_offset == 8'h3C;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %MinGnt : uint (bits:8);
    keep soft MinGnt == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Max_Lat'reg_type csco_ht_device_reg {
    keep reg_size == Eight;
    keep byte_offset == 2'b11;
    keep doubleword_offset == 8'h3C;
    keep persistance_mask == %{8'h00};
    keep soft write_mask == %{8'h00};
    %MaxLat : uint (bits:8);
    keep soft MaxLat == 8'h00;
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
'>
