////////////////////////////////////////////////////////////////////////////       // -
// Title         : HT Host Bridge
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////       // -
// File          : csco_ht_host_bridge.e
// Author        : Brian Van Essen  <vanessen@cisco.com>
// Created       : 25.01.2001
// Last modified : 02.05.2001
////////////////////////////////////////////////////////////////////////////       // -
// Description :
// This file is an extension of the csco_ht_device.e file and provides a simple Host
// Bridge device.  This host bridge will correctly size the HT chain and is
// capable of random and directed traffic.  This file is essentially
// "when Host_Bridge'csco_ht_device_function csco_ht_device {};" block.
////////////////////////////////////////////////////////////////////////////////
// Modification history :
// 25.01.2001 : created
//////////////////////////////////////////////////////////////////////////////-

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****

<'
extend Host_Bridge'device_function csco_ht_device {
  // A flag to indicate that this host bridge is the master host bridge on the chain
  // currently, this field is not in use.
  master_bridge : bool;
  keep soft master_bridge == TRUE;
  // Specify that this device only has 1 HT interface.
  keep control_block.num_device_csco_ht_interfaces == 1;
  keep csco_ht_interfaces.size() == 1;
  // A simple keyed list that acts as a generic memory.
  !device_memory : list (key:addr) of csco_ht_memory_element;
  csco_ht_init_scene : csco_ht_scenario;
  keep soft csco_ht_init_scene.kind == BASIC_INIT;
  // A flag used to display information about the logical initialization of the HT chain.
  display_init_info : bool;
  keep soft display_init_info == FALSE;
  hb_type : csco_ht_hostbridge_type;
  keep soft hb_type == NA;
  ////////////////////////////////////////////////////////////////////////////////
  // These are a collection of generic methods that are initially define in csco_ht_device.e and
  // are redefined here to provide specific functionality for an individual device.
  ////////////////////////////////////////////////////////////////////////////////
  // memory_write() defines how accesses to the Hosts memory are handled.
  ////////////////////////////////////////////////////////////////////////////////
  memory_write(packet : csco_ht_packet):bool @csco_ht_device_clock is {
    var mem_latency : uint;
    gen mem_latency keeping {
      it <= max_memory_latency;
    };
    wait [mem_latency] * cycle;

    generic_memory_write(device_memory, packet);
    result = FALSE;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // pci_io_write() defines how accesses to the Hosts IO space are handled.
  ////////////////////////////////////////////////////////////////////////////////
  pci_io_write(packet : csco_ht_packet):bool @csco_ht_device_clock is {
    dut_error("A Write was sent to the PCI IO space on this device, which is not implemented.");
    var io_delay : uint;
    gen io_delay keeping {
      it <= max_io_delay;
    };
    wait [io_delay] * cycle;
    result = TRUE;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // memory_read() defines how accesses to the Hosts memory are handled.
  ////////////////////////////////////////////////////////////////////////////////
  memory_read(packet : csco_ht_packet):list of bit @csco_ht_device_clock is {
    var mem_latency : uint;
    gen mem_latency keeping {
      it <= max_memory_latency;
    };
    wait [mem_latency] * cycle;

    result = generic_memory_read(device_memory, packet);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // pci_io_read() defines how accesses to the Hosts IO space are handled.
  ////////////////////////////////////////////////////////////////////////////////
  pci_io_read(packet : csco_ht_packet):list of bit @csco_ht_device_clock is {
    dut_error("A Read was sent to the PCI IO space on this device, which is not implemented.");
    var io_delay : uint;
    gen io_delay keeping {
      it <= max_io_delay;
    };
    wait [io_delay] * cycle;
    result = {};
  };
  ////////////////////////////////////////////////////////////////////////////////
  // process_forwarded_traffic() defines how the Host will respond to inbound packets that are
  // not destined for the Host Bridge.  In this case, error responses are generated.
  ////////////////////////////////////////////////////////////////////////////////
  process_forwarded_traffic(inbound_channel : csco_ht_channel_interface, outbound_channel : csco_ht_channel_interface, current_packet : csco_ht_packet)@csco_ht_device_clock is {
    var tunneling_delay : uint;
    gen tunneling_delay keeping {
      it <= max_tunneling_delay;
    };
    wait [tunneling_delay] * cycle;
    generate_forwarding_error(inbound_channel, current_packet);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // print_device_records() extends the csco_ht_device.e method to include the print out of the Host memory.
  ////////////////////////////////////////////////////////////////////////////////
  print_device_records() is also {
    if(display_records == TRUE) {
      messagef(LOW, "\n  Main Memory for: %s(%s)\n", control_block.device_reference_name, device_function) {
        outf("////////// Main Memory\n");
        for each in device_memory.sort(.addr) {
          it.print_mem();
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // A Host Bridge needs to initialize the chain.
  ////////////////////////////////////////////////////////////////////////////////
//  run() is also {
  activate_TCMs(reset_type : csco_ht_reset_type)@csco_ht_device_clock is also {
//    start csco_ht_init_scene.run_scenario(csco_ht_interfaces[0]);
    start csco_ht_init_scene.init_scenario(csco_ht_interfaces[0]);
//    start print_traffic_params();
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////

extend csco_ht_address_region : [INIT_REGION];
extend INIT_REGION'region csco_ht_address_map {
  keep soft base_addr == 40'h00_0000_0000;
  keep top_addr == 40'hFC_FFFF_FFFF;
};

extend csco_ht_scenario_kind  : [BASIC_INIT];
extend BASIC_INIT'kind csco_ht_scenario {
  keep soft addr_map.region == INIT_REGION;
  keep target_address_gen_complete == TRUE;
  // Specifies the starting address for the BARs of devices on the HT chain.
  csco_ht_base_address : uint (bits:64);
  keep csco_ht_base_address[63:40] == 24'h00_0000;
  keep csco_ht_base_address >= 64'h0000_0000_0000_0000;
  keep csco_ht_base_address <= 64'h0000_00FC_FFFF_FFFF; // Specifies that the address must be in the range of DRAM/Memory-Mapped IO
  keep csco_ht_base_address[39:0] == addr_map.base_addr;

  NextFreeID : uint (bits:5);
  keep NextFreeID == 5'h01;
  temp_csco_ht_packet : csco_ht_packet;
  previous_device : csco_ht_device_profile;
  current_device : csco_ht_device_profile;
  current_state : csco_ht_init_stages;
  keep current_state == Generic_0;
  next_state : csco_ht_init_stages;
  keep next_state == Generic_0;
  extended_address_decoder : bool;
  keep extended_address_decoder == FALSE;
  csco_ht_reserved_address_decoder : bool;
  keep csco_ht_reserved_address_decoder == TRUE;
  memory_region_size : uint;
  // This flag is used to indicate if there is a pregenerated chain.
  // If the chain is not pregenerated don't check the return value of fields like VendorID & DeviceID.
  generic_topology : bool;
  keep soft generic_topology == TRUE;
  keep parent_device is a Host_Bridge'device_function csco_ht_device(tmp) and tmp.chain_topology.size() == 0 => generic_topology == TRUE;
  keep parent_device is a Host_Bridge'device_function csco_ht_device(tmp) and tmp.chain_topology.size() != 0 => generic_topology == FALSE;
  chain_index : uint; // Keep track of which element on the chain is currently being configured.
  keep chain_index == 0;

  generate_transactions()@parent_device.csco_ht_device_clock is {
    // Now that the Host Bridge has been programmed, configure the chain.
    // If a read or write error occures during initialization just bail on the whole thing.
    // There is no recovery mechanism for this.
    var host_bridge : Host_Bridge'device_function csco_ht_device;
    host_bridge = parent_device.as_a(Host_Bridge'device_function csco_ht_device);
    if(host_bridge.control_block.capability_register.links_initialized()[0] != 1
       || host_bridge.control_block.capability_register.links_crc_errors()[0] != 4'b0000) {
         messagef(MEDIUM, "The Host Bridge is seeing CRC Errors on it's outbound link or the link is not initialized");
         // Write the EOC bit and Transmitter off for the Link Control register on the Host Bridge
         var write_error : bool;
         write_error = host_bridge.control_block.csr_write(40'hFD_FE00_0044, 4'b0001, 32'h0000_00C0);
         current_state = Chain_Init_Complete;
       };

    case(current_state) {
      Generic_0: {
        // Keep a pointer to the last configured device.
        previous_device = current_device;
        if(generic_topology) {
          gen current_device;
          host_bridge.chain_topology.add(current_device);
          chain_index += 1;
        }else {
          current_device = host_bridge.chain_topology[chain_index];
          chain_index += 1;
        };

        if(generic_topology) {
          current_device.device_header.regs.key(VendorID).ignore_data_compare = TRUE;
          current_device.device_header.regs.key(DeviceID).ignore_data_compare = TRUE;
        };
        // Check the Device ID and the Vendor ID
        if(current_device.device_header.regs.key(VendorID).read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        if(current_device.device_header.regs.key(DeviceID).read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        if(generic_topology) {
          current_device.device_header.regs.key(VendorID).ignore_data_compare = FALSE;
          current_device.device_header.regs.key(DeviceID).ignore_data_compare = FALSE;
        };

        // Set the Bus Master Enable, the I/O Space Enable and the Memory Space Enable
        var command_reg : Command'reg_type csco_ht_device_reg;
        command_reg = current_device.device_header.regs.key(Command).as_a(Command'reg_type csco_ht_device_reg);
        command_reg.IO_Space_Enable = 1'b1;
        command_reg.Memory_Space_Enable = 1'b1;
        command_reg.Bus_Master_Enable = 1'b1;
        if(command_reg.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        // Read the Status Register of the Device Header to see if there is a capabilities list.
        if(current_device.device_header.regs.key(Status).read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        if(current_device.device_header.regs.key(Status).as_a(Status'reg_type csco_ht_device_reg).Capabilities_List == 1'b1) {
          // Now that a capabilities pointer has been found go to the next state to explore the device abilities.
          next_state = Generic_1;
        }else {
          // The host bridge has found a device that does not have a capabilities registers.
          // Just terminate this chain initialization sequence.
          next_state = Chain_Init_Complete;
        };
      };
      Generic_1: {
        if(generic_topology) {
          current_device.device_header.regs.key(RevisionID).ignore_data_compare = TRUE;
          current_device.device_header.regs.key(ClassCode).ignore_data_compare = TRUE;
        };
        // Read the Class Code, Vendor ID and Device ID.  If it is an HT device, write the Base Unit ID.
        // Revision ID & Class Code
        if(current_device.device_header.regs.key(RevisionID).read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        if(current_device.device_header.regs.key(ClassCode).read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        if(generic_topology) {
          current_device.device_header.regs.key(RevisionID).ignore_data_compare = FALSE;
          current_device.device_header.regs.key(ClassCode).ignore_data_compare = FALSE;
        };

        // Capabilities Pointer
        if(current_device.device_header.regs.key(Capabilities_Pointer).read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        // Check the type of HT Capabilities.
        var temp_CapID_block : Capability_ID'reg_type csco_ht_capability_reg;
        var tmp_csco_ht_dev_reg :  Capabilities_Pointer'reg_type csco_ht_device_reg;
        tmp_csco_ht_dev_reg =  current_device.device_header.regs.key(Capabilities_Pointer).as_a(Capabilities_Pointer'reg_type csco_ht_device_reg);
        gen temp_CapID_block keeping {
          it.block_offset == tmp_csco_ht_dev_reg.Offset;
        };

        if(temp_CapID_block.read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        if(temp_CapID_block.ID == 8'h08) {
          // This is an HT device
          next_state = HT_1;
        }else {
          // This is not an HT device
          next_state = Non_HT_1;
        };
      };
      HT_1: {
        // Check the devices CRC Error Bits, if errors are received chain sizing is complete.
        // Otherwise the CRC Fatal Enable bit should be set. (I don't know what bit this is.);
        var temp_link_control : Link_Control0'reg_type csco_ht_capability_reg;


        var tmp_csco_ht_dev_reg :  Capabilities_Pointer'reg_type csco_ht_device_reg;
        tmp_csco_ht_dev_reg =  current_device.device_header.regs.key(Capabilities_Pointer).as_a(Capabilities_Pointer'reg_type csco_ht_device_reg);
        gen temp_link_control keeping {
          it.block_offset == tmp_csco_ht_dev_reg.Offset;
        };

        // Set the CSR so that the returned data will not be checked.
        temp_link_control.ignore_data_compare = TRUE;
        if(temp_link_control.read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        // Set the CSR so that the returned data will be checked.
        temp_link_control.ignore_data_compare = FALSE;

        if(temp_link_control.CRC_Error != 4'b0000) {
          // The link is seeing CRC Errors on its upstream interface. So, the chain needs to be terminated prior to this
          // device.
          // Write the EOC bit and Transmitter off for the second Link Control register on the previous device.
          var previous_link_control1_reg : Link_Control1'reg_type csco_ht_capability_reg;
          previous_link_control1_reg = (previous_device.capability_header.regs.key(Link_Control1)).as_a(Link_Control1'reg_type csco_ht_capability_reg);
          previous_link_control1_reg.End_of_Chain = 1;
          previous_link_control1_reg.Transmitter_Off = 1;
          if(previous_link_control1_reg.write_ht(channel) != NO_ERROR) {
            current_state = Chain_Init_Error;
            return;
          };
          // Take the last device off of the list.
          host_bridge.chain_topology.delete(host_bridge.chain_topology.size()-1);
          next_state = Chain_Init_Complete;
        };

        var current_bar_regs : list (key : reg_type) of csco_ht_device_reg;
        // Create a sublist of all of the BARs
        current_bar_regs = current_device.device_header.regs.all(it.doubleword_offset >= 8'h10 && it.doubleword_offset <= 8'h24);
        check that (current_bar_regs.size() == 6) else dut_error("Problem occured while the Host is trying to init BARs.");
        extended_address_decoder = FALSE;
        var current_bar : uint;
        // All HT devices have BAR's so set these now.
        for each (BAR) in current_bar_regs do {
          if(csco_ht_base_address[39:0] > 40'hFC_FFFF_FFFF && csco_ht_base_address[63:40] != 24'h00_0000) {
            sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR, "The entire addressable memory region for HT has been used.", host_bridge.control_block, 0);
            break;
          };
          if(extended_address_decoder == FALSE) {
            // Write all one's to the BAR
            BAR.write_shadow(%{4'hF}, %{32'hFFFF_FFFF});
            if(BAR.write_ht(channel) != NO_ERROR) {
              current_state = Chain_Init_Error;
              return;
            };
            // Set the CSR so that the returned data will not be checked.
            BAR.ignore_data_compare = TRUE;
            // Read back that same BAR
            if(BAR.read_ht(channel) != NO_ERROR) {
              current_state = Chain_Init_Error;
              return;
            };
            // Set the CSR so that the returned data will be checked.
            BAR.ignore_data_compare = FALSE;
            current_bar = (BAR.read_local(%{4'hF}))[31:0];

            case(current_bar[2:0]) { // current_bar[0:0] == 1 => IO Decoder.
              3'b000: { // 32 bit Memory Map decoder
                extended_address_decoder = FALSE;
                csco_ht_reserved_address_decoder = FALSE;

                if(csco_ht_base_address[63:32] != 32'h0000_0000) {
                  sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR, "The only available address are greater than 2^32 and this is only a 32 bit BAR", host_bridge.control_block, 0);
                  messagef(NONE, "ERROR: csco_ht_base_address: %04x_%04x_%04x_%04x\n",
                       csco_ht_base_address[63:42], csco_ht_base_address[41:32],
                       csco_ht_base_address[31:16], csco_ht_base_address[15:0]);
                  break;
                };
              };
              3'b100: { // 64 bit Memory Map decoder
                extended_address_decoder = TRUE;
              };
              3'b110: { // A 40 bit HT Memory Map decoder (A special use of the reserved bits for expanding the BAR's)
                csco_ht_reserved_address_decoder = TRUE;
              };
              default: {
                // Reserved
                sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR, "Unknown BAR decoder encountered.", host_bridge.control_block, 0);
                break;
              };
            };
            memory_region_size = 0;
            for i from 4 to 31 do {
              // Scan the BAR for the first 1 which will indicate the size of the memory range addressed by the BAR
              if(current_bar[i:i] == 1'b1) {
                memory_region_size = i;
                break;
              };
            };
            // Update the write mask of the local BAR to match that of the target BAR.
            if(memory_region_size == 0) {
              BAR.write_mask = %{32'h0000_0000};
            } else {
              BAR.write_mask = %{(32'hFFFF_FFFF << memory_region_size)};
            };

            if(csco_ht_reserved_address_decoder == TRUE) {
              BAR.write_local(%{4'hF}, %{csco_ht_base_address[39:8]});
            }else {
              BAR.write_local(%{4'hF}, %{csco_ht_base_address[31:0]});
            };
            if(BAR.write_ht(channel) != NO_ERROR) {
              current_state = Chain_Init_Error;
              return;
            };
          }else { // extended_address_decoder is TRUE
            // This BAR is the base of a 64 bit BAR and the addressable size is greated than 2^32.
            if(memory_region_size == 0) {
              // Write all one's to the BAR
              BAR.write_shadow(%{4'hF}, %{32'hFFFF_FFFF});
              if(BAR.write_ht(channel) != NO_ERROR) {
                current_state = Chain_Init_Error;
                return;
              };
              // Set the CSR so that the returned data will not be checked.
              BAR.ignore_data_compare = TRUE;
              // Read back that same BAR
              if(BAR.read_ht(channel) != NO_ERROR) {
                current_state = Chain_Init_Error;
                return;
              };
              // Set the CSR so that the returned data will be checked.
              BAR.ignore_data_compare = FALSE;
              current_bar = (BAR.read_local(%{4'hF}))[31:0];
              for i from 0 to 31 do {
                // Scan the BAR for the first 1 which will indicate the size of the memory range addressed by the BAR
                if(current_bar[i:i] == 1'b1) {
                  memory_region_size = i + 32; // This bar is the second of a 64 bit range
                  break;
                };
              };
            };
            // Update the write mask of the local BAR to match that of the target BAR.
            if(memory_region_size == 0) {
              BAR.write_mask = %{32'h0000_0000};
            }else if(memory_region_size < 32) { // The lower BAR has already been set, so this entire BAR must be set.
              BAR.write_mask = %{32'hFFFF_FFFF};
            } else {
              BAR.write_mask = %{(32'hFFFF_FFFF << (memory_region_size-32))};
            };

            // Since this is a 64 bit BAR, write the upper register.
            BAR.write_local(%{4'hF}, %{csco_ht_base_address[63:32]});
            if(BAR.write_ht(channel) != NO_ERROR) {
              current_state = Chain_Init_Error;
              return;
            };
            extended_address_decoder = FALSE;
          };

          if(extended_address_decoder == FALSE) { // Either this was a 32 bit decoder, or the 64 bit decoder has completed configuration.
            if(memory_region_size != 0) { // If the memory_region_size is 0 it means that the BAR is not in use.
              if(csco_ht_reserved_address_decoder) {
                // The HT reserved address decoder allows a 32 bit BAR to represent a 40 bit address range,
                // by shifting the value in the BAR left by 8 bits.  So if this type of BAR was detected, the BAR
                // would have been configured as a 32 bit BAR, but the address space it represents is increased.
                memory_region_size += 8;
              };
              csco_ht_base_address += ~(64'hFFFF_FFFF_FFFF_FFFF << memory_region_size) + 1;
            };
          };
        };

        // Check the type of HT Capabilities.
        var temp_command_block : Command'reg_type csco_ht_capability_reg;

        tmp_csco_ht_dev_reg =  current_device.device_header.regs.key(Capabilities_Pointer).as_a(Capabilities_Pointer'reg_type csco_ht_device_reg);
        gen temp_command_block keeping {
          it.block_offset == tmp_csco_ht_dev_reg.Offset;
        };

        // Set the CSR so that the returned data will not be checked.
        temp_command_block.ignore_data_compare = TRUE;
        if(temp_command_block.read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        // Set the CSR so that the returned data will be checked.
        temp_command_block.ignore_data_compare = FALSE;

        if(temp_command_block.Capability_Type == 3'b000) { // Capability Header is Primary type
          // Check to see if the device is a Tunnel or EOC, this will allow the correct type of header to be created.
          // Check the second Link Config Reg to see if this is an EOC device or a Tunnel.
          var temp_link_control1 : Link_Control1'reg_type csco_ht_capability_reg;

          tmp_csco_ht_dev_reg =  current_device.device_header.regs.key(Capabilities_Pointer).as_a(Capabilities_Pointer'reg_type csco_ht_device_reg);
          gen temp_link_control1 keeping {
          it.block_offset == tmp_csco_ht_dev_reg.Offset;
          };

          // Set the CSR so that the returned data will not be checked.
          temp_link_control1.ignore_data_compare = TRUE;
          if(temp_link_control1.read_ht(channel) != NO_ERROR) {
            current_state = Chain_Init_Error;
            return;
          };
          // Set the CSR so that the returned data will be checked.
          temp_link_control1.ignore_data_compare = FALSE;
          var temp_device_function : csco_ht_device_encoding;
          if(temp_link_control1.Initialization_Complete == 1'b1) { // The device is a Tunnel Device
            temp_device_function = Tunnel_IO_Device;
          }else { // The device is an EOC Device;
            temp_device_function = EOC_IO_Device;
          };
          tmp_csco_ht_dev_reg =  current_device.device_header.regs.key(Capabilities_Pointer).as_a(Capabilities_Pointer'reg_type csco_ht_device_reg);
          gen current_device.capability_header keeping {
            it.device_function == temp_device_function;
            it.block_offset == tmp_csco_ht_dev_reg.Offset;
          };
          // Since the Command block and the Link_Control1 block have already been read, push that data into the new
          // header.
          current_device.capability_header.regs.key(Command).write_shadow(%{2'b11}, temp_command_block.data);
          current_device.capability_header.regs.key(Command).update_expected_data();
          current_device.capability_header.regs.key(Link_Control1).write_shadow(%{2'b11}, temp_link_control1.data);
          current_device.capability_header.regs.key(Link_Control1).update_expected_data();
          next_state = HT_Slave_Primary_1;
        }else {  // Capability Header is Secondary type
          tmp_csco_ht_dev_reg =  current_device.device_header.regs.key(Capabilities_Pointer).as_a(Capabilities_Pointer'reg_type csco_ht_device_reg);
          gen current_device.capability_header keeping {
            it.device_function == Host_Bridge;
            it.block_offset == tmp_csco_ht_dev_reg.Offset;
          };
          // Since the Command block has already been read, push that data into the new header.
          current_device.capability_header.regs.key(Command).write_shadow(%{2'b11}, temp_command_block.data);
          current_device.capability_header.regs.key(Command).update_expected_data();

          next_state = HT_Host_1;
        };
      };
      HT_Slave_Primary_1: {
        // Perform a write to the command register to set the Master Host Bit.
          if(current_device.capability_header.regs.key(Command).write_ht(channel) != NO_ERROR) {
            current_state = Chain_Init_Error;
            return;
           };

        // Read the HT RevID
        if(current_device.capability_header.regs.key(HT_Rev_ID).read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        // If the second Control register has a Initialized Complete bit of 0, then the device is EOC.
        if(current_device.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).Initialization_Complete == 1'b0) {
          next_state = HT_EOC_1;
        }else {
          next_state = HT_Tunnel_1;
        };
      };
      HT_EOC_1: {
        current_device.target_device_function = EOC_IO_Device;
        // Write the EOC bit and Transmitter off for the second Link Control register on this device.
        var temp_link_control1 : Link_Control1'reg_type csco_ht_capability_reg;
        temp_link_control1 = current_device.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg);
        temp_link_control1.End_of_Chain = 1'b1;
        temp_link_control1.Transmitter_Off = 1'b1;
        if(temp_link_control1.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        // Set the base unit id for this device.

        var temp_command_block : Command'reg_type Primary'interface_type csco_ht_capability_reg;
        if current_device.capability_header.regs.key(Command) is a Command'reg_type Primary'interface_type csco_ht_capability_reg (cpreg) {
                temp_command_block = cpreg ;
        } else {
                error("Testbench error, expecting register to be correct subtype because....");
        };
        temp_command_block.Base_UnitID = NextFreeID;

        if(temp_command_block.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        // Make sure that the target_UnitID is also updated for the device and capability header.
        current_device.device_header.update_target_UnitID(NextFreeID);
        current_device.capability_header.update_target_UnitID(NextFreeID);
        NextFreeID += temp_command_block.Unit_Count;

        // Verify that the base unit id got set correctly. (Use a new version of a Command reg.)
        var new_command_block : Primary'interface_type Command'reg_type csco_ht_capability_reg;
        gen new_command_block keeping {
          it is a Primary'interface_type Command'reg_type csco_ht_capability_reg(tmp)                                       and
          tmp.block_offset == current_device.capability_header.block_offset                                             and
          tmp.target_UnitID == temp_command_block.Base_UnitID                                                           and
          tmp.Base_UnitID == temp_command_block.Base_UnitID;
        };
        if(new_command_block.read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        if(temp_command_block.Base_UnitID != new_command_block.Base_UnitID) {
          sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR, "Host bridge failed to set UnitID for device.", host_bridge.control_block, 0);
        };
        next_state = Chain_Init_Complete;
      };
      HT_Tunnel_1: {
        current_device.target_device_function = Tunnel_IO_Device;
        // Check the outbound links on this device for CRC Errors, if errors are received finish configuring this
        // device and mark it as the end of chain.  Additionally check the Initialization Complete bit which would indicate
        // that there is no other device attached to this Tunnel.
        var outbound_link_control : Link_Control1'reg_type csco_ht_capability_reg;
        outbound_link_control = current_device.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg);
        if(outbound_link_control.read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        if(outbound_link_control.CRC_Error != 4'b0000 || outbound_link_control.Initialization_Complete == 1'b0) {
          // This device is seeing CRC Errors on its downstream interface, or
          // The downstream interface on this device never go initialized.
          // Write the EOC bit and Transmitter off for the second Link Control register on this device.
          outbound_link_control.End_of_Chain = 1'b1;
          outbound_link_control.Transmitter_Off = 1'b1;
          if(outbound_link_control.write_ht(channel) != NO_ERROR) {
            current_state = Chain_Init_Error;
            return;
          };
          next_state = Chain_Init_Complete;
        }else {
          next_state = Generic_0; // This device is complete, start over with the next one on the chain.
        };

       // Set the base unit id for this device.

        var temp_command_block : Command'reg_type Primary'interface_type csco_ht_capability_reg;
        if current_device.capability_header.regs.key(Command) is a Command'reg_type Primary'interface_type csco_ht_capability_reg (cpreg) {
                temp_command_block = cpreg ;
        } else {
                error("Testbench error, expecting register to be correct subtype because....");
        };
        temp_command_block.Base_UnitID = NextFreeID;



        if(temp_command_block.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        // Make sure that the target_UnitID is also updated for the device and capability header.
        current_device.device_header.update_target_UnitID(NextFreeID);
        current_device.capability_header.update_target_UnitID(NextFreeID);
        NextFreeID += temp_command_block.Unit_Count;

        // Verify that the base unit id got set correctly.

        var new_command_block : Primary'interface_type Command'reg_type csco_ht_capability_reg;
        gen new_command_block keeping {
          it is a Primary'interface_type Command'reg_type csco_ht_capability_reg(tmp)                                       and
          tmp.block_offset == current_device.capability_header.block_offset                                             and
          tmp.target_UnitID == temp_command_block.Base_UnitID                                                           and
          tmp.Base_UnitID == temp_command_block.Base_UnitID;
        };


        if(new_command_block.read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        if(temp_command_block.Base_UnitID != new_command_block.Base_UnitID) {
          sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR, "Host bridge failed to set UnitID for device.", host_bridge.control_block, 0);
          dut_error("Host bridge failed to set UnitID for device.");
        };
      };
      HT_Host_1: {
        current_device.target_device_function = Host_Bridge;
        // Read the HT RevID
        if(current_device.capability_header.regs.key(HT_Rev_ID).read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        // Set the double ended bit for the other host bridge.

        var temp_command_block : Command'reg_type Secondary'interface_type csco_ht_capability_reg;
        if current_device.capability_header.regs.key(Command) is a Command'reg_type Secondary'interface_type csco_ht_capability_reg (cpreg) {
                 temp_command_block = cpreg ;
         } else {
                 error("Testbench error, expecting register to be correct subtype because....");
         };
         temp_command_block.Double_Ended = 1'b1;

        if(temp_command_block.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        next_state = Chain_Init_Complete;
      };
      Non_HT_1: {
        next_state = Generic_0;
      };
      Chain_Init_Complete: {
        if(host_bridge.display_init_info == TRUE) {
          messagef(LOW, "HT Chain Sizing is complete!\n");
          host_bridge.print_chain_topology();
        };
        // Signal that the chain has been initialized
        host_bridge.control_block.chain_initialization_complete = TRUE;
        // Indicate that the scenario is complete.
        done = TRUE;
      };
      Chain_Init_Error: {
        sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR, "The Host Bridge encountered an error while trying to configure the chain.", host_bridge.control_block, 0);

        if(host_bridge.display_init_info == TRUE) {
          messagef(NONE, "ERROR : HT Chain Sizing is complete because an error occured!\n");
          host_bridge.print_chain_topology();
        };
        // Signal that the chain has been initialized
        host_bridge.control_block.chain_initialization_complete = TRUE;
        // Indicate that the scenario is complete.
        done = TRUE;
      };
      default: {
        dut_error("Host Bridge initialization FSM has entered into an unknown state.");
      };
    };
    if(host_bridge.display_init_info == TRUE) {
      messagef(HIGH, "Current State = %s Next State = %s\n",
               current_state, next_state);
    };
    current_state = next_state;
    wait cycle;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // initialize_csco_ht_chain() is a Host Bridge specific method that will walk and size the HT chain.
  // Additionally, it will configure the BAR's in the Host bridge.  By definition of the HT spec,
  // the allocation of UnitID's is always done the same way, but the BAR's are configures as consequtive regions
  // of memory starting at csco_ht_base_address. (Which is user configurable.)
  // If a read or write error occurs while trying to configure the chain, then the Host Bridge will immediately
  // quit chain sizing.  There is no recovery mechanism for errors during sizing.
  ////////////////////////////////////////////////////////////////////////////////
  initialize_csco_ht_chain()@parent_device.csco_ht_device_clock is {
    wait true(parent_device.control_block.physical_link_initialized(channel.channel_identifier) == TRUE);

    // Initially configure your own Device Header.
    // Set the Bus Master Enable, the I/O Space Enable and the Memory Space Enable
    var csr_write_error : bool;
    csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0004, 4'h3, 32'h0000_0007);

    var current_bar : uint;
    var bar_index : uint;
    // All HT devices have BAR's so set these now.
    for{bar_index = 0; bar_index < 6; bar_index += 1} {
      // Write all one's to the BAR
      csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, 32'hFFFF_FFFF);

      // Read back that same BAR
      current_bar = parent_device.control_block.csr_read(40'hFD_FE00_0010+(bar_index*4), 4'hF)[31:0];

      case(current_bar[2:1]) {
        2'b00: { // 32 bit decoder
          extended_address_decoder = FALSE;
          csco_ht_reserved_address_decoder = FALSE;
        };
        2'b10: { // 64 bit decoder
          extended_address_decoder = TRUE;
        };
        3'b11: { // A 40 bit HT Memory Map decoder (A special use of the reserved bits for expanding the BAR's)
          csco_ht_reserved_address_decoder = TRUE;
        };
        default: {
          // Reserved
        };
      };
      memory_region_size = 0;
      for i from 4 to 31 do {
        // Scan the BAR for the first 1 which will indicate the size of the memory range addressed by the BAR
        if(current_bar[i:i] == 1'b1) {
          memory_region_size = i;
          break;
        };
      };
      if(current_bar[0:0] == 1'b0) { // This is a memory decoder
        // Since this is a memory_map BAR write the next free base address into it.
        if(extended_address_decoder == FALSE && csco_ht_reserved_address_decoder == FALSE && csco_ht_base_address[63:32] != 32'h0000_0000) {
          dut_error("The only available address are greater than 2^32 and this is only a 32 bit BAR");
        };
        if(extended_address_decoder == FALSE && csco_ht_reserved_address_decoder == TRUE && csco_ht_base_address[63:40] != 24'h00_0000) {
          dut_error("The only available address are greater than 2^40 and this is only a 40 bit BAR");
        };
        if(csco_ht_base_address[39:0] > 40'hFC_FFFF_FFFF && csco_ht_base_address[63:40] != 24'h00_0000) {
          dut_error("The entire addressable memory region for HT has been used.");
        };
        if(csco_ht_reserved_address_decoder == TRUE) {
          csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, csco_ht_base_address[39:8]);
        }else {
          csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, csco_ht_base_address[31:0]);
        };
        if(extended_address_decoder == TRUE) {
          // This BAR is the base of a 64 bit BAR and the addressable size is greated than 2^32.
          bar_index += 1;

          if(memory_region_size == 0) {
            // Write all one's to the BAR
            csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, 32'hFFFF_FFFF);

            // Read back that same BAR
            current_bar = parent_device.control_block.csr_read(40'hFD_FE00_0010+(bar_index*4), 4'hF)[31:0];

            for i from 0 to 31 do {
              // Scan the BAR for the first 1 which will indicate the size of the memory range addressed by the BAR
              if(current_bar[i:i] == 1'b1) {
                memory_region_size = i + 32; // This bar is the second of a 64 bit range
                break;
              };
            };
          };
          // Since this is a 64 bit BAR, write the upper register.
          csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, csco_ht_base_address[63:32]);
        };
        if(memory_region_size != 0) { // If the memory_region_size is 0 it means that the BAR is not in use.
          if(csco_ht_reserved_address_decoder) {
            // The HT reserved address decoder allows a 32 bit BAR to represent a 40 bit address range,
            // by shifting the value in the BAR left by 8 bits.  So if this type of BAR was detected, the BAR
            // would have been configured as a 32 bit BAR, but the address space it represents is increased.
            memory_region_size += 8;
          };
          csco_ht_base_address += ~(64'hFFFF_FFFF_FFFF_FFFF << memory_region_size) + 1;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  run_scenario(channel : csco_ht_channel_interface)@parent_device.csco_ht_device_clock is first {
    me.channel = channel; // This is a hack to all the run_scenario to be easily extended.
    initialize_csco_ht_chain();
  };
  ////////////////////////////////////////////////////////////////////////////////
  run_scenario(channel : csco_ht_channel_interface)@parent_device.csco_ht_device_clock is also {
    // Once the Host Bridge has completed the initialization issue an event to set a flag in csco_ht_test_control.
    emit sys.csco_ht_test_control.csco_ht_model_initialized;
    // If the model is being used as a development system, issued the csco_ht_start_traffic event to kick off the devices.
    if(sys.csco_ht_development_test == TRUE) {
      emit sys.csco_ht_test_control.csco_ht_start_traffic;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
};

// scenario to init the csco_ht_chain, this should only be run in place of BASIC_INIT if
// the user has configured the chain himself (or herself) - this includes generating and
// configured the capability regs accordingly, as well as setting the BARs the the initial readback value
// this will write the BARs accordingly, sets the Base UnitID and some other regs


extend csco_ht_scenario_kind : [FAST_BASIC_INIT];
extend FAST_BASIC_INIT'kind csco_ht_scenario {
  keep addr_map.region == INIT_REGION;
  keep target_address_gen_complete == TRUE;
  // Specifies the starting address for the BARs of devices on the HT chain.
  csco_ht_base_address : uint (bits:64);
  keep csco_ht_base_address[63:40] == 24'h00_0000;
  keep csco_ht_base_address >= 64'h0000_0000_0000_0000;
  // Specifies that the address must be in the range of DRAM/Memory-Mapped IO
  keep csco_ht_base_address <= 64'h0000_00FC_FFFF_FFFF;
  keep csco_ht_base_address[39:0] == addr_map.base_addr;

  NextFreeID : uint (bits:5);
  keep NextFreeID == 5'h01;
  current_device : csco_ht_device_profile;
  current_state : csco_ht_init_stages;
  keep current_state == Generic_0;
  next_state : csco_ht_init_stages;
  keep next_state == Generic_0;
  extended_address_decoder : bool;
  keep extended_address_decoder == FALSE;
  csco_ht_reserved_address_decoder : bool;
  keep csco_ht_reserved_address_decoder == TRUE;
  memory_region_size : uint;
  chain_index : uint; // Keep track of which element on the chain is currently being configured.
  keep chain_index == 0;

  generate_transactions()@parent_device.csco_ht_device_clock is {
    // Now that the Host Bridge has been programmed, configure the chain.
    // If a read or write error occures during initialization just bail on the whole thing.
    // There is no recovery mechanism for this.
    var host_bridge : Host_Bridge'device_function csco_ht_device;
    host_bridge = parent_device.as_a(Host_Bridge'device_function csco_ht_device);
    if(host_bridge.control_block.capability_register.links_initialized()[0] != 1
       || host_bridge.control_block.capability_register.links_crc_errors()[0] != 4'b0000) {
         messagef(MEDIUM, "The Host Bridge is seeing CRC Errors on it's outbound link or the link is not initialized");
         // Write the EOC bit and Transmitter off for the Link Control register on the Host Bridge
         var write_error : bool;
         write_error = host_bridge.control_block.csr_write(40'hFD_FE00_0044, 4'b0001, 32'h0000_00C0);
         current_state = Chain_Init_Complete;
       };

    case(current_state) {
      Generic_0: {
        current_device = host_bridge.chain_topology[chain_index];
        chain_index += 1;

        // Set the Bus Master Enable, the I/O Space Enable and the Memory Space Enable
        var command_reg : Command'reg_type csco_ht_device_reg;
        command_reg = current_device.device_header.regs.key(Command).as_a(Command'reg_type csco_ht_device_reg);
        command_reg.IO_Space_Enable = 1'b1;
        command_reg.Memory_Space_Enable = 1'b1;
        command_reg.Bus_Master_Enable = 1'b1;
        if(command_reg.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        if(current_device.device_header.regs.key(Status).as_a(Status'reg_type csco_ht_device_reg).Capabilities_List == 1'b1) {
          // Now that a capabilities pointer has been found go to the next state to explore the device abilities.
          if(current_device.capability_header.regs.key(Capability_ID).as_a(Capability_ID'reg_type csco_ht_capability_reg).ID == 8'h08) {
            // This is an HT device
            next_state = HT_1;
          }else {
            // This is not an HT device
            next_state = Non_HT_1;
          };
          //          next_state = Generic_1;
        }else {
          // The host bridge has found a device that does not have a capabilities registers.
          // Just terminate this chain initialization sequence.
          next_state = Chain_Init_Complete;
        };
      };
      HT_1: {
        var current_bar_regs : list (key : reg_type) of csco_ht_device_reg;
        // Create a sublist of all of the BARs
        current_bar_regs = current_device.device_header.regs.all(it.doubleword_offset >= 8'h10 && it.doubleword_offset <= 8'h24);
        check that (current_bar_regs.size() == 6) else dut_error("Problem occured while the Host is trying to init BARs.");
        extended_address_decoder = FALSE;
        var current_bar : uint;
        // All HT devices have BAR's so set these now.
        for each (BAR) in current_bar_regs do {
          if(csco_ht_base_address[39:0] > 40'hFC_FFFF_FFFF && csco_ht_base_address[63:40] != 24'h00_0000) {
            sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR,
                                                 "The entire addressable memory region for HT has been used.",
                                                 host_bridge.control_block, 0);
            break;
          };
          if(extended_address_decoder == FALSE) {
            current_bar = (BAR.read_local(%{4'hF}))[31:0];

            case(current_bar[2:0]) { // current_bar[0:0] == 1 => IO Decoder.
              3'b000: { // 32 bit Memory Map decoder
                extended_address_decoder = FALSE;
                csco_ht_reserved_address_decoder = FALSE;

                if(csco_ht_base_address[63:32] != 32'h0000_0000) {
                  sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR,
                                                       "The only available address are greater than 2^32 and this is only a 32 bit BAR",
                                                       host_bridge.control_block, 0);
                  messagef(NONE, "ERROR: csco_ht_base_address: %04x_%04x_%04x_%04x\n",
                       csco_ht_base_address[63:42], csco_ht_base_address[41:32],
                       csco_ht_base_address[31:16], csco_ht_base_address[15:0]);
                  break;
                };
              };
              3'b100: { // 64 bit Memory Map decoder
                extended_address_decoder = TRUE;
              };
              3'b110: { // A 40 bit HT Memory Map decoder (A special use of the reserved bits for expanding the BAR's)
                csco_ht_reserved_address_decoder = TRUE;
              };
              default: {
                // Reserved
                sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR, "Unknown BAR decoder encountered.", host_bridge.control_block, 0);
                break;
              };
            };
            memory_region_size = 0;
            for i from 4 to 31 do {
              // Scan the BAR for the first 1 which will indicate the size of the memory range addressed by the BAR
              if(current_bar[i:i] == 1'b1) {
                memory_region_size = i;
                break;
              };
            };
            // Update the write mask of the local BAR to match that of the target BAR.
            if(memory_region_size == 0) {
              BAR.write_mask = %{32'h0000_0000};
            } else {
              BAR.write_mask = %{(32'hFFFF_FFFF << memory_region_size)};
            };

            if(csco_ht_reserved_address_decoder == TRUE) {
              BAR.write_local(%{4'hF}, %{csco_ht_base_address[39:8]});
            }else {
              BAR.write_local(%{4'hF}, %{csco_ht_base_address[31:0]});
            };
            if(BAR.write_ht(channel) != NO_ERROR) {
              current_state = Chain_Init_Error;
              return;
            };
          } else { // extended_address_decoder is TRUE
            // This BAR is the base of a 64 bit BAR and the addressable size is greated than 2^32.
            if(memory_region_size == 0) {
              current_bar = (BAR.read_local(%{4'hF}))[31:0];
              for i from 0 to 31 do {
                // Scan the BAR for the first 1 which will indicate the size of the memory range addressed by the BAR
                if(current_bar[i:i] == 1'b1) {
                  memory_region_size = i + 32; // This bar is the second of a 64 bit range
                  break;
                };
              };
            };
            // Update the write mask of the local BAR to match that of the target BAR.
            if(memory_region_size == 0) {
              BAR.write_mask = %{32'h0000_0000};
            }else if(memory_region_size < 32) { // The lower BAR has already been set, so this entire BAR must be set.
              BAR.write_mask = %{32'hFFFF_FFFF};
            } else {
              BAR.write_mask = %{(32'hFFFF_FFFF << (memory_region_size-32))};
            };

            // Since this is a 64 bit BAR, write the upper register.
            BAR.write_local(%{4'hF}, %{csco_ht_base_address[63:32]});
            if(BAR.write_ht(channel) != NO_ERROR) {
              current_state = Chain_Init_Error;
              return;
            };
            extended_address_decoder = FALSE;
          };

          if(extended_address_decoder == FALSE) { // Either this was a 32 bit decoder, or the 64 bit decoder has completed configuration.
            if(memory_region_size != 0) { // If the memory_region_size is 0 it means that the BAR is not in use.
              if(csco_ht_reserved_address_decoder) {
                // The HT reserved address decoder allows a 32 bit BAR to represent a 40 bit address range,
                // by shifting the value in the BAR left by 8 bits.  So if this type of BAR was detected, the BAR
                // would have been configured as a 32 bit BAR, but the address space it represents is increased.
                memory_region_size += 8;
              };
              csco_ht_base_address += ~(64'hFFFF_FFFF_FFFF_FFFF << memory_region_size) + 1;
            };
          };
        };

        // Capability Header is Primary type
        if(current_device.capability_header.regs.key(Command).as_a(Command'reg_type csco_ht_capability_reg).Capability_Type == 3'b000) {
          // If the second Control register has a Initialized Complete bit of 0, then the device is EOC.
          if(current_device.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg ).Initialization_Complete == 1'b0) {
            next_state = HT_EOC_1;
          }else {
            next_state = HT_Tunnel_1;
          };
        }else {  // Capability Header is Secondary type
          next_state = HT_Host_1;
        };
      };

      HT_EOC_1: {
        // Set the base unit id for this device.

        var temp_command_block : Command'reg_type Primary'interface_type csco_ht_capability_reg;
        if current_device.capability_header.regs.key(Command) is a Command'reg_type Primary'interface_type csco_ht_capability_reg (cpreg) {
                temp_command_block = cpreg ;
        } else {
                error("Testbench error, expecting register to be correct subtype because....");
        };

        if(temp_command_block.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        // Make sure that the target_UnitID is also updated for the device and capability header.
        current_device.device_header.update_target_UnitID(NextFreeID);
        current_device.capability_header.update_target_UnitID(NextFreeID);
        NextFreeID += temp_command_block.Unit_Count;

        // Verify that the base unit id got set correctly. (Use a new version of a Command reg.)
        var new_command_block : Command'reg_type Primary'interface_type csco_ht_capability_reg;
        gen new_command_block keeping {
          it.block_offset == current_device.capability_header.block_offset;
          it.target_UnitID == temp_command_block.Base_UnitID;
          it.Base_UnitID == temp_command_block.Base_UnitID;
        };

        if(new_command_block.read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        if(temp_command_block.Base_UnitID != new_command_block.Base_UnitID) {
          sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR,
                                               "Host bridge failed to set UnitID for device.",
                                               host_bridge.control_block, 0); }; next_state = Chain_Init_Complete;
      };
      HT_Tunnel_1: {
        // Set the base unit id for this device.

        var temp_command_block : Command'reg_type Primary'interface_type csco_ht_capability_reg;
        if current_device.capability_header.regs.key(Command) is a Command'reg_type Primary'interface_type csco_ht_capability_reg (cpreg) {
                temp_command_block = cpreg ;
        } else {
                error("Testbench error, expecting register to be correct subtype because....");
        };

        if(temp_command_block.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        // Make sure that the target_UnitID is also updated for the device and capability header.
        current_device.device_header.update_target_UnitID(NextFreeID);
        current_device.capability_header.update_target_UnitID(NextFreeID);
        NextFreeID += temp_command_block.Unit_Count;

        // Verify that the base unit id got set correctly.
        var new_command_block : Command'reg_type Primary'interface_type csco_ht_capability_reg;
        gen new_command_block keeping {
        it.block_offset == current_device.capability_header.block_offset;
        it.target_UnitID == temp_command_block.Base_UnitID;
        it.Base_UnitID == temp_command_block.Base_UnitID;
        };

        if(new_command_block.read_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };

        if((temp_command_block.Base_UnitID) != (new_command_block.Base_UnitID)) {
          sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR,
                                               "Host bridge failed to set UnitID for device.",
                                               host_bridge.control_block, 0);
          dut_error("Host bridge failed to set UnitID for device.");
        };
      next_state = Generic_0; // This device is complete, start over with the next one on the chain.
      };
      HT_Host_1: {
        // Set the double ended bit for the other host bridge.
        var temp_command_block : Command'reg_type csco_ht_capability_reg;
        temp_command_block = current_device.capability_header.regs.key(Command).as_a(Command'reg_type csco_ht_capability_reg);
        if(temp_command_block.write_ht(channel) != NO_ERROR) {
          current_state = Chain_Init_Error;
          return;
        };
        next_state = Chain_Init_Complete;
      };
      Non_HT_1: {
        next_state = Generic_0;
      };
      Chain_Init_Complete: {
        if(host_bridge.display_init_info == TRUE) {
          messagef(LOW, "HT Chain Sizing is complete!\n");
          host_bridge.print_chain_topology();
        };
        // Signal that the chain has been initialized
        host_bridge.control_block.chain_initialization_complete = TRUE;
        // Indicate that the scenario is complete.
        done = TRUE;
      };
      Chain_Init_Error: {
        sys.csco_ht_errors.report_protocol_error(CHAIN_CONFIG_ERROR,
                                             "The Host Bridge encountered an error while trying to configure the chain.",
                                             host_bridge.control_block, 0);

        if(host_bridge.display_init_info == TRUE) {
          messagef(NONE, "ERROR : HT Chain Sizing is complete because an error occured\n");
          host_bridge.print_chain_topology();
        };
        // Signal that the chain has been initialized
        host_bridge.control_block.chain_initialization_complete = TRUE;
        // Indicate that the scenario is complete.
        done = TRUE;
      };
      default: {
        dut_error("Host Bridge initialization FSM has entered into an unknown state.");
      };
    };
    if(host_bridge.display_init_info == TRUE) {
      messagef(HIGH, "Current State = %s Next State = %s\n", current_state, next_state);
    };
    current_state = next_state;
    wait cycle;
  };

  ////////////////////////////////////////////////////////////////////////////////
  // initialize_csco_ht_chain() is a Host Bridge specific method that will walk and size the HT chain.
  // Additionally, it will configure the BAR's in the Host bridge.  By definition of the HT spec,
  // the allocation of UnitID's is always done the same way, but the BAR's are configures as consequtive regions
  // of memory starting at csco_ht_base_address. (Which is user configurable.)
  // If a read or write error occurs while trying to configure the chain, then the Host Bridge will immediately
  // quit chain sizing.  There is no recovery mechanism for errors during sizing.
  ////////////////////////////////////////////////////////////////////////////////
  initialize_csco_ht_chain()@parent_device.csco_ht_device_clock is {
    wait true(parent_device.control_block.physical_link_initialized(channel.channel_identifier) == TRUE);

    // Initially configure your own Device Header.
    // Set the Bus Master Enable, the I/O Space Enable and the Memory Space Enable
    var csr_write_error : bool;
    csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0004, 4'h3, 32'h0000_0007);

    var current_bar : uint;
    var bar_index : uint;
    // All HT devices have BAR's so set these now.
    for{bar_index = 0; bar_index < 6; bar_index += 1} {
      // Write all one's to the BAR
      csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, 32'hFFFF_FFFF);

      // Read back that same BAR
      current_bar = parent_device.control_block.csr_read(40'hFD_FE00_0010+(bar_index*4), 4'hF)[31:0];

      case(current_bar[2:1]) {
        2'b00: { // 32 bit decoder
          extended_address_decoder = FALSE;
          csco_ht_reserved_address_decoder = FALSE;
        };
        2'b10: { // 64 bit decoder
          extended_address_decoder = TRUE;
        };
        3'b11: { // A 40 bit HT Memory Map decoder (A special use of the reserved bits for expanding the BAR's)
          csco_ht_reserved_address_decoder = TRUE;
        };
        default: {
          // Reserved
        };
      };
      memory_region_size = 0;
      for i from 4 to 31 do {
        // Scan the BAR for the first 1 which will indicate the size of the memory range addressed by the BAR
        if(current_bar[i:i] == 1'b1) {
          memory_region_size = i;
          break;
        };
      };
      if(current_bar[0:0] == 1'b0) { // This is a memory decoder
        // Since this is a memory_map BAR write the next free base address into it.
        if((extended_address_decoder == FALSE) &&
           (csco_ht_reserved_address_decoder == FALSE) &&
           (csco_ht_base_address[63:32] != 32'h0000_0000)) {
           dut_error("The only available address are greater than 2^32 and this is only a 32 bit BAR");
        };
        if(extended_address_decoder == FALSE && csco_ht_reserved_address_decoder == TRUE && csco_ht_base_address[63:40] != 24'h00_0000) {
          dut_error("The only available address are greater than 2^40 and this is only a 40 bit BAR");
        };
        if(csco_ht_base_address[39:0] > 40'hFC_FFFF_FFFF && csco_ht_base_address[63:40] != 24'h00_0000) {
          dut_error("The entire addressable memory region for HT has been used.");
        };
        if(csco_ht_reserved_address_decoder == TRUE) {
          csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, csco_ht_base_address[39:8]);
        }else {
          csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, csco_ht_base_address[31:0]);
        };
        if(extended_address_decoder == TRUE) {
          // This BAR is the base of a 64 bit BAR and the addressable size is greated than 2^32.
          bar_index += 1;

          if(memory_region_size == 0) {
            // Write all one's to the BAR
            csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, 32'hFFFF_FFFF);

            // Read back that same BAR
            current_bar = parent_device.control_block.csr_read(40'hFD_FE00_0010+(bar_index*4), 4'hF)[31:0];

            for i from 0 to 31 do {
              // Scan the BAR for the first 1 which will indicate the size of the memory range addressed by the BAR
              if(current_bar[i:i] == 1'b1) {
                memory_region_size = i + 32; // This bar is the second of a 64 bit range
                break;
              };
            };
          };
          // Since this is a 64 bit BAR, write the upper register.
          csr_write_error = parent_device.control_block.csr_write(40'hFD_FE00_0010+(bar_index*4), 4'hF, csco_ht_base_address[63:32]);
        };
        if(memory_region_size != 0) { // If the memory_region_size is 0 it means that the BAR is not in use.
          if(csco_ht_reserved_address_decoder) {
            // The HT reserved address decoder allows a 32 bit BAR to represent a 40 bit address range,
            // by shifting the value in the BAR left by 8 bits.  So if this type of BAR was detected, the BAR
            // would have been configured as a 32 bit BAR, but the address space it represents is increased.
            memory_region_size += 8;
          };
          csco_ht_base_address += ~(64'hFFFF_FFFF_FFFF_FFFF << memory_region_size) + 1;
        };
      };
    };

    if(parent_device.as_a(Host_Bridge'device_function csco_ht_device).display_init_info == TRUE) {
      messagef(MEDIUM, "done configuring hostbridge bars\n");
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  run_scenario(channel : csco_ht_channel_interface)@parent_device.csco_ht_device_clock is first {
    me.channel = channel; // This is a hack to all the run_scenario to be easily extended.
    initialize_csco_ht_chain();
  };
  ////////////////////////////////////////////////////////////////////////////////
  run_scenario(channel : csco_ht_channel_interface)@parent_device.csco_ht_device_clock is also {
    // Once the Host Bridge has completed the initialization issue an event to set a flag in csco_ht_test_control.
    emit sys.csco_ht_test_control.csco_ht_model_initialized;
    // If the model is being used as a development system, issued the csco_ht_start_traffic event to kick off the devices.
    if(sys.csco_ht_development_test == TRUE) {
      emit sys.csco_ht_test_control.csco_ht_start_traffic;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
};

extend Host_Bridge'device_function csco_ht_device {
  ////////////////////////////////////////////////////////////////////////////////
  // print chain topology configuration
  ////////////////////////////////////////////////////////////////////////////////
  print_chain_topology() is {
    messagef(LOW, "current topology:\n") {
      for each in chain_topology {
        outf("  device %d - %s\n", index, it.target_device_function);
      };
    };
  };

//  print_csco_ht_chain_config()@sys.csco_ht_test_control.csco_ht_model_initialized is {
  print_csco_ht_chain_config()@csco_ht_device_clock is {
    var unit_id: list of uint(bits:5);
    var device_func: string;
    unit_id.add(0); // init first item to zero

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @reset @reset;
      };
      {
        wait @sys.csco_ht_test_control.csco_ht_model_initialized;
    device_func = ((control_block.device_function == Host_Bridge) ? "Host Dev" : ((control_block.device_function == Tunnel_IO_Device) ? "Tunnel Dev" : "EOC Dev"));

    messagef(LOW, "HT CHAIN CONFIGURATION\n");
    messagef(LOW, "configured via %s scenario\n", csco_ht_init_scene.kind);
    for each (prof) in chain_topology {
      // check if the capability header exists.. in case something went wrong during init
      if(prof.capability_header != NULL) {
        unit_id = prof.capability_header.read_UnitID();
      } else {
        unit_id[0] = 0;
      };

      messagef(LOW, "Device: %s, UnitID: %d\n", prof.target_device_function, unit_id[0]) {
        outf("  Device Header Registers\n");
        outf("    DeviceID Reg: ID=0x%04x\n",
             prof.device_header.regs.key(DeviceID).as_a(DeviceID'reg_type csco_ht_device_reg).ID);

        outf("    VendorID Reg: ID=0x%04x\n",
             prof.device_header.regs.key(VendorID).as_a(VendorID'reg_type csco_ht_device_reg).ID);

        outf("    Status Reg: Capabilities_List=0x%x, Signaled_Target_Abort=0x%x\n",
             prof.device_header.regs.key(Status).as_a(Status'reg_type csco_ht_device_reg).Capabilities_List,
             prof.device_header.regs.key(Status).as_a(Status'reg_type csco_ht_device_reg).Signaled_Target_Abort);

        outf("    Received_Target_Abort=0x%x, Received_Master_Abort=0x%x Signaled_System_Error=0x%x\n",
             prof.device_header.regs.key(Status).as_a(Status'reg_type csco_ht_device_reg).Received_Target_Abort,
             prof.device_header.regs.key(Status).as_a(Status'reg_type csco_ht_device_reg).Received_Master_Abort,
             prof.device_header.regs.key(Status).as_a(Status'reg_type csco_ht_device_reg).Signaled_System_Error);

        outf("    Command Reg: IO_Space_Enable=0x%x, Memory_Space_Enable=0x%x\n",
             prof.device_header.regs.key(Command).as_a(Command'reg_type csco_ht_device_reg).IO_Space_Enable,
             prof.device_header.regs.key(Command).as_a(Command'reg_type csco_ht_device_reg).Memory_Space_Enable);

        outf("                 Bus_Master_Enable=0x%x, SERR_Enable=0x%x\n",
             prof.device_header.regs.key(Command).as_a(Command'reg_type csco_ht_device_reg).Bus_Master_Enable,
             prof.device_header.regs.key(Command).as_a(Command'reg_type csco_ht_device_reg).SERR_Enable);

        outf("    Class Code Reg: Code=0x%x\n",
             prof.device_header.regs.key(ClassCode).as_a(ClassCode'reg_type csco_ht_device_reg).Code);

        outf("    RevisionID Reg: ID=0x%x\n",
             prof.device_header.regs.key(RevisionID).as_a(RevisionID'reg_type csco_ht_device_reg).ID);

        outf("    BIST Reg: BIST=0x%x\n",
             prof.device_header.regs.key(BIST).as_a(BIST'reg_type csco_ht_device_reg).BIST);

        outf("    Header Type Reg: Header=0x%x\n",
             prof.device_header.regs.key(Header_Type).as_a(Header_Type'reg_type csco_ht_device_reg).Header);

        outf("    Latency Timer Reg: Timer=0x%x\n",
             prof.device_header.regs.key(Latency_Timer).as_a(Latency_Timer'reg_type csco_ht_device_reg).Timer);

        outf("    Cache Line Size Reg: Cache_Line=0x%x\n",
             prof.device_header.regs.key(Cache_Line_Size).as_a(Cache_Line_Size'reg_type csco_ht_device_reg).Cache_Line);

        outf("    BAR0 Reg: BAR=0x%04x_%04x\n",
             prof.device_header.regs.key(BAR0).as_a(BAR0'reg_type csco_ht_device_reg).bar[31:16],
             prof.device_header.regs.key(BAR0).as_a(BAR0'reg_type csco_ht_device_reg).bar[15:0]);

        outf("    BAR1 Reg: BAR=0x%04x_%04x\n",
             prof.device_header.regs.key(BAR1).as_a(BAR1'reg_type csco_ht_device_reg).bar[31:16],
             prof.device_header.regs.key(BAR1).as_a(BAR1'reg_type csco_ht_device_reg).bar[15:0]);

        outf("    BAR2 Reg: BAR=0x%04x_%04x\n",
             prof.device_header.regs.key(BAR2).as_a(BAR2'reg_type csco_ht_device_reg).bar[31:16],
             prof.device_header.regs.key(BAR2).as_a(BAR2'reg_type csco_ht_device_reg).bar[15:0]);

        outf("    BAR3 Reg: BAR=0x%04x_%04x\n",
             prof.device_header.regs.key(BAR3).as_a(BAR3'reg_type csco_ht_device_reg).bar[31:16],
             prof.device_header.regs.key(BAR3).as_a(BAR3'reg_type csco_ht_device_reg).bar[15:0]);

        outf("    BAR4 Reg: BAR=0x%04x_%04x\n",
             prof.device_header.regs.key(BAR4).as_a(BAR4'reg_type csco_ht_device_reg).bar[31:16],
             prof.device_header.regs.key(BAR4).as_a( BAR4'reg_type csco_ht_device_reg).bar[15:0]);

        outf("    BAR5 Reg: BAR=0x%04x_%04x\n",
             prof.device_header.regs.key(BAR5).as_a(BAR5'reg_type csco_ht_device_reg).bar[31:16],
             prof.device_header.regs.key(BAR5).as_a(BAR5'reg_type csco_ht_device_reg).bar[15:0]);

        outf("    CardBus CIS Pointer Reg: Pointer=0x%04x_%04x\n",
             prof.device_header.regs.key(Cardbus_CIS_Pointer).as_a(Cardbus_CIS_Pointer'reg_type csco_ht_device_reg).Pointer[31:16],
             prof.device_header.regs.key(Cardbus_CIS_Pointer).as_a(Cardbus_CIS_Pointer'reg_type csco_ht_device_reg).Pointer[15:0]);

        outf("    SubsystemID Reg: ID=0x%04x\n",
             prof.device_header.regs.key(SubSystem_ID).as_a(SubSystem_ID'reg_type csco_ht_device_reg).ID);

        outf("    Subsystem Vendor ID Reg: ID=0x%04x\n",
             prof.device_header.regs.key(SubSystem_Vendor_ID).as_a(SubSystem_Vendor_ID'reg_type csco_ht_device_reg).ID);

        outf("    Expansion ROM Base Address: ROM Address=0x%04x_%04x\n",
             prof.device_header.regs.key(Expansion_ROM_Base_Address).as_a(Expansion_ROM_Base_Address'reg_type csco_ht_device_reg).ROM_Addr[31:16],
             prof.device_header.regs.key(Expansion_ROM_Base_Address).as_a(Expansion_ROM_Base_Address'reg_type csco_ht_device_reg).ROM_Addr[15:0]);

        outf("    Cababilities Pointer Reg: Offset=0x%x\n",
             prof.device_header.regs.key(Capabilities_Pointer).as_a(Capabilities_Pointer'reg_type csco_ht_device_reg).Offset);

        outf("    Max_Lat Reg: MaxLat=0x%x\n",
             prof.device_header.regs.key(Max_Lat).as_a(Max_Lat'reg_type csco_ht_device_reg).MaxLat);

        outf("    Min Gnt Reg: MinGnt=0x%x\n",
             prof.device_header.regs.key(Min_Gnt).as_a(Min_Gnt'reg_type csco_ht_device_reg).MinGnt);

        outf("    Interrupt Pin Reg: Pin=0x%x\n",
             prof.device_header.regs.key(Interrupt_Pin).as_a(Interrupt_Pin'reg_type csco_ht_device_reg).Pin);

        outf("    Interrupt Line Reg: Line=0x%x\n\n",
             prof.device_header.regs.key(Interrupt_Line).as_a(Interrupt_Line'reg_type csco_ht_device_reg).Line);

      // check if the capability header exists.. in case something went wrong during init
        if(prof.capability_header != NULL) {
          outf("  Capability Registers\n");
          if(prof.capability_header.regs.key(Command) is a Primary'interface_type Command'reg_type csco_ht_capability_reg(tmp) ) {
            outf("    Command Reg: Base_UnitID=0x%02x, Unit_Count=0x%02x, Master_Host=0x%x\n",
                 tmp.Base_UnitID,
                 tmp.Unit_Count,
                 tmp.Master_Host);

            outf("                 Default_Direction=0x%x, Capability_Type=0x%x\n",
                 tmp.Default_Direction,
                 tmp.Capability_Type);
          } else if(prof.capability_header.regs.key(Command) is a Secondary'interface_type Command'reg_type csco_ht_capability_reg(tmp2) ){
            outf("    Command Reg: Warm_Reset=0x%02x, Double_Ended=0x%02x, Capability_Type=0x%x\n",
                 tmp2.Warm_Reset,
                 tmp2.Double_Ended,
                 tmp2.Capability_Type);
          };

          outf("    Capabilities Pointer Reg: Capability Pointer=0x%02x\n",
               prof.capability_header.regs.key(Capabilities_Pointer).as_a(Capabilities_Pointer'reg_type csco_ht_capability_reg).capability_pointer);

          outf("    CapabilityID Reg: ID=0x%02x\n",
               prof.capability_header.regs.key(Capability_ID).as_a(Capability_ID'reg_type csco_ht_capability_reg).ID);
          
          outf("    Link_Config0 Reg: MaxLinkWidthIn=0x%x, DwFcIn=0x%x\n",
               prof.capability_header.regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).MaxLinkWidthIn,
               prof.capability_header.regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).DwFcIn);

          outf("                      MaxLinkWidthOut=0x%x, DwFcOut=0x%x\n",
               prof.capability_header.regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).MaxLinkWidthOut,
               prof.capability_header.regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).DwFcOut);
          
          outf("                      LinkWidthIn=0x%x, DwFcInEn=0x%x\n",
               prof.capability_header.regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).LinkWidthIn,
               prof.capability_header.regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).DwFcInEn);
          
          outf("                      LinkWidthOut=0x%x, DwFcOutEn=0x%x\n",
               prof.capability_header.regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).LinkWidthOut,
               prof.capability_header.regs.key(Link_Config0).as_a(Link_Config0'reg_type csco_ht_capability_reg).DwFcOutEn);
          
          outf("    Link_Control0 Reg: CRC_Flood_Enable=0x%x, CRC_Start_Test=0x%x\n",
               prof.capability_header.regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).CRC_Flood_Enable,
               prof.capability_header.regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).CRC_Start_Test);
          
          outf("                       CRC_Force_Error=0x%x, Link_Failure=0x%x\n",
               prof.capability_header.regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).CRC_Force_Error,
               prof.capability_header.regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).Link_Failure);

          outf("                       Initialization_Complete=0x%x, End_of_Chain=0x%x\n",
               prof.capability_header.regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).Initialization_Complete,
               prof.capability_header.regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).End_of_Chain);

          outf("                       Transmitter_Off=0x%x, CRC_Error=0x%x\n",
               prof.capability_header.regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).Transmitter_Off,
               prof.capability_header.regs.key(Link_Control0).as_a(Link_Control0'reg_type csco_ht_capability_reg).CRC_Error);

          outf("    Link_Config1 Reg: MaxLinkWidthIn=0x%x, DwFcIn=0x%x\n",
               prof.capability_header.regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).MaxLinkWidthIn,
               prof.capability_header.regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).DwFcIn);

          outf("                      MaxLinkWidthOut=0x%x, DwFcOut=0x%x\n",
               prof.capability_header.regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).MaxLinkWidthOut,
               prof.capability_header.regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).DwFcOut);

          outf("                      LinkWidthIn=0x%x, DwFcInEn=0x%x\n",
               prof.capability_header.regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).LinkWidthIn,
               prof.capability_header.regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).DwFcInEn);

          outf("                      LinkWidthOut=0x%x, DwFcOutEn=0x%x\n",
               prof.capability_header.regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).LinkWidthOut,
               prof.capability_header.regs.key(Link_Config1).as_a(Link_Config1'reg_type csco_ht_capability_reg).DwFcOutEn);

          outf("    Link_Control1 Reg: CRC_Flood_Enable=0x%x, CRC_Start_Test=0x%x\n",
               prof.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).CRC_Flood_Enable,
               prof.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).CRC_Start_Test);

          outf("                       CRC_Force_Error=0x%x, Link_Failure=0x%x\n",
               prof.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).CRC_Force_Error,
               prof.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).Link_Failure);

          outf("                       Initialization_Complete=0x%x, End_of_Chain=0x%x\n",
               prof.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).Initialization_Complete,
               prof.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).End_of_Chain);

          outf("                       Transmitter_Off=0x%x, CRC_Error=0x%x\n",
               prof.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).Transmitter_Off,
               prof.capability_header.regs.key(Link_Control1).as_a(Link_Control1'reg_type csco_ht_capability_reg).CRC_Error);


          if(prof.capability_header.regs.key(Command) is a Primary'interface_type csco_ht_capability_reg(tmp) ) {
            outf("    Link Frequency Reg: Freq0=0x%x, Freq1=0x%x\n",
                 prof.capability_header.regs.key(Link_Freq).as_a(Link_Freq'reg_type csco_ht_capability_reg).Freq0,
                 prof.capability_header.regs.key(Link_Freq).as_a(Primary'interface_type Link_Freq'reg_type csco_ht_capability_reg).Freq1);
          } else {
            outf("    Link Frequency Reg: Freq0=0x%x\n",
                 prof.capability_header.regs.key(Link_Freq).as_a(Link_Freq'reg_type csco_ht_capability_reg).Freq0);
          };

          outf("    Link Revision ID Reg: MajorRev=0x%x, MinorRev=0x%x\n",
               prof.capability_header.regs.key(HT_Rev_ID).as_a(HT_Rev_ID'reg_type csco_ht_capability_reg).MajorRev,
               prof.capability_header.regs.key(HT_Rev_ID).as_a(HT_Rev_ID'reg_type csco_ht_capability_reg).MinorRev);
        };
      };
    };
      };
    };
  };

//  run() is also {
  activate_TCMs(reset_type : csco_ht_reset_type)@csco_ht_device_clock is also {
    start print_csco_ht_chain_config();
  };
};
'>
