////////////////////////////////////////////////////////////////////////////       // -
// Title         : HT Link Interface
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////       // -
// File          : csco_ht_link_interface.e
// Author        : Brian Van Essen  <vanessen@vanessen-u60.cisco.com>
// Created       : 25.01.2001
// Last modified : 30.04.2001
////////////////////////////////////////////////////////////////////////////       // -
// Description :
// The link interface is the lowest level of the HT model and is the part that actually interacts with the
// bus at a wire level.  It is the single point of contact for the model to interact with the DUT and is
// responsible for framing a HT packet into the appropriate number of bit times and generating/checking
// the CRC for each bit lane.
////////////////////////////////////////////////////////////////////////////////
// Modification history :
// 25.01.2001 : created
//////////////////////////////////////////////////////////////////////////////-

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****


<'

////////////////////////////////////////////////////////////////////////////////
// The link interface is the lowest level of the HT model and is the part that actually interacts with the
// bus at a wire level.  It is the single point of contact for the model to interact with the DUT and is
// responsible for framing a HT packet into the appropriate number of bit times and generating/checking
// the CRC for each bit lane.
////////////////////////////////////////////////////////////////////////////////

unit csco_ht_link_interface {
  parent_channel : csco_ht_channel_interface;
  keep parent_channel == get_enclosing_unit(csco_ht_channel_interface);
  // A reference to the CSR and general control block.
  control_block : csco_ht_device_control_block;
  keep control_block == parent_channel.control_block;
  // A number used to identify which interface this is on a given device.
  link_identifier : uint;
  // Flag used to indicate that this link actually hooks up to a verilog simulator/model
  verilog_links : bool;
  keep soft verilog_links == FALSE;
  keep sys.csco_ht_using_verilog_simulator == FALSE => verilog_links == FALSE;
  keep sys.csco_ht_using_verilog_simulator == TRUE => verilog_links == TRUE;

  // A flag used to disable proper generation of CRC's when this is set, invalid CRC's will be transmitted.
  generate_bad_CRCs : bool;
  keep soft generate_bad_CRCs == FALSE;

  // A flag used to print link level information to the user interface.
  debug_link : bool;
  keep soft debug_link == FALSE;
  // A flag that will print CRC infomation to the user interface.
  debug_crc : bool;
  keep soft debug_crc == FALSE;
  // A flag that will print initialization information to the user interface.
  debug_initialization : bool;
  keep soft debug_initialization == FALSE;
  // A flag that will cause the link interface to record every bit time that is transmitted and received.
  record_link_traffic : bool;
  keep soft record_link_traffic == FALSE;
  ////////////////////////////////////////////////////////////////////////////////
  // Clocks, Events and Initialization
  ////////////////////////////////////////////////////////////////////////////////
  // This is the primary clock event for this link interface.  It will either be a specman derived event, or driven
  // from the simulator (DUT).  The use of this signal allow all links to have an independent direct drive clock signal.
  // If the use_direct_csco_ht_clock flag is false, this clock must represent 1/12th ns so that the appropriate TX clocks
  // can be generated.  This should be supplied by the DUT. Otherwise this clock needs to be 2x as fast as the link
  // speed for each device.
  event csco_ht_link_fundamental_clock;
  on csco_ht_link_fundamental_clock {
    // Drive the csco_ht_device_fundamental_clock so that it is always atleast as fast as the fastest link clock.
    // This will provide the function of logically OR'ing the link clocks together on devices with multiple links.
    emit parent_channel.parent_device.csco_ht_device_fundamental_clock;
  };

  event csco_ht_physical_tx_clk;
  // A logical clk signal used for all higher layer views of the HT bus.  The crc cycles are removed.
  event csco_ht_tx_clk;
  event csco_ht_tx_buffer_ready;
  event csco_ht_tx_complete;
//  event csco_ht_tx_4_byte_boundary;
  event csco_ht_tx_initialization_complete;

  event csco_ht_physical_rx_clk;
  // A logical clk signal used for all higher layer views of the HT bus.  The crc cycles are removed.
  event csco_ht_rx_clk;
  event csco_ht_rx_buffer_ready;
  event csco_ht_rx_initialization_complete;

  // On the HT bus the smallest granularity of traffic is a 4 byte block (32 bits).
  // These flags indicate when these boundaries occur.  (Note: The first clock cycle after initialization is a legal boundary.)
  csco_ht_tx_4_byte_boundary : bool;
  keep csco_ht_tx_4_byte_boundary == TRUE;
  csco_ht_rx_4_byte_boundary : bool;
  keep csco_ht_rx_4_byte_boundary == TRUE;

  // Use this state bit to determine what value to drive onto the clock wires.
  csco_ht_tx_clk_state : bit;
  keep csco_ht_tx_clk_state == 0;

  on csco_ht_physical_tx_clk {
    csco_ht_tx_clk_state = ~csco_ht_tx_clk_state;
  };

  ////////////////////////////////////////////////////////////////////////////////
  // A standardized message format for informational messages.
  ////////////////////////////////////////////////////////////////////////////////
  short_name(): string is {
    if(control_block.read_link_orientation(link_identifier) == TX_Upstream_RX_Downstream_Interface) {
      result = "TX_Up_RX_Down";
    } else {
      result = "TX_Down_RX_Up";
    };
  };

  ////////////////////////////////////////////////////////////////////////////////
  // These flags indicate that the low level initialization for the tx/rx paths is complete.
  tx_link_initialized : bool;
  keep tx_link_initialized == FALSE;
  rx_link_initialized : bool;
  keep rx_link_initialized == FALSE;
  on csco_ht_tx_initialization_complete {
    tx_link_initialized = TRUE;
    messagef(LOW, "TX path initialized");
  };
  on csco_ht_rx_initialization_complete {
    rx_link_initialized = TRUE;
    messagef(LOW, "RX path initialized");
  };

  // These fields are used to determine that current link width.  Note: These are only changed at reset.
  tx_negotiated_link_width : csco_ht_link_size;
  keep tx_negotiated_link_width == control_block.read_current_tx_width(link_identifier);
  rx_negotiated_link_width : csco_ht_link_size;
  keep rx_negotiated_link_width == control_block.read_current_rx_width(link_identifier);
  // A field used to determine the current clock frequency.  Note: This is only changed at reset.
  tx_negotiated_clock_freq : uint;
  keep tx_negotiated_clock_freq == control_block.read_current_tx_clock_freq(link_identifier);
  // Defining the physical level clock based on the current clock frequency.
  event csco_ht_physical_tx_clk is only {@csco_ht_physical_tx_clk; [tx_negotiated_clock_freq]} or {@session.start_of_test} @csco_ht_link_fundamental_clock;

  // Fields used to hold the current bit time to be transmitted or received.
  // Events are used to communicate that these fields are valid and ready to be read.
  // csco_ht_tx_buffer_ready and csco_ht_rx_buffer_ready
  !tx_bit_time : csco_ht_bit_time;
  !rx_bit_time : csco_ht_bit_time;

  ////////////////////////////////////////////////////////////////////////////////
  // Fields used for interacting with the HT bus.
  ////////////////////////////////////////////////////////////////////////////////
  // These fields are used to reference the specman links used when not interacting with a verilog DUT.
  tx_link : csco_ht_link is instance ;
  rx_link : csco_ht_link is instance ;

  ////////////////////////////////////////////////////////////////////////////////
  // These are strings used to hold the names of the verilog signals that this link interface
  // will be communicating with.
  ////////////////////////////////////////////////////////////////////////////////
  // Fundamental Clock
  csco_ht_link_fundamental_clock_signal : string;
  keep soft csco_ht_link_fundamental_clock_signal == "";

  // Non Differential signals
  clk_tx_signal : string;
  keep soft clk_tx_signal == "";
  ctl_tx_signal : string;
  keep soft ctl_tx_signal == "";
  cad_tx_signal : string;
  keep soft cad_tx_signal == "";

  clk_rx_signal : string;
  keep soft clk_rx_signal == "";
  ctl_rx_signal : string;
  keep soft ctl_rx_signal == "";
  cad_rx_signal : string;
  keep soft cad_rx_signal == "";

  // Differential Signals
  clk_tx_H : string;
  keep soft clk_tx_H == "";
  clk_tx_L : string;
  keep soft clk_tx_L == "";
  ctl_tx_H : string;
  keep soft ctl_tx_H == "";
  ctl_tx_L : string;
  keep soft ctl_tx_L == "";
  cad_tx_H : string;
  keep soft cad_tx_H == "";
  cad_tx_L : string;
  keep soft cad_tx_L == "";

  clk_rx_H : string;
  keep soft clk_rx_H == "";
  clk_rx_L : string;
  keep soft clk_rx_L == "";
  ctl_rx_H : string;
  keep soft ctl_rx_H == "";
  ctl_rx_L : string;
  keep soft ctl_rx_L == "";
  cad_rx_H : string;
  keep soft cad_rx_H == "";
  cad_rx_L : string;
  keep soft cad_rx_L == "";

  ////////////////////////////////////////////////////////////////////////////////
  // Fields used for the generation and checking of the bit lane CRC's
  ////////////////////////////////////////////////////////////////////////////////
  csco_ht_tx_bit_time_count : uint (bits:9);
  keep csco_ht_tx_bit_time_count == 0;
  initial_tx_crc_window : bool;
  keep initial_tx_crc_window == TRUE;
  crc_tx_in_progress : bool;
  keep crc_tx_in_progress == FALSE;
  crc_tx_bit_time_count : uint;
  keep crc_tx_bit_time_count == 0;

  csco_ht_rx_bit_time_count : uint (bits:9);
  keep csco_ht_rx_bit_time_count == 0;
  initial_rx_crc_window : bool;
  keep initial_rx_crc_window == TRUE;
  crc_rx_in_progress : bool;
  keep crc_rx_in_progress == FALSE;
  crc_rx_bit_time_count : uint;
  keep crc_rx_bit_time_count == 0;


  // There is one crc for each byte lane.
  !current_tx_crc_result : list of csco_ht_crc_byte_lane;
  !completed_tx_crc_result : list of csco_ht_crc_byte_lane;

  !current_rx_crc_result : list of csco_ht_crc_byte_lane;
  !completed_rx_crc_result : list of csco_ht_crc_byte_lane;
  ////////////////////////////////////////////////////////////////////////////////
  // Methods used for the generation of the byte lane CRC's
  ////////////////////////////////////////////////////////////////////////////////
  on csco_ht_physical_tx_clk {
    if(!initial_tx_crc_window) {
//      if(csco_ht_tx_bit_time_count == 64) {
      if(csco_ht_tx_bit_time_count == 64 && control_block.Sync_Flood_Init == FALSE) { // If a sync flood is in progress, don't issue a CRC
        if(!crc_tx_in_progress) {
          crc_tx_in_progress = TRUE;
          crc_tx_bit_time_count = 0;
        }else if(crc_tx_in_progress && crc_tx_bit_time_count >= 4) {
          crc_tx_in_progress = FALSE;
        };
      };
    }else {
      if(csco_ht_tx_bit_time_count == 511) {
        initial_tx_crc_window = FALSE;
      };
    };

    if(!crc_tx_in_progress) {
      emit csco_ht_tx_clk;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Generate the byte lane CRC
  ////////////////////////////////////////////////////////////////////////////////
  generate_crc()@csco_ht_tx_clk is {
    var ctl_pad : bit;
    var tmp_crc_byte_lane : csco_ht_crc_byte_lane;

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_channel.parent_device.reset @parent_channel.parent_device.reset;
      };
      {
        sync @csco_ht_tx_initialization_complete;
        // Once a sync flood has started, stop generating a crc
        while(control_block.capability_register.links_active()[link_identifier] != 1 && control_block.Sync_Flood_Init == FALSE) {
          // Once the transmit buffer is ready, compute the crc.
          sync @csco_ht_tx_buffer_ready;

          for each in current_tx_crc_result {
            if(index == 0) {
              ctl_pad = tx_bit_time.ctl;
            }else {
              ctl_pad = 0;
            };
            it.crc_window_buffer.add(%{ctl_pad, tx_bit_time.cad[(index+1)*8 - 1: index*8]});
          };

          if(csco_ht_tx_bit_time_count == 511) {
            if(debug_crc) {
              messagef(HIGH, "completed TX CRC window");
            };
            for each in current_tx_crc_result {
              it.compute_crc();
            };
            completed_tx_crc_result = current_tx_crc_result;

            current_tx_crc_result = {};
            for i from 0 to ((tx_negotiated_link_width.as_a(uint) / 8) - 1) {
              tmp_crc_byte_lane = new;
              tmp_crc_byte_lane.csco_ht_poly = 32'h04c11db7;
              current_tx_crc_result.add(tmp_crc_byte_lane);
            };
          };

          csco_ht_tx_bit_time_count += 1;
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Methods used for the checking of the byte lane CRC's
  ////////////////////////////////////////////////////////////////////////////////
  on csco_ht_physical_rx_clk {
    if(!initial_rx_crc_window) {
//      if(csco_ht_rx_bit_time_count == 64) {
      if(csco_ht_rx_bit_time_count == 64 && control_block.Sync_Flood_Init == FALSE) { // If a sync flood is in progress, don't expect a CRC
        if(!crc_rx_in_progress) {
          crc_rx_in_progress = TRUE;
          crc_rx_bit_time_count = 0;
        }else if(crc_rx_in_progress && crc_rx_bit_time_count >= 4) {
          crc_rx_in_progress = FALSE;
          for each in completed_rx_crc_result {
            if(it.check_crc_match() == FALSE) {
              var error_string : string;
              error_string = appendf("CRC Error : Expected CRC=%x Received CRC=%x", it.crc, it.transmitted_crc);
              sys.csco_ht_errors.report_protocol_error(CRC_ERROR, error_string, control_block, link_identifier);
            };
          };
        };
      };
    }else {
      if(csco_ht_rx_bit_time_count == 511) {
        initial_rx_crc_window = FALSE;
      };
    };

    if(!crc_rx_in_progress) {
      emit csco_ht_rx_clk;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Verify the byte lane CRC
  ////////////////////////////////////////////////////////////////////////////////
  verify_crc()@csco_ht_rx_clk is {
    var ctl_pad : bit;
    var tmp_crc_byte_lane : csco_ht_crc_byte_lane;

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_channel.parent_device.reset @parent_channel.parent_device.reset;
      };
      {
        sync @csco_ht_rx_initialization_complete;
        // Once a sync flood has started, stop preparing to verify a crc
        while(control_block.capability_register.links_active()[link_identifier] != 1 && control_block.Sync_Flood_Init == FALSE) {
          sync @csco_ht_rx_buffer_ready;

          for each in current_rx_crc_result {
            if(index == 0) {
              ctl_pad = rx_bit_time.ctl;
            }else {
              ctl_pad = 0;
            };
            it.crc_window_buffer.add(%{ctl_pad, rx_bit_time.cad[(index+1)*8 - 1: index*8]});
          };

          if(csco_ht_rx_bit_time_count == 511) {
            if(debug_crc) {
              messagef(HIGH, "completed RX CRC window");
            };
            for each in current_rx_crc_result {
              it.compute_crc();
            };
            completed_rx_crc_result = current_rx_crc_result;

            current_rx_crc_result = {};
            for i from 0 to ((rx_negotiated_link_width.as_a(uint) / 8) - 1) {
              tmp_crc_byte_lane = new;
              tmp_crc_byte_lane.csco_ht_poly = 32'h04c11db7;
              current_rx_crc_result.add(tmp_crc_byte_lane);
            };
          };

          csco_ht_rx_bit_time_count += 1;
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Maintain the flags for the 4 byte boundaries.
  ////////////////////////////////////////////////////////////////////////////////
  csco_ht_tx_4_byte_count : uint;
  keep csco_ht_tx_4_byte_count == 0;
  !reset_tx_counter : bool;
  on csco_ht_tx_clk {
    if(tx_link_initialized) {
      if(reset_tx_counter == TRUE) {
        csco_ht_tx_4_byte_count = 0;
        reset_tx_counter = FALSE;
      }else {
        csco_ht_tx_4_byte_count += tx_negotiated_link_width.as_a(int)/8;
      };
    }else {
      csco_ht_tx_4_byte_count = 0;
    };
    if(csco_ht_tx_4_byte_count % 4 == 0) {
      csco_ht_tx_4_byte_boundary = TRUE;
    }else {
      csco_ht_tx_4_byte_boundary = FALSE;
    };
  };
  csco_ht_rx_4_byte_count : uint;
  keep csco_ht_rx_4_byte_count == 0;
  !reset_rx_counter : bool;
  on csco_ht_rx_clk {
    if(rx_link_initialized) {
      if(reset_rx_counter == TRUE) {
        csco_ht_rx_4_byte_count = 0;
        reset_rx_counter = FALSE;
      }else {
        csco_ht_rx_4_byte_count += rx_negotiated_link_width.as_a(int)/8;
      };
    }else {
      csco_ht_rx_4_byte_count = 0;
    };
    if(csco_ht_rx_4_byte_count % 4 == 0) {
      csco_ht_rx_4_byte_boundary = TRUE;
    }else {
      csco_ht_rx_4_byte_boundary = FALSE;
    };
  };
  reset_tx_clock_counters() is {
    messagef(HIGH, "Reseting the TX Clock Counters");
    reset_tx_counter = TRUE;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // csco_ht_tx and csco_ht_rx are the core methods that are responsible for the actual transmission
  // and reception of bit times on the links.
  ////////////////////////////////////////////////////////////////////////////////
  // Use a flag to enable packet logging/history
  !csco_ht_tx_history : list of csco_ht_bit_time_record;
  !csco_ht_rx_history : list of csco_ht_bit_time_record;

  // Two methods that are used to allow the interaction with verilog or specman links.
  tx_data(temp_tx_bit_time : csco_ht_bit_time)@csco_ht_link_fundamental_clock is empty;
  rx_data():csco_ht_bit_time is empty;
  ////////////////////////////////////////////////////////////////////////////////
  csco_ht_tx()@csco_ht_physical_tx_clk is {
    // Wait until the transmit side has had a chance to prepare the transmit buffer.
    var temp_tx_bit_time : csco_ht_bit_time;

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_channel.parent_device.reset @parent_channel.parent_device.reset;
      };
      {
        while(control_block.capability_register.links_active()[link_identifier] != 1) {
          if(crc_tx_in_progress) {
            temp_tx_bit_time = new;
            temp_tx_bit_time.link_width = tx_negotiated_link_width;
            temp_tx_bit_time.ctl = 1;
            for i from 0 to (temp_tx_bit_time.link_width.as_a(uint)/8)-1 do {
              temp_tx_bit_time.clk.add(csco_ht_tx_clk_state);
            };

            if(generate_bad_CRCs == FALSE) { // If the generate_bad_CRCs flag is true, then the CAD bits will not be properly set.
              // Set up the proper data for the CAD bits.
              for each (byte_lane) in completed_tx_crc_result {
                temp_tx_bit_time.cad[(index+1)*8 - 1: index*8] = byte_lane.crc[(crc_tx_bit_time_count+1)*8 - 1: crc_tx_bit_time_count*8];
              };
            };
            if(debug_crc) {
              var message : string;
              messagef(FULL, "TX crc bit time %d, CAD=0x%x",
                       csco_ht_tx_bit_time_count,
                       temp_tx_bit_time.cad[temp_tx_bit_time.link_width.as_a(int)-1:0]);
            };
            crc_tx_bit_time_count += 1;
          }else {
            sync @csco_ht_tx_buffer_ready;
            temp_tx_bit_time = tx_bit_time;
          };

          // Transmit the data
          tx_data(temp_tx_bit_time);
          // Add to transmit history
          if(record_link_traffic && tx_link_initialized) {
            var new_bit_time_record : csco_ht_bit_time_record = new;
            new_bit_time_record.bit_time = temp_tx_bit_time;
            new_bit_time_record.time = sys.time;
            csco_ht_tx_history.add(new_bit_time_record);
          };
          if((debug_link && tx_link_initialized) || (debug_crc && crc_tx_in_progress) || debug_initialization) {
            var message : string;
            messagef(FULL, "TX Port Clk=%b Ctl=%b CAD=%x",
                     temp_tx_bit_time.clk[(temp_tx_bit_time.link_width.as_a(int)/8)-1:0],
                     temp_tx_bit_time.ctl, temp_tx_bit_time.cad[temp_tx_bit_time.link_width.as_a(int)-1:0]);
          };
          emit csco_ht_tx_complete;
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  csco_ht_rx()@csco_ht_physical_rx_clk is {
    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_channel.parent_device.reset @parent_channel.parent_device.reset;
      };
      {
        while(control_block.capability_register.links_active()[link_identifier] != 1) {
          rx_bit_time = rx_data();
          // Add to transmit history
          if(record_link_traffic && rx_link_initialized) {
            var new_bit_time_record : csco_ht_bit_time_record = new;
            new_bit_time_record.bit_time = rx_bit_time;
            new_bit_time_record.time = sys.time;

            csco_ht_rx_history.add(new_bit_time_record);
          };
          if((debug_link && rx_link_initialized) || (debug_crc && crc_rx_in_progress) || debug_initialization) {
            var message : string;
            messagef(FULL, "RX Port Clk=%b Ctl=%b CAD=%x",
                     rx_bit_time.clk[(rx_bit_time.link_width.as_a(int)/8)-1:0],
                     rx_bit_time.ctl,
                     rx_bit_time.cad[rx_bit_time.link_width.as_a(int)-1:0]);
          };
          if(crc_rx_in_progress) {
            if(debug_crc) {
              var message : string;
              messagef(FULL, "RX crc bit time %d, CAD=0x%x",
                       csco_ht_rx_bit_time_count,
                       rx_bit_time.cad[rx_bit_time.link_width.as_a(int)-1:0]);
            };
            for each (byte_lane) in completed_rx_crc_result {
              byte_lane.transmitted_crc[(crc_rx_bit_time_count+1)*8 - 1: crc_rx_bit_time_count*8] = rx_bit_time.cad[(index+1)*8 - 1: index*8];
            };
            crc_rx_bit_time_count += 1;
          }else {
            emit csco_ht_rx_buffer_ready;
          };
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Special methods that allow the interaction with verilog or 'e' specman links.
  ////////////////////////////////////////////////////////////////////////////////
  // Using an 'e' link structure.
  ////////////////////////////////////////////////////////////////////////////////
  when FALSE'verilog_links csco_ht_link_interface {
    // Define the source of the fundamental link clock.
    event csco_ht_link_fundamental_clock is only {@csco_ht_link_fundamental_clock; [2]} or {@session.start_of_test} @sys.any;
    // Define the incomming rx clock
    event csco_ht_physical_rx_clk is only change(rx_link.clk[0]);

    // Use 'e' objects for the passing of data.
    tx_data(temp_tx_bit_time : csco_ht_bit_time)@csco_ht_link_fundamental_clock is only {
      tx_link.ctl = temp_tx_bit_time.ctl;
      tx_link.cad = temp_tx_bit_time.cad;
      wait [tx_negotiated_clock_freq/2];
      tx_link.clk = temp_tx_bit_time.clk;
    };
    rx_data():csco_ht_bit_time is only {
      var temp_rx_bit_time : csco_ht_bit_time = new;
      temp_rx_bit_time.link_width = rx_negotiated_link_width;
      temp_rx_bit_time.clk = rx_link.clk;
      temp_rx_bit_time.ctl = rx_link.ctl;
      temp_rx_bit_time.cad = rx_link.cad;
      result = temp_rx_bit_time;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Using a verilog link structure.
  ////////////////////////////////////////////////////////////////////////////////
  when TRUE'verilog_links csco_ht_link_interface {
    // Define the source of the fundamental link clock.
    event csco_ht_link_fundamental_clock is only change('(csco_ht_link_fundamental_clock_signal)')@sim;
    // Allow the use of differential or non differential signals and bussed or non bussed signals.
    use_differential_bit_channels : bool;
    keep soft use_differential_bit_channels == TRUE;
    ////////////////////////////////////////////////////////////////////////////////
    when TRUE'use_differential_bit_channels csco_ht_link_interface {
      // Derive the rx clock off of the first clk signal of the link.  In actual hardware you would want
      // to have a different clock for each byte lane.  This is simplified for this model.
      event csco_ht_physical_rx_clk is only change('(clk_rx_H)[0]')@sim;
      tx_data(temp_tx_bit_time : csco_ht_bit_time)@csco_ht_link_fundamental_clock is only {
        '(ctl_tx_H)' = temp_tx_bit_time.ctl;
        '(ctl_tx_L)' = ~temp_tx_bit_time.ctl;
        '(cad_tx_H)' = temp_tx_bit_time.cad;
        '(cad_tx_L)' = ~temp_tx_bit_time.cad;
        wait [tx_negotiated_clock_freq/2];
        '(clk_tx_H)' = temp_tx_bit_time.clk[:];
        '(clk_tx_L)' = ~temp_tx_bit_time.clk[:];
      };
      rx_data():csco_ht_bit_time is only {
        result = new;
        result.link_width = me.rx_negotiated_link_width;
        result.ctl = '(ctl_rx_H)';
        result.clk = '(clk_rx_H)';
        result.cad[result.link_width.as_a(uint)-1:0] = '(cad_rx_H)';
      };
    };
    ////////////////////////////////////////////////////////////////////////////////
    when FALSE'use_differential_bit_channels csco_ht_link_interface {
      // Derive the rx clock off of the first clk signal of the link.  In actual hardware you would want
      // to have a different clock for each byte lane.  This is simplified for this model.
      event csco_ht_physical_rx_clk is only change('(clk_rx_signal)')@sim;
      tx_data(temp_tx_bit_time : csco_ht_bit_time)@csco_ht_link_fundamental_clock is only {
        '(ctl_tx_signal)' = temp_tx_bit_time.ctl;
        '(cad_tx_signal)' = temp_tx_bit_time.cad;
        wait [tx_negotiated_clock_freq/2];
        '(clk_tx_signal)' = temp_tx_bit_time.clk[:];
      };
      rx_data():csco_ht_bit_time is only {
        result = new;
        result.link_width = me.rx_negotiated_link_width;
        result.clk = '(clk_rx_signal)';
        result.ctl = '(ctl_rx_signal)';
        result.cad = '(cad_rx_signal)';
      };
    };
    ////////////////////////////////////////////////////////////////////////////////
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Methods used for the initialization of the link interface.
  ////////////////////////////////////////////////////////////////////////////////
  // Before the initialization sequencing begins a transmitter needs to drive CLK, a 0 CTL bit and 1's on the CAD.
  ////////////////////////////////////////////////////////////////////////////////
  pre_reset_tx_path_initialization()@csco_ht_physical_tx_clk is {
    var temp_tx_bit_time : csco_ht_bit_time;
    var max_tx_width : csco_ht_link_size;
    max_tx_width = control_block.read_max_tx_width(link_identifier);

    while(parent_channel.parent_device.RESET_L == TRUE || parent_channel.parent_device.PWROK == FALSE) {
      temp_tx_bit_time = new;
      temp_tx_bit_time.link_width = max_tx_width;
      temp_tx_bit_time.ctl = 0;
      for i from 0 to (max_tx_width.as_a(uint)/8)-1 do {
        temp_tx_bit_time.clk.add(csco_ht_tx_clk_state);
      };
      temp_tx_bit_time.cad = ((1 << max_tx_width.as_a(uint)) - 1);

      // Transmit the data
      tx_data(temp_tx_bit_time);
      wait cycle;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Initialize all of the appropriate variables in the tx path. Start the transmitter.
  ////////////////////////////////////////////////////////////////////////////////
  initialize_tx_path()@csco_ht_physical_tx_clk is {
    var tmp_crc_byte_lane : csco_ht_crc_byte_lane;

    // Initialize all of the relevant variables.
    tx_negotiated_link_width = control_block.read_current_tx_width(link_identifier);
    tx_negotiated_clock_freq = control_block.read_current_tx_clock_freq(link_identifier);

    csco_ht_tx_bit_time_count = 0;
    initial_tx_crc_window = TRUE;
    crc_tx_in_progress = FALSE;
    crc_tx_bit_time_count = 0;

    current_tx_crc_result = {};
    for i from 0 to ((tx_negotiated_link_width.as_a(uint) / 8) - 1) {
      tmp_crc_byte_lane = new;
      tmp_crc_byte_lane.csco_ht_poly = 32'h04c11db7;
      current_tx_crc_result.add(tmp_crc_byte_lane);
    };

    completed_tx_crc_result = {};

    csco_ht_tx_4_byte_count = 0;

    csco_ht_tx_history = {};

    start csco_ht_tx();
    start generate_crc();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Initialize all of the appropriate variables in the rx path. Start the receiver.
  ////////////////////////////////////////////////////////////////////////////////
  initialize_rx_path()@csco_ht_physical_rx_clk is {
    var tmp_crc_byte_lane : csco_ht_crc_byte_lane;
    rx_negotiated_link_width = control_block.read_current_rx_width(link_identifier);

    csco_ht_rx_bit_time_count = 0;
    initial_rx_crc_window = TRUE;
    crc_rx_in_progress = FALSE;
    crc_rx_bit_time_count = 0;

    current_rx_crc_result = {};
    for i from 0 to ((rx_negotiated_link_width.as_a(uint) / 8) - 1) {
      tmp_crc_byte_lane = new;
      tmp_crc_byte_lane.csco_ht_poly = 32'h04c11db7;
      current_rx_crc_result.add(tmp_crc_byte_lane);
    };

    completed_rx_crc_result = {};

    csco_ht_rx_4_byte_count = 0;

    csco_ht_rx_history = {};

    start csco_ht_rx();
    start verify_crc();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Once both of the tx and rx paths have been initialized set the link_initialized bit in the Link Control CSR.
  ////////////////////////////////////////////////////////////////////////////////
  initialize_path()@csco_ht_physical_tx_clk is {
    all of {
      {
        wait @csco_ht_tx_initialization_complete;
      };
      {
        wait @csco_ht_rx_initialization_complete;
      };
    };
    // Once both of the initialization signals have been received set the Initialization Complete bit in the control block
    control_block.link_initialization_complete(link_identifier);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // reset()
  ////////////////////////////////////////////////////////////////////////////////
  reset() is {
    tx_link_initialized = FALSE;
    rx_link_initialized = FALSE;
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// A structure used to represent a single bit time on a link, independent of the link width.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_bit_time {
  link_width : csco_ht_link_size;
  clk : list of bit;
  keep clk.size() == link_width.as_a(int)/8;
  ctl : bit;
  cad : uint;
};
////////////////////////////////////////////////////////////////////////////////
// A structure used to manage the running 512 byte CRC window for each byte lane.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_crc_byte_lane {
  !crc : uint;
  !crc_window_buffer : list of bit;
  !transmitted_crc : uint;

  csco_ht_poly : uint;
  keep csco_ht_poly == 32'h04c11db7;

  compute_crc() is {
    // Seed it with 1's
    crc = 32'hFFFFFFFF;
    for i from 0 to 511 do {
      crc = compute_csco_ht_crc(crc_window_buffer[9*(i+1)-1:9*i], crc);
    };
    // Invert the crc before sending it out, or comparison
    crc = ~crc;
  };

  compute_csco_ht_crc(data : uint, crc : uint):uint is {
    var tmp : bit;

    for i from 0 to 8 do {
      tmp = crc >> 31;
      crc = (crc << 1) | ((data >> i) & 1);
      crc = (tmp==1) ? crc ^ csco_ht_poly: crc;
    };
    result = crc;
  };

  check_crc_match():bool is {
    if(crc == transmitted_crc) {
      result = TRUE;
    }else {
      result = FALSE;
    };
  };
};
////////////////////////////////////////////////////////////////////////////////
// A structure used to record a given bit time and the associated simulator time.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_bit_time_record {
  bit_time : csco_ht_bit_time;
  time : uint;
};
////////////////////////////////////////////////////////////////////////////////
'>
