//-----------------------------------------------------------------------------
// Title         : HT Packet Definition
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////       // -
// File          : csco_ht_packet_definition.e
// Author        : Brian Van Essen  <vanessen@cisco.com>
// Created       : 25.01.2001
// Last modified : 01.05.2001
////////////////////////////////////////////////////////////////////////////       // -
// Description :
// The definition of an HT packet.
// This definition is HT 1.0 compliant and allows a NonPosted Broadcast packet.
// The enumerated types used by this file are defined in csco_ht_declarations.e
////////////////////////////////////////////////////////////////////////////////
// Modification history :
// 25.01.2001 : created
// 18.07.2001 : Added support for Interrupt Packets
// 04.22.2002 : fixed Obsolete field access syntax i.e. ' casting. for Specman 4.0
////////////////////////////////////////////////////////////////////////////       // -

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****

<'
////////////////////////////////////////////////////////////////////////////////
// The fundamental HT packet
////////////////////////////////////////////////////////////////////////////////
type channel_instance_e : [uncovered];


struct csco_ht_packet {
  cover_instance : channel_instance_e;
    keep soft cover_instance == uncovered;

  // The HT command
  command : csco_ht_packet_encoding;

  packet_type : csco_ht_packet_type;

  // Packet Attribute Fields
  vChan : csco_ht_virtual_channel;
  passing_rule : csco_ht_passing_rules;

  // The length of the packet is indicated by the number of bytes in the packet.  However, each
  // packet must be a multiple of a double word. (i.e. increments of 4 bytes.)
  // Number of bytes in the packet.
  packet_length : uint;
  !packed_data : list of bit;
  packet_string : string;

  // Use these fields to manage the scheduling and transmission of each packet.
  !scheduling_index : uint;
  !weight : uint;
  stall : bool;
  keep stall == FALSE;

  // packing attribute to support different endian policies
  // (little endian, big endian, big endian w/ match bit lane)
  endian_type : csco_ht_endian_type;
  keep soft endian_type == little_endian;

  unpack_packet() is empty;

  // coverage events....
  event rx_pkt_cov_event;
  event tx_pkt_cov_event;
  event rd_tx_pkt_cov_event;
  event wr_tx_pkt_cov_event;
  event bc_tx_pkt_cov_event;

  ////////////////////////////////////////////////////////////////////////////////
  // These are methods which are used to extract information from the various types of HT packets.
  // If they are called on a packet type without that field they will throw a DUT error.
  ////////////////////////////////////////////////////////////////////////////////
  // Used to determine what UnitID this packet is associated with.
  UnitID():uint (bits:5) is {
    dut_error("UnitID() was called on a packet that does not have a UnitID field.");
  };
  // Used to determine what IO Stream this packet is associate with.  This takes into account the bridge
  // bit, and hence the up or downstream direction of this packet.
  IOStream():uint (bits:5) is {
    result = UnitID();
  };
  SeqID():uint (bits:4) is {
    dut_error("SeqID() was called on a packet that does not have a SeqID field.");
  };
  BridgeBit():bit is {
    dut_error("BridgeBit() was called on a packet other than a response.");
  };
  CompatBit():bit is {
    dut_error("CompatBit() was called on a packet that does not contain a compat bit.");
  };
  Mask_Count():uint (bits:4) is {
    dut_error("Mask_Count was called on a packet that does not contain a Mask/Count field.");
  };
  Address():uint (bits:40) is {
    dut_error("Address() was called on a packet that does not contain an address.");
  };
  DataType():csco_ht_packet_encoding is {
    dut_error("DataType() was called on a packet other than a Read or a Write.");
  };
  // Returns the number of bytes in the data packet.
  DataLength():uint is {
    result = (Mask_Count() + 1) * 4;
  };
  SrcTag():uint (bits:5) is {
    dut_error("SrcTag() was called an a packet with no SrcTag field.");
  };
  ErrorFlag():bit is {
    dut_error("Error() was called on a packet other than a RdResponse or a TgtDone.");
  };
  NXA():bit is {
    dut_error("NXA() was called on a packet other than a RdResponse or a TgtDone.");
  };
  attach_data_packet(new_data : csco_ht_packet) is {
    dut_error("attach_data_packet() was called on a ctl packet with no associated Data Packet.");
  };
  retrieve_data_packet():csco_ht_packet is {
    dut_error("retrieve_data_packet() was called on a packet with no associated Data Packet.");
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Methods for Nop's
  ////////////////////////////////////////////////////////////////////////////////
  PostCmd():uint (bits:2) is {
    dut_error("PostCmd() was called on a packet other than a Nop.");
  };
  PostData():uint (bits:2) is {
    dut_error("PostData() was called on a packet other than a Nop.");
  };
  NonPostCmd():uint (bits:2) is {
    dut_error("NonPostCmd() was called on a packet other than a Nop.");
  };
  NonPostData():uint (bits:2) is {
    dut_error("NonPostData() was called on a packet other than a Nop.");
  };
  Response():uint (bits:2) is {
    dut_error("Response() was called on a packet other than a Nop.");
  };
  ResponseData():uint (bits:2) is {
    dut_error("ResponseData() was called on a packet other than a Nop.");
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Methods for interrupt packets
  ////////////////////////////////////////////////////////////////////////////////
  MessageType():csco_ht_interrupt_type is {
    dut_error("MessageType() was called on a packet other than a Interrupt.");
  };
  TriggerMode():csco_ht_interrupt_mode is {
    dut_error("TriggerMode() was called on a packet other than a Interrupt.");
  };
  DestinationMode():csco_ht_interrupt_destination is {
    dut_error("DestinationMode() was called on a packet other than a Interrupt.");
  };
  IntrDest():byte is {
    dut_error("IntrDest() was called on a packet other than a Interrupt.");
  };
  Vector():byte is {
    dut_error("Vector() was called on a packet other than a Interrupt.");
  };
  ////////////////////////////////////////////////////////////////////////////////
  // These are common methods that are valid for (almost) all packet types.
  ////////////////////////////////////////////////////////////////////////////////
  print_csco_ht_packet() is {
    outf("%s\n", print_csco_ht_packet_string());
  };
  print_csco_ht_packet_string():string is {
    result = packet_string;
  };
  create_csco_ht_packet_string() is empty;
  command():csco_ht_packet_encoding is {
    result = command;
  };
  vChan():csco_ht_virtual_channel is {
    if(vChan == NA) {
      dut_error("vChan() was called on a packet that does not belong in a normal VC.");
    };
    result = vChan;
  };
  PassPW():bool is {
    if(passing_rule == NA) {
      dut_error("PassPW() was called on a packet that does not have a PassPW field.");
    };
    case(passing_rule) {
      NA: {
        result = FALSE;
      };
      NoPassPW: {
        result = FALSE;
      };
      PassPW: {
        result = TRUE;
      };
    };
  };
  // Generate a token for a given packet.
  generate_token(): uint (bits:10) is {
    if(me.vChan() == Response) {
      // Check to see if the packet is a response.  If it is, the token may have to be generated differently.
      if(me.BridgeBit() == 1) {
        // This is a response from a Host Bridge so the UnitID is the source of the request.
        // Hence the UnitID is used to compute the transaction token.
        result = %{me.UnitID(), me.SrcTag()};
      }else {
        // Otherwise this is a response from an IO Device and the UnitID is the target of the request.
        // Thus the UnitID of 0 should be used to compute the transaction token to align with the
        // requestors (Host Bridge) UnitID.
        result = %{5'b00000, me.SrcTag()};
      };
    }else {
      result = %{me.UnitID(), me.SrcTag()};
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Packet Definitions
  ////////////////////////////////////////////////////////////////////////////////
  // Control Packets
  // When a packet is a control packet then the ctl signal should be active.
  ////////////////////////////////////////////////////////////////////////////////
  when Nop'command csco_ht_packet {
    keep packet_type == Info;
    keep vChan == NA;
    keep passing_rule == NA;

    keep me.packet_length == 4;
    Cmd              : uint (bits:6);
    keep me.Cmd == 0;
    DisCon           : bit;
    keep soft me.DisCon == 0;
    PostCmd          : uint (bits:2);
    PostData         : uint (bits:2);
    Response         : uint (bits:2);
    ResponseData     : uint (bits:2);
    NonPostCmd       : uint (bits:2);
    NonPostData      : uint (bits:2);
    Diag             : bit;
    keep soft me.Diag == 0;
    Zero             : bit;
    keep me.Zero == 0;
    Rsv              : uint (bits:11);
    keep soft me.Rsv == 0;

    PostCmd():uint (bits:2) is only {
      result = PostCmd;
    };
    PostData():uint (bits:2) is only {
      result = PostData;
    };
    NonPostCmd():uint (bits:2) is only {
      result = NonPostCmd;
    };
    NonPostData():uint (bits:2) is only {
      result = NonPostData;
    };
    Response():uint (bits:2) is only {
      result = Response;
    };
    ResponseData():uint (bits:2) is only {
      result = ResponseData;
    };
    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : Buffer Releases : PostCmd=0d%d PostData=0d%d Response=0d%d ResponseData=0d%d NonPostCmd=0d%d NonPostData=0d%d",
                              command, PostCmd, PostData, Response, ResponseData, NonPostCmd, NonPostData);
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, DisCon, Rsv[0:0], PostCmd, PostData, Response, ResponseData, NonPostCmd, NonPostData,
                         Rsv[2:1], Diag, Zero, Rsv[10:3]);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, DisCon, Rsv[0:0], PostCmd, PostData, Response, ResponseData, NonPostCmd, NonPostData,
             Rsv[2:1], Diag, Zero, Rsv[10:3]);
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Flush'command csco_ht_packet {
    keep packet_type == Ctl;
    keep vChan == NonPosted;
    keep passing_rule == NoPassPW or passing_rule == PassPW;
    // Typically a flush has the PassPW bit clear to push all request ahead of it.
    // A set PassPW bit is legal but the result is unpredicatable.
    keep soft passing_rule == NoPassPW;

    keep me.packet_length == 4;
    Cmd              : uint (bits:6);
    keep me.Cmd == 6'b000010;
    SeqID            : uint (bits:4);
    UnitID           : uint (bits:5);
    PassPW           : bit;
    keep passing_rule == NoPassPW => me.PassPW == 0;
    keep passing_rule == PassPW => me.PassPW == 1;
    SrcTag           : uint (bits:5);
    Rsv              : uint (bits:11);
    keep soft me.Rsv == 0;

    UnitID():uint (bits:5) is only {
      result = UnitID;
    };
    SeqID():uint (bits:4) is only {
      result = SeqID;
    };
    SrcTag():uint (bits:5) is only {
      result = SrcTag;
    };
    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : SeqID=0x%0x UnitID=0x%0x PassPW=0x%0x SrcTag=0x%0x",
                              command, SeqID, UnitID, PassPW, SrcTag);
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, SrcTag, Rsv);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, SrcTag, Rsv);
      if(PassPW == 0) {
        passing_rule = NoPassPW;
      }else {
        passing_rule = PassPW;
      };
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Write'command csco_ht_packet {
    // Packet Attribute Fields
    keep packet_type == Ctl;
    keep vChan == NonPosted or vChan == Posted;
    keep passing_rule == NoPassPW or passing_rule == PassPW;
    data_type : csco_ht_data_type;
    BW_latency_req : csco_ht_BW_latency_req;
    access_type : csco_ht_mem_access_type;

    // Support for interrupt packets
    interrupt : bool;
//    keep soft interrupt == FALSE;
    keep Addr[39:24] == 16'hFDF8 => interrupt == TRUE; // Interrupt space in System Address Map
    keep Addr[39:24] != 16'hFDF8 => interrupt == FALSE;
    keep interrupt == TRUE => vChan == Posted;
    keep interrupt == TRUE => data_type == Byte;

    keep me.packet_length == 8;
    Cmd              : uint (bits:6);
    keep vChan == NonPosted => me.Cmd[5:5] == 1'b0;
    keep vChan == Posted => me.Cmd[5:5] == 1'b1;
    keep me.Cmd[4:3] == 2'b01;
    keep data_type == Byte => me.Cmd[2:2] == 1'b0;
    keep data_type == DoubleWord => me.Cmd[2:2] == 1'b1;
    keep BW_latency_req == Normal => me.Cmd[1:1] == 1'b0;
    keep BW_latency_req == Isochronous => me.Cmd[1:1] == 1'b1;
    keep access_type == NonCoherent => me.Cmd[0:0] == 1'b0;
    keep access_type == Coherent => me.Cmd[0:0] == 1'b1;
    SeqID            : uint (bits:4);
    UnitID           : uint (bits:5);
    PassPW           : bit;
    keep passing_rule == NoPassPW => me.PassPW == 0;
    keep passing_rule == PassPW => me.PassPW == 1;
    SrcTag_Rsv       : uint (bits:5);
    // The SrcTag_Rsv field is reserved for Posted Writes.
    keep vChan == Posted => SrcTag_Rsv == 5'b00000;
    Compat           : bit;
    keep interrupt == TRUE => Compat == 1'b0;
    Mask_Count       : uint (bits:4);
    keep gen (Addr) before (Mask_Count);
    // If a Byte Write is performed then the count is from 0 to 8 indicating 1 to 9 doublewords of data & mask.
    keep data_type == Byte => Mask_Count in [0..8];
    keep interrupt == FALSE && data_type == Byte => Mask_Count in [1..8]; // Non interrupt packets must have data
    keep interrupt == TRUE && data_type == Byte => Mask_Count == 0; // Interrupt packets only have the byte mask
    // Ensure that a Write does not cross a 32 byte boundary on a Byte transaction and a 64 byte boundary on a DoubleWord transaction.
    keep data_type == Byte => Mask_Count <= 4'b1000 - Addr[4:2];
    keep data_type == DoubleWord => Mask_Count < 5'b10000 - Addr[5:2];
    Addr             : uint (bits:40);
    keep Addr[1:0] == 2'b00;
    // This is a pointer to the associated data packet.
    assoc_data : csco_ht_packet;
    keep data_type == Byte => assoc_data.command == Byte_Write;
    keep data_type == DoubleWord => assoc_data.command == DoubleWord_Write;
    keep assoc_data.vChan == vChan;
    keep assoc_data.packet_length == (Mask_Count + 1) * 4;
    keep assoc_data.endian_type == endian_type;

    MT               : csco_ht_interrupt_type;
    keep MT != EOI;
    TM               : csco_ht_interrupt_mode;
    // Non-Vectored interrupts are always edge-triggered.
    keep me.MT == SMI or me.MT == NMI or me.MT == INIT or me.MT == Startup or me.MT == Ext_Int => me.TM == Edge;
    DM               : csco_ht_interrupt_destination;
    IntrDest         : byte;
    Vector           : byte;
    keep interrupt == TRUE => Addr[39:24] == 16'hFDF8; // Interrupt space in System Address Map
    // Map the interrupt specific fields onto the Addr.
    keep interrupt == TRUE => Addr[23:16] == Vector;
    keep interrupt == TRUE => Addr[15:8] == IntrDest;
    keep interrupt == TRUE => Addr[7:7] == 1'b0;
    keep interrupt == TRUE => Addr[6:6] == DM.as_a(bit);
    keep interrupt == TRUE => Addr[5:5] == TM.as_a(bit);
    keep interrupt == TRUE => Addr[4:2] == MT.as_a(uint (bits:3));
    keep interrupt == TRUE and (assoc_data is a Byte_Write'command csco_ht_packet(tmp_assoc_data)) == TRUE => tmp_assoc_data.Mask== 32'h0000_0000;
    //KMM keep interrupt == TRUE => assoc_data.Byte_Write'command'Mask == 32'h0000_0000;


    UnitID():uint (bits:5) is only {
      result = UnitID;
    };
    SeqID():uint (bits:4) is only {
      result = SeqID;
    };
    CompatBit():bit is only {
      result = Compat;
    };
    Mask_Count():uint (bits:4) is only {
      result = Mask_Count;
    };
    Address():uint (bits:40) is only {
      result = Addr;
    };
    DataType():csco_ht_packet_encoding is only {
      if(Cmd[2:2] == 0) {
        result = Byte_Write;
      }else {
        result = DoubleWord_Write;
      };
    };
    SrcTag():uint (bits:5) is only {
      if(Cmd[5:5] == 0) {
        result = SrcTag_Rsv;
      }else {
        dut_error("SrcTag() was called an a Posted Sized Write, this field is reserved.");
      };
    };
    MessageType():csco_ht_interrupt_type is only {
      if(interrupt == TRUE) {
        result = MT;
      }else {
        dut_error("MessageType() was called on a Write packet that is not an interrupt.");
      };
    };
    TriggerMode():csco_ht_interrupt_mode is only {
      if(interrupt == TRUE) {
        result = TM;
      }else {
        dut_error("TriggerMode() was called on a Write packet that is not an interrupt.");
      };
    };
    DestinationMode():csco_ht_interrupt_destination is {
      if(interrupt == TRUE) {
        result = DM;
      }else {
        dut_error("DestinationMode() was called on a Write packet that is not an interrupt.");
      };
    };
    IntrDest():byte is {
      if(interrupt == TRUE) {
        result = IntrDest;
      }else {
        dut_error("IntrDest() was called on a Write packet that is not an interrupt.");
      };
    };
    Vector():byte is {
      if(interrupt == TRUE) {
        result = Vector;
      }else {
        dut_error("Vector() was called on a Write packet that is not an interrupt.");
      };
    };
    attach_data_packet(new_data : csco_ht_packet) is only {
      assoc_data = new_data;
    };
    retrieve_data_packet():csco_ht_packet is only {
      result = assoc_data;
    };
    create_csco_ht_packet_string() is {
      if(interrupt == TRUE) {
        packet_string = appendf("%s : %s %s %s %s : SeqID=0x%0x UnitID=0x%0x PassPW=0x%0x MT=0x%0x TM=0x%0x DM=0x%0x IntrDest=0x%0x Vector=0x%0x Addr=0x%010x",
                                command, vChan, data_type, BW_latency_req, access_type, SeqID, UnitID, PassPW, MT, TM, DM, IntrDest, Vector, Addr);
      }else {
        packet_string = appendf("%s : %s %s %s %s : SeqID=0x%0x UnitID=0x%0x PassPW=0x%0x SrcTag=0x%0x Compat=0x%0x Mask_Count=0x%0x Addr=0x%010x",
                                command, vChan, data_type, BW_latency_req, access_type, SeqID, UnitID, PassPW, SrcTag_Rsv, Compat, Mask_Count, Addr);
      };
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, SrcTag_Rsv, Compat, Mask_Count, Addr[39:2]);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, SrcTag_Rsv, Compat, Mask_Count, Addr[39:2]);
      if(Addr[39:24] == 16'hFDF8) {
        interrupt = TRUE; // Interrupt space in System Address Map
      };
      Addr[1:0] = 2'b00;
      if(interrupt == TRUE) {
        // Map the interrupt specific fields onto the Addr.
        Vector = Addr[23:16];
        IntrDest = Addr[15:8];
        DM = Addr[6:6].as_a(csco_ht_interrupt_destination);
        TM = Addr[5:5].as_a(csco_ht_interrupt_mode);
        MT = Addr[4:2].as_a(csco_ht_interrupt_type);
      };
      if(Cmd[5:5] == 0) {
        vChan = NonPosted;
      }else {
        vChan = Posted;
      };
      if(Cmd[2:2] == 0) {
        data_type = Byte;
      }else {
        data_type = DoubleWord;
      };
      if(Cmd[1:1] == 0) {
        BW_latency_req = Normal;
      }else {
        BW_latency_req = Isochronous;
      };
      if(Cmd[0:0] == 0) {
        access_type = NonCoherent;
      }else {
        access_type = Coherent;
      };
      if(PassPW == 0) {
        passing_rule = NoPassPW;
      }else {
        passing_rule = PassPW;
      };
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Read'command csco_ht_packet {
    // Packet Attribute Fields
    keep packet_type == Ctl;
    keep vChan == NonPosted;
    keep passing_rule == NoPassPW or passing_rule == PassPW;
    response_passing_rule : csco_ht_passing_rules;
    keep response_passing_rule == NoPassPW or response_passing_rule == PassPW;
    data_type : csco_ht_data_type;
    BW_latency_req : csco_ht_BW_latency_req;
    access_type : csco_ht_mem_access_type;

    keep me.packet_length == 8;
    Cmd              : uint (bits:6);
    keep me.Cmd[5:4] == 2'b01;
    keep response_passing_rule == NoPassPW => me.Cmd[3:3] == 1'b0;
    keep response_passing_rule == PassPW => me.Cmd[3:3] == 1'b1;
    keep data_type == Byte => me.Cmd[2:2] == 1'b0;
    keep data_type == DoubleWord => me.Cmd[2:2] == 1'b1;
    keep BW_latency_req == Normal => me.Cmd[1:1] == 1'b0;
    keep BW_latency_req == Isochronous => me.Cmd[1:1] == 1'b1;
    keep access_type == NonCoherent => me.Cmd[0:0] == 1'b0;
    keep access_type == Coherent => me.Cmd[0:0] == 1'b1;
    SeqID            : uint (bits:4);
    UnitID           : uint (bits:5);
    PassPW           : bit;
    keep passing_rule == NoPassPW => me.PassPW == 0;
    keep passing_rule == PassPW => me.PassPW == 1;
    SrcTag_Rsv       : uint (bits:5);
    Compat           : bit;
    Mask_Count       : uint (bits:4);
    keep gen (Addr) before (Mask_Count);
    // Ensure that a Read does not cross a 64 byte boundary on a DoubleWord transaction.
    keep data_type == DoubleWord => Mask_Count < 5'b10000 - Addr[5:2];
    Addr             : uint (bits:40);
    keep Addr[1:0] == 2'b00;

    UnitID():uint (bits:5) is only {
      result = UnitID;
    };
    SeqID():uint (bits:4) is only {
      result = SeqID;
    };
    CompatBit():bit is only {
      result = Compat;
    };
    Mask_Count():uint (bits:4) is only {
      result = Mask_Count;
    };
    Address():uint (bits:40) is only {
      result = Addr;
    };
    DataType():csco_ht_packet_encoding is only {
      if(Cmd[2:2] == 0) {
        result = Byte_Read;
      }else {
        result = DoubleWord_Read;
      };
    };
    SrcTag():uint (bits:5) is only {
      result = SrcTag_Rsv;
    };
    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : %s %s %s : SeqID=0x%0x UnitID=0x%0x PassPW=0x%0x SrcTag=0x%0x Compat=0x%0x Mask_Count=0x%0x Addr=0x%010x RespPassPW=0x%0x",
                              command, data_type, BW_latency_req, access_type, SeqID, UnitID, PassPW, SrcTag_Rsv, Compat, Mask_Count, Addr, Cmd[3:3]);
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, SrcTag_Rsv, Compat, Mask_Count, Addr[39:2]);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, SrcTag_Rsv, Compat, Mask_Count, Addr[39:2]);
      Addr[1:0] = 2'b00;
      if(Cmd[3:3] == 0) {
        response_passing_rule = NoPassPW;
      }else {
        response_passing_rule = PassPW;
      };
      if(Cmd[2:2] == 0) {
        data_type = Byte;
      }else {
        data_type = DoubleWord;
      };
      if(Cmd[1:1] == 0) {
        BW_latency_req = Normal;
      }else {
        BW_latency_req = Isochronous;
      };
      if(Cmd[0:0] == 0) {
        access_type = NonCoherent;
      }else {
        access_type = Coherent;
      };
      if(PassPW == 0) {
        passing_rule = NoPassPW;
      }else {
        passing_rule = PassPW;
      };
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when RdResponse'command csco_ht_packet {
    keep packet_type == Ctl;
    keep vChan == Response;
    keep passing_rule == NoPassPW or passing_rule == PassPW;

    keep me.packet_length == 4;
    Cmd              : uint (bits:6);
    keep me.Cmd == 6'b110000;
    UnitID           : uint (bits:5);
    Bridge           : bit;
    PassPW           : bit;
    keep passing_rule == NoPassPW => me.PassPW == 0;
    keep passing_rule == PassPW => me.PassPW == 1;
    SrcTag           : uint (bits:5);
    ErrorFlag        : bit;
    Count            : uint (bits:4);
    NXA              : bit;
    Rsv              : uint (bits:8);
    keep soft me.Rsv == 0;

    // This is a pointer to the associated data packet.
    assoc_data : csco_ht_packet;
    keep Count == 0 => assoc_data.command == Byte_Read or assoc_data.command == DoubleWord_Read;
    keep Count != 0 => assoc_data.command == DoubleWord_Read;
    keep assoc_data.vChan == vChan;
    keep assoc_data.packet_length == (Count + 1) * 4;
    keep assoc_data.endian_type == endian_type;

    UnitID():uint (bits:5) is only {
      result = UnitID;
    };
    IOStream():uint (bits:5) is only {
      if(Bridge == 1) {
        // If this was sent from a bridge, then the UnitID is the source of the corresponding request.
        result = 0;
      }else {
        result = UnitID;
      };
    };
    BridgeBit():bit is only {
      result = Bridge;
    };
    SrcTag():uint (bits:5) is only {
      result = SrcTag;
    };
    Mask_Count():uint (bits:4) is only {
      result = Count;
    };
    ErrorFlag():bit is only {
      result = ErrorFlag;
    };
    NXA():bit is only {
      result = NXA;
    };
    attach_data_packet(new_data : csco_ht_packet) is only {
      assoc_data = new_data;
    };
    retrieve_data_packet():csco_ht_packet is only {
      result = assoc_data;
    };
    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : UnitID=0x%0x Bridge=0x%0x PassPW=0x%0x SrcTag=0x%0x Error=0x%0x Count=0x%0x NXA=0x%0x",
                              command, UnitID, Bridge, PassPW, SrcTag, ErrorFlag, Count, NXA);
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, Rsv[1:0], UnitID, Rsv[2:2], Bridge, PassPW, SrcTag, ErrorFlag, Count, Rsv[5:3], NXA, Rsv[7:6]);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, Rsv[1:0], UnitID, Rsv[2:2], Bridge, PassPW, SrcTag, ErrorFlag, Count, Rsv[5:3], NXA, Rsv[7:6]);
      if(PassPW == 0) {
        passing_rule = NoPassPW;
      }else {
        passing_rule = PassPW;
      };
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when TgtDone'command csco_ht_packet {
    keep packet_type == Ctl;
    keep vChan == Response;
    keep passing_rule == NoPassPW or passing_rule == PassPW;

    keep me.packet_length == 4;
    Cmd              : uint (bits:6);
    keep me.Cmd == 6'b110011;
    UnitID           : uint (bits:5);
    Bridge           : bit;
    PassPW           : bit;
    keep passing_rule == NoPassPW => me.PassPW == 0;
    keep passing_rule == PassPW => me.PassPW == 1;
    SrcTag           : uint (bits:5);
    ErrorFlag        : bit;
    NXA              : bit;
    Rsv              : uint (bits:12);
    keep soft me.Rsv == 0;

    UnitID():uint (bits:5) is only {
      result = UnitID;
    };
    IOStream():uint (bits:5) is only {
      if(Bridge == 1) {
        // If this was sent from a bridge, then the UnitID is the source of the corresponding request.
        result = 0;
      }else {
        result = UnitID;
      };
    };
    BridgeBit():bit is only {
      result = Bridge;
    };
    SrcTag():uint (bits:5) is only {
      result = SrcTag;
    };
    ErrorFlag():bit is only {
      result = ErrorFlag;
    };
    NXA():bit is only {
      result = NXA;
    };
    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : UnitID=0x%0x Bridge=0x%0x PassPW=0x%0x SrcTag=0x%0x Error=0x%0x NXA=0x%0x",
                              command, UnitID, Bridge, PassPW, SrcTag, ErrorFlag, NXA);
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, Rsv[1:0], UnitID, Rsv[2:2], Bridge, PassPW, SrcTag, ErrorFlag, Rsv[9:3], NXA, Rsv[11:10]);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, Rsv[1:0], UnitID, Rsv[2:2], Bridge, PassPW, SrcTag, ErrorFlag, Rsv[9:3], NXA, Rsv[11:10]);
      if(PassPW == 0) {
        passing_rule = NoPassPW;
      }else {
        passing_rule = PassPW;
      };
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Broadcast'command csco_ht_packet {
    keep packet_type == Ctl;
    keep vChan == NonPosted or vChan == Posted;
    keep passing_rule == NoPassPW or passing_rule == PassPW;

    // Support for interrupt packets
    interrupt : bool;
    keep soft interrupt == FALSE;
    keep Addr[39:24] == 16'hFDF8 => interrupt == TRUE; // Interrupt space in System Address Map
    keep interrupt == TRUE => vChan == Posted;

    keep me.packet_length == 8;
    // Posted Broadcasts do not require a response, NonPosted Broadcasts require a response.
    Cmd              : uint (bits:6);
    // This manner of constraint is not compatible with v3.2.4, v3.3 or higher needs to be used.
    keep me.Cmd[5:1] == 5'b11101;
    keep vChan == NonPosted => me.Cmd[0:0] == 1'b1;
    keep vChan == Posted => me.Cmd[0:0] == 1'b0;
    SeqID            : uint (bits:4);
    UnitID           : uint (bits:5);
    PassPW           : bit;
    keep passing_rule == NoPassPW => me.PassPW == 0;
    keep passing_rule == PassPW => me.PassPW == 1;
    SrcTag_Rsv       : uint (bits:5);
    // The SrcTag_Rsv field is reserved for Posted Writes.
    keep vChan == Posted => SrcTag_Rsv == 5'b00000;
    Rsv              : uint (bits:5);
    keep soft me.Rsv == 0;
    Addr             : uint (bits:40);
    keep Addr[1:0] == 2'b00;

    MT               : csco_ht_interrupt_type;
    keep MT == EOI;
    Vector           : byte;
    keep interrupt == TRUE => Addr[39:24] == 16'hFDF8; // Interrupt space in System Address Map
    // Map the interrupt specific fields onto the Addr.
    keep interrupt == TRUE => Addr[23:16] == Vector;
    keep interrupt == TRUE => Addr[15:5] == 11'b00000000000;
    keep interrupt == TRUE => Addr[4:2] == MT.as_a(uint (bits:3));

    UnitID():uint (bits:5) is only {
      result = UnitID;
    };
    SeqID():uint (bits:4) is only {
      result = SeqID;
    };
    SrcTag():uint (bits:5) is only {
      if(Cmd[0:0] == 1) {
        result = SrcTag_Rsv;
      }else {
        dut_error("SrcTag() was called an a Posted Broadcast, this field is reserved.");
      };
    };
    Address():uint (bits:40) is only {
      result = Addr;
    };
    MessageType():csco_ht_interrupt_type is only {
      if(interrupt == TRUE) {
        result = MT;
      }else {
        dut_error("MessageType() was called on a Broadcast packet that is not an EOI.");
      };
    };
    Vector():byte is {
      if(interrupt == TRUE) {
        result = Vector;
      }else {
        dut_error("Vector() was called on a Broadcast packet that is not an EOI.");
      };
    };
    create_csco_ht_packet_string() is {
      if(interrupt == TRUE) {
        packet_string = appendf("%s : SeqID=0x%0x UnitID=0x%0x PassPW=0x%0x MT=0x%0x Vector=0x%0x Addr=0x%010x",
                                command, SeqID, UnitID, PassPW, MT, Vector, Addr);
      }else {
        packet_string = appendf("%s : SeqID=0x%0x UnitID=0x%0x PassPW=0x%0x SrcTag=0x%0x Addr=0x%010x",
                                command, SeqID, UnitID, PassPW, SrcTag_Rsv, Addr);
      };
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, SrcTag_Rsv, Rsv, Addr[39:2]);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, SrcTag_Rsv, Rsv, Addr[39:2]);
      if(Addr[39:24] == 16'hFDF8) {
        interrupt = TRUE; // Interrupt space in System Address Map
      };
      Addr[1:0] = 2'b00;
      if(interrupt == TRUE) {
        // Map the interrupt specific fields onto the Addr.
        Vector = Addr[23:16];
        MT = Addr[4:2].as_a(csco_ht_interrupt_type);
      };
      if(Cmd[0:0] == 0) {
        vChan = Posted;
      }else {
        vChan = NonPosted;
      };
      if(PassPW == 0) {
        passing_rule = NoPassPW;
      }else {
        passing_rule = PassPW;
      };
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Fence'command csco_ht_packet {
    keep packet_type == Ctl;
    keep vChan == Posted;
    keep passing_rule == NoPassPW or passing_rule == PassPW;

    keep me.packet_length == 4;
    Cmd              : uint (bits:6);
    keep me.Cmd == 6'b111100;
    SeqID            : uint (bits:4);
    UnitID           : uint (bits:5);
    PassPW           : bit;
    keep passing_rule == NoPassPW => me.PassPW == 0;
    keep passing_rule == PassPW => me.PassPW == 1;
    Rsv              : uint (bits:16);
    keep soft me.Rsv == 0;

    UnitID():uint (bits:5) is only {
      result = UnitID;
    };
    IOStream():uint (bits:5) is only {
      result = UnitID;
    };
    SeqID():uint (bits:4) is only {
      result = SeqID;
    };
    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : SeqID=0x%0x UnitID=0x%0x PassPW=0x%0x",
                              command, SeqID, UnitID, PassPW);
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, Rsv);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, SeqID[3:2], UnitID, SeqID[1:0], PassPW, Rsv);
      if(PassPW == 0) {
        passing_rule = NoPassPW;
      }else {
        passing_rule = PassPW;
      };
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Sync'command csco_ht_packet {
    keep packet_type == Info;
    keep vChan == NA;
    keep passing_rule == NA;

    keep me.packet_length == 4;
    Cmd              : uint (bits:6);
    keep me.Cmd == 6'b111111;
    Sync             : uint (bits:26);
    keep me.Sync == 26'h3FFFFFF;

    create_csco_ht_packet_string() is {
      packet_string = appendf("%s", command);
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, Sync);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, Sync);
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Create a Reserved packet that is the same as a sync packet for completeness.
  when Reserved'command csco_ht_packet {
    keep packet_type == Info;
    keep vChan == NA;
    keep passing_rule == NA;

    keep me.packet_length == 4;
    Cmd              : uint (bits:6);
    keep me.Cmd == 6'b111111;
    Sync             : uint (bits:26);
    keep me.Sync == 26'h3FFFFFF;

    create_csco_ht_packet_string() is {
      packet_string = appendf("%s", command);
    };
    post_generate() is also {
      packed_data = pack(packing.low, Cmd, Sync);
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      unpack(packing.low, packed_data, Cmd, Sync);
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Data Packets
  ////////////////////////////////////////////////////////////////////////////////
  // Double Word reads and writes have data packets that range from 1 to 16
  // DW's of data.

  // Byte reads always return a single DW of data.  The portion of data that
  // is valid is determined by the Mask field in the original request.

  // Byte writes contain one DW of mask information followed by 1 to 8 DW's
  // of data.  The validity of the data is determined by the mask included in
  // the packet.
  ////////////////////////////////////////////////////////////////////////////////
  when DoubleWord_Read'command csco_ht_packet {
    keep packet_type == Data;
    keep vChan == Response;
    keep passing_rule == NA;

    // DW operations transfer from 1 to 16 DW's of data.
    keep me.packet_length in [4..64];
    Data : list of bit;
    keep me.Data.size() == me.packet_length * 8;

    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : Data=0x%08x Packed Data=0x%08x",
                              command, Data[Data.size()-1:0], packed_data[packed_data.size()-1:0]);
    };
    post_generate() is also {
      if(endian_type == little_endian) {
        packed_data = pack(packing.low, Data);
      } else if(endian_type == big_endian) {
        packed_data = pack(packing.csco_ht_big_endian, Data);
      } else if(endian_type == big_endian_bit_match) {
        packed_data = pack(packing.csco_ht_big_endian_bit_match, Data);
      } else {
        dut_error("DoubleWord_Read packet - endian policy undefined!!!");
      };
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      if(endian_type == little_endian) {
        unpack(packing.low, packed_data, Data);
      } else if(endian_type == big_endian) {
        unpack(packing.csco_ht_big_endian, packed_data, Data);
      } else if(endian_type == big_endian_bit_match) {
        unpack(packing.csco_ht_big_endian_bit_match, packed_data, Data);
      } else {
        dut_error("DoubleWord_Read packet - endian policy undefined!!!");
      };

      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Byte_Read'command csco_ht_packet {
    keep packet_type == Data;
    keep vChan == Response;
    keep passing_rule == NA;

    keep me.packet_length == 4;
    Data : list of bit;
    keep me.Data.size() == me.packet_length * 8;

    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : Data=0x%08x Packed Data=0x%08x",
                              command, Data[Data.size()-1:0], packed_data[packed_data.size()-1:0]);
    };
    post_generate() is also {
      if(endian_type == little_endian) {
        packed_data = pack(packing.low, Data);
      } else if(endian_type == big_endian) {
        packed_data = pack(packing.csco_ht_big_endian, Data);
      } else if(endian_type == big_endian_bit_match) {
        packed_data = pack(packing.csco_ht_big_endian_bit_match, Data);
      } else {
        dut_error("DoubleWord_Read packet - endian policy undefined!!!");
      };
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      if(endian_type == little_endian) {
        unpack(packing.low, packed_data, Data);
      } else if(endian_type == big_endian) {
        unpack(packing.csco_ht_big_endian, packed_data, Data);
      } else if(endian_type == big_endian_bit_match) {
        unpack(packing.csco_ht_big_endian_bit_match, packed_data, Data);
      } else {
        dut_error("DoubleWord_Read packet - endian policy undefined!!!");
      };
      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when DoubleWord_Write'command csco_ht_packet {
    keep packet_type == Data;
    keep vChan == NonPosted or vChan == Posted;
    keep passing_rule == NA;

    // DW operations transfer from 1 to 16 DW's of data.
    keep me.packet_length in [4..64];
    Data : list of bit;
    keep me.Data.size() == me.packet_length * 8;

    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : Data=0x%08x Packed Data=0x%08x",
                              command, Data[Data.size()-1:0], packed_data[packed_data.size()-1:0]);
    };
    post_generate() is also {
      if(endian_type == little_endian) {
        packed_data = pack(packing.low, Data);
      } else if(endian_type == big_endian) {
        packed_data = pack(packing.csco_ht_big_endian, Data);
      } else if(endian_type == big_endian_bit_match) {
        packed_data = pack(packing.csco_ht_big_endian_bit_match, Data);
      } else {
        dut_error("DoubleWord_Read packet - endian policy undefined!!!");
      };
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      if(endian_type == little_endian) {
        unpack(packing.low, packed_data, Data);
      } else if(endian_type == big_endian) {
        unpack(packing.csco_ht_big_endian, packed_data, Data);
      } else if(endian_type == big_endian_bit_match) {
        unpack(packing.csco_ht_big_endian_bit_match, packed_data, Data);
      } else {
        dut_error("DoubleWord_Read packet - endian policy undefined!!!");
      };

      create_csco_ht_packet_string();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Byte_Write'command csco_ht_packet {
    keep packet_type == Data;
    keep vChan == NonPosted or vChan == Posted;
    keep passing_rule == NA;

    // A Byte_Write will always transfer a 32 bit mask, and from 0 to 8 DW's of data.
    keep me.packet_length in [4..36];
    // Currently specman 3.3.1 does not support relating a bit slice constraint to itself, so
    // keep me.Mask[3:0] == 4'b0000 => me.Mask[31:4] == 28'h0000_000; does not work.  Hence a temp field is used.
    temp_mask : uint (bits:4);
    temp_flag : bool;
    Mask : uint;
     // If the Byte Mask is not all zero's then there must be one set bit associated with the first DW of data.
    keep me.Mask[3:0] == me.temp_mask;
    keep me.temp_mask == 4'b0000 => me.Mask[31:4] == 28'h000_0000;

    Data : list of bit;
    keep me.Data.size() == (me.packet_length - 4) * 8;

    create_csco_ht_packet_string() is {
      packet_string = appendf("%s : Mask=0x%04x ", command, Mask);
      if(Data.size() != 0) {
        var masked_data: string;
        for i from 0 to (Data.size()/8)-1 do {
          if(Mask[i:i] == 0) {
            masked_data = appendf("//%s", masked_data);
          }else {
            masked_data = appendf("%02x%s", Data[(i+1)*8-1:i*8], masked_data);
          };
        };
        packet_string = appendf("Data=0x%08x Masked Data=%s", Data[Data.size()-1:0], masked_data);
      };
    };
    // Some checking needs to be added here to ensure that the corrollation between Mask and Data
    // fields are correct.
    post_generate() is also {
      if(endian_type == little_endian) {
        packed_data = pack(packing.low, Mask, Data);
      } else if(endian_type == big_endian) {
        packed_data = pack(packing.csco_ht_big_endian, Data, %{Mask});
      } else if(endian_type == big_endian_bit_match) {
        packed_data = pack(packing.csco_ht_big_endian_bit_match, Data, %{Mask});
      } else {
        dut_error("DoubleWord_Read packet - endian policy undefined!!!");
      };

      if(Mask[31:4] != 28'h000_0000) {
        check that Mask[3:0] != 4'b0000 else dut_error("Byte_Write packet has illegal Mask field");
      };
      create_csco_ht_packet_string();
    };
    unpack_packet() is only {
      if(endian_type == little_endian) {
        unpack(packing.low, packed_data, Mask, Data);
      } else if(endian_type == big_endian) {
        unpack(packing.csco_ht_big_endian, packed_data, Data, Mask);
      } else if(endian_type == big_endian_bit_match) {
        unpack(packing.csco_ht_big_endian_bit_match, packed_data, Data, Mask);
      } else {
        dut_error("DoubleWord_Read packet - endian policy undefined!!!");
      };

      if(Mask[31:4] != 28'h000_0000) {
        check that Mask[3:0] != 4'b0000 else dut_error("Byte_Write packet has illegal Mask field");
      };
      create_csco_ht_packet_string();
    };
  };

  ////////////////////////////////////////////////////////////////////////////////
  // Coverage collection routine.
  ////////////////////////////////////////////////////////////////////////////////
  do_cover() is empty;

  when uncovered csco_ht_packet {
    do_cover() is {};
  };

  ////////////////////////////////////////////////////////////////////////////////
  // Check that the size of the packed packet and the expected packet size are the same.
  post_generate() is also {
    check that packed_data.size() == 8 * me.packet_length else dut_error("Generated packet has invalid size.");
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
'>
